#pragma rtGlobals=3

#pragma IgorVersion = 6.20

#pragma IndependentModule = Localizer

#include <Resize Controls>
#include <WaveSelectorWidget>
#include <SaveRestoreWindowCoords>
#include <ImageSlider>

// Igor-related constants
constant kMaxWaveName = 31
constant kUserAbort = 57

// constants for the dimensions of the interface
constant kViewerGraph_EdgeMargin = 10
constant kViewerControls_Width = 477
constant kViewerControls_Height = 40
constant kViewerControls_Margin = 10
constant kAnalysisControls_Width = 284
constant kAnalysisControls_Height = 315
constant kAnalysisControls_Margin = 10
constant kHistogram_Margin = 15
constant kTabHeight = 20

// camera types (for data loading)
constant CAMERA_TYPE_WINSPEC = 0
constant CAMERA_TYPE_ANDOR = 1
constant CAMERA_TYPE_HAMAMATSU = 2
constant CAMERA_TYPE_TIFF = 3
constant CAMERA_TYPE_PDE = 4	// a custom, very simple image format.
constant CAMERA_TYPE_ZEISS = 5		// Zeiss .lsm files. Currently unused since these are really TIFF files
constant CAMERA_TYPE_IGOR_WAVE = 6	// load data from an Igor wave

// threshold methods
constant THRESHOLD_METHOD_GLRT = 0
constant THRESHOLD_METHOD_ISODATA = 1
constant THRESHOLD_METHOD_TRIANGLE = 2
constant THRESHOLD_METHOD_DIRECT = 3
constant THRESHOLD_METHOD_SMOOTHSIGMA = 4

// preprocessing
constant  PREPROCESSOR_NONE = 0
constant  PREPROCESSOR_3X3MEDIAN = 1
constant  PREPROCESSOR_5X5MEDIAN = 2
constant  PREPROCESSOR_1X1GAUSSIAN = 3
constant  PREPROCESSOR_2X2GAUSSIAN = 4
constant  PREPROCESSOR_3X3MEAN = 5
constant  PREPROCESSOR_5X5MEAN = 6

// postprocessing
constant POSTPROCESSOR_NONE = 0
constant POSTPROCESSOR_REMOVE_ISOLATED = 1

// localization methods
constant LOCALIZATION_GAUSS_FITTING = 0
constant LOCALIZATION_GAUSS_FITTING_FIX = 1
constant LOCALIZATION_MULTIPLICATION = 2
constant LOCALIZATION_CENTROID = 3
constant LOCALIZATION_ZEISSPALM = 4
constant LOCALIZATION_ELLIPSOIDAL2DGAUSS = 5
constant LOCALIZATION_MLEWG = 6

// particle finding methods
constant PARTICLEFINDER_ADJACENT4 = 0
constant PARTICLEFINDER_ADJACENT8 = 1
constant PARTICLEFINDER_RADIUS = 2

// particle verification methods
constant PARTICLEVERIFIER_NONE = 0
constant PARTICLEVERIFIER_2DGAUSS = 1
constant PARTICLEVERIFIER_ELLIPSEGAUSS = 2
constant PARTICLEVERIFIER_OVERLAP = 3

// CCD processing methods
constant PROCESSCCDIMAGE_SUBAVG = 0
constant PROCESSCCDIMAGE_DIFFIMAGE = 1
constant PROCESSCCDIMAGE_CONVERTFORMAT = 2
constant PROCESSCCDIMAGE_CROP = 3
constant PROCESSCCDIMAGE_CONVERTPHOTONS = 4

// CCD processing output formats
constant IMAGE_OUTPUT_TYPE_TIFF = 0;
constant IMAGE_OUTPUT_TYPE_COMPR_TIFF = 1
constant IMAGE_OUTPUT_TYPE_IGOR = 2
constant IMAGE_OUTPUT_TYPE_PDE = 3

// CCD images analysis methods
constant ANALYZECCD_SUMMEDINTENSITYTRACE = 0
constant ANALYZECCD_AVERAGETRACE = 1
constant ANALYZECCD_AVERAGE_IMAGE = 2
constant ANALYZECCD_VARIANCE_IMAGE = 3
constant ANALYZECCD_BLEACHING_ANALYSIS = 4

// PALM bitmap error estimators
constant PALMBITMAP_DEV_SAME = 0
constant PALMBITMAP_DEV_FITUNCERTAINTY = 1
constant PALMBITMAP_DEV_GAUSSIANMASK = 2

// structure that makes it more easy to pass data to e.g. the fitting routines
structure FitData
	Variable directThresholdLevel
	Variable cameraType	// currently unused
	variable background		// currently unused
	variable currentImage
	variable minDistanceBetweenParticles	// currently unused
	variable PFA
	variable smoothSigmaFactor
	variable PSFWidth
	variable xSize	// currently unused
	variable ySize	// currently unused
	variable numberOfImages
	variable firstFrameToAnalyze
	variable lastFrameToAnalyze
	variable thresholdMethod
	variable preprocessing
	variable postprocessing
	variable particlefinder
	variable localizationMethod
	string CCDFilePath
	
	// use or don't use a particular particle verification algorithm
	// to not use it set the corresponding value to zero
	// to use it set it to the corresponding constant defined above
	variable PVerOverlap
	variable PVerSymm
	variable PVerEllipse
endstructure

// structure used to specify a progress reporter function to the operations
Constant kLocalizerProgStructVersion = 1
Structure LocalizerProgStruct
	uint32 version
	FUNCREF ProgressFunctionPrototype func
EndStructure

// included for ease of writing and reading .pde files if required
// (provided for convenience; not currently used in this Igor code)
Structure PDEFormatHeader
	uint32 magic			// set to 27
	uint32 version			// set to 1
	uint32 nImages
	uint32 xSize
	uint32 ySize
	uint32 storageType		// see immediately below
EndStructure
// use the following constants to specify the storage type
// (this is an excerpt from the XOP source code)
//const int STORAGE_TYPE_INT4 = 0;
//const int STORAGE_TYPE_UINT4 = 1;
//const int STORAGE_TYPE_INT8 = 2;
//const int STORAGE_TYPE_UINT8 = 3;
//const int STORAGE_TYPE_INT16 = 4;
//const int STORAGE_TYPE_UINT16 = 5;
//const int STORAGE_TYPE_INT32 = 6;
//const int STORAGE_TYPE_UINT32 = 7;
//const int STORAGE_TYPE_INT64 = 8;
//const int STORAGE_TYPE_UINT64 = 9;
//const int STORAGE_TYPE_FP32 = 10;
//const int STORAGE_TYPE_FP64 = 11;

Menu "Macros"
	Submenu "Localizer"
		SubMenu "Read CCD Data"
			"Read CCD data from disk...", /Q, SetupImageViewer_loadFromDisk()
			"Read CCD data from Igor wave...", /Q, SetupImageViewer_IgorWaves()
			"Read entire CCD file into memory...", /Q, ReadCCDFramesIntoMemory()
			"-"
			"Load localized positions from file...", /Q, LoadPositionsFromTextFile()
		End
		SubMenu "Manipulate localized positions"
			"Merge positions...", /Q, MergePositions()
			"Consolidate identical emitters...", /Q, ConsolidateSameEmitters_menu()
			SubMenu "Two-Color Imaging"
				"Create registration map...", /Q, CreateRegistrationMap_Menu()
				"Apply registration map...", /Q, ApplyRegistrationMap_Menu()
			End
			"-"
			"Save positions to text file...", /Q, SavePositionsToTextFile()
		End
		SubMenu "Analyze Localized Positions"
			"L-Clustering analysis...", /Q, DoLClustering()
		End
		SubMenu "Make Localization Images"
			"Generate scatter plot...", /Q, GenerateScatterPlot_Menu()
			"Generate accumulated image...", /Q, MakeAccumulatedImage_Menu()
			"Generate bitmap image...", /Q, MakePALMImage_Menu()
		End
		SubMenu "Make Tracking Images"
			"Plot all tracks...", /Q, MakeParticleTrackingScatterPlot()
		End
		"Save topmost movie as...", /Q, SaveCCDImagesAs()
		"Run batch localization...", /Q, SetupBatchLocalizationPanel()
	End
End

Menu "GraphMarquee"
	SubMenu "Localization Analysis"
		"Average Intensity Trajectory", /Q, MakeAverageTraceFromMarquee()
		"Crop Images", /Q, MakeCroppedStackFromMarquee()
		SubMenu "Clustering"
			"Single cluster", /Q, Clustering_Marquee("Direct")
			"L function", /Q, Clustering_Marquee("LClustering")
		End
	End
End

Menu "TracePopup" 
	"-"
	SubMenu "Particle tracking"
		"Combine two tracks...", /Q, CombineTracks_TracePopup()
		"Split track here...", /Q, SplitTracks_TracePopup()
		"-"
		"Delete track...", /Q, DeleteTrack_TracePopup()
	End
End

Function SetupImageViewer_loadFromDisk()
	variable refNum
	string filePaths, currentFile
	
	Open /D/R/T="????" /MULT=1 RefNum
	if (strlen(S_Filename) == 0)	// The user cancelled the dialog
		return -1
	endif
	filePaths = S_Filename
	variable nFiles = ItemsInList(filePaths, "\r")
	variable i
	
	for (i = 0; i < nFiles; i+=1)
		currentFile = StringFromList(i, filePaths, "\r")
		NewInterfacePanel(currentFile)
	endfor
End

Function SetupImageViewer_IgorWaves()
	
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:Localizer
	
	// make a listbox window to choose from
	DoWindow /F candidateWavesViewer
	if (V_flag == 0)
		NewPanel /K=1/W=(101,286,401,486) /N=candidateWavesViewer as "Choose a matrix wave"
		ListBox ListBoxCandidateWaves,pos={1,2},size={298,183}
		MakeListIntoWaveSelector("candidateWavesViewer", "ListBoxCandidateWaves", content = WMWS_Waves, selectionMode = WMWS_SelectionNonContiguous,  nameFilterProc="WaveSelectorFilter2DAnd3DWaves")
		WS_SetNotificationProc("candidateWavesViewer", "ListBoxCandidateWaves", "wavesViewer_NotificationProc")
		ListBox ListBoxCandidateWaves,userdata(ResizeControlsInfo)= A"!!,<7!!#7a!!#BO!!#AFz!!#](Aon\"Qzzzzzzzzzzzzzz!!#o2B4uAezz"
		ListBox ListBoxCandidateWaves,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#u:Du]k<zzzzzzzzzzz"
		ListBox ListBoxCandidateWaves,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
		SetWindow kwTopWin,hook(ResizeControls)=ResizeControls#ResizeControlsHook
		SetWindow kwTopWin,userdata(ResizeControlsInfo)= A"!!*'\"z!!#BP!!#AWzzzzzzzzzzzzzzzzzzzzz"
		SetWindow kwTopWin,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzzzzzzzzzzzzzzz"
		SetWindow kwTopWin,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzzzzzzzzz!!!"
		
		// if the window was opened previously then restore its size and position
		WC_WindowCoordinatesRestore("candidateWavesViewer")
		SetWindow candidateWavesViewer hook(WindowCoordsHook)=WC_WindowCoordinatesNamedHook
	endif
End

Function WaveSelectorFilter2DAnd3DWaves(aName, contents)
	String aName
	Variable contents
	
	if (contents != WMWS_Waves)	// only accept waves
		return 0
	endif
	
	wave proposedWave = $aName
	
	// only accept if the wave is 2D or 3D
	if ((DimSize(proposedWave, 1) != 0) && (DimSize(proposedWave, 3) == 0))
		return 1
	endif
	return 0
end

Function wavesViewer_NotificationProc(SelectedItem, EventCode)
	String selectedItem
	Variable eventCode
	
	switch (eventCode)
		case WMWS_DoubleClick:
			string wavePaths = WS_SelectedObjectsList("candidateWavesViewer", "ListBoxCandidateWaves")
			DoWindow /K candidateWavesViewer
			variable nWaves = ItemsInList(wavePaths)
			variable i
			
			for (i = 0; i < nWaves; i+=1)
				wave /Z dataWave = $StringFromList(i, wavePaths)
				if (WaveExists(dataWave))
					NewInterfacePanel(StringFromList(i, wavePaths))
				else
					Abort "Unable to find the selected wave!"
				endif
			endfor
			break
		default:
			return 0
	endswitch
End

Function NewInterfacePanel(filePath)
	String filePath
	String windowName
	
	// check if the user already has this file open
	// in another file. If so then do nothing
	if (CheckIfPanelFromSameFileExists(filePath) != 0)
		return 0
	endif
	
	string fileName = ParseFilePath(5, filePath, ":", 0, 1)
	fileName = ParseFilePath(3, fileName, ":", 0, 1)
	
	// get the coordinates for the panel
	variable panelLeft, panelTop, panelRight, panelBottom
	GetNewInterfacePanelCoordinates(panelLeft, panelTop, panelRight, panelBottom)
	
	// get the name for the panel
	// make sure that no panel ever shares the same name
	// (except when the user used an older version of this package)
	// by keeping an incrementing counter
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:Localizer
	variable /G root:Packages:Localizer:V_nextWindowIndex
	NVAR nextWindowIndex = root:Packages:Localizer:V_nextWindowIndex
	
	// if the user is mixing windows created with an older version
	// then the name is not necessarily unique
	// so do some trickery for that
	do
		sprintf windowName, "LocalizerViewer%d", nextWindowIndex
		nextWindowIndex += 1
		DFREF testDF = root:Packages:Localizer:$windowName
	while (DataFolderRefStatus(testDF) != 0)
	
	// now make the panel itself
	NewPanel /K=1 /N=$windowName /W=(panelLeft,panelTop,panelRight,panelBottom) as fileName
	Assert(StringMatch(windowName, S_name) == 1)
	windowName = S_name
	
	// now that we know the name of the panel, set up all of the
	// other options we will use
	NewDataFolder /O root:Packages:Localizer:$windowName
	DFREF windowDataFolder = GetWindowDataFolder(windowName)
	DFREF packageDataFolder = root:Packages:Localizer
	
	// now read the actual image
	ReadCCDImages /C=1 /S=0 /DEST=windowDataFolder:M_CCDImage filePath
	if (V_flag != 0)
		DoWindow /K $windowName
		KillDataFolder /Z windowDataFolder
		DoAlert 0, "unable to read from " + filePath
		return 1
	endif
	Redimension /N=(-1, -1) windowDataFolder:M_CCDImage
	
	string /G windowDataFolder:S_filePath = filePath
	variable /G windowDataFolder:V_nImages
	variable /G windowDataFolder:V_currentImage
	
	variable /G windowDataFolder:V_CCDPixelSize
	variable /G windowDataFolder:V_currentThreshold
	variable /G windowDataFolder:V_PFA
	variable /G windowDataFolder:V_smoothSigmaFactor
	variable /G windowDataFolder:V_gaussianWidth
	variable /G windowDataFolder:V_particleVerifierOverlap
	variable /G windowDataFolder:V_particleVerifierSymm
	variable /G windowDataFolder:V_particleVerifierEllipse
	variable /G windowDataFolder:V_firstFrameInLocalization
	variable /G windowDataFolder:V_lastFrameInLocalization
	variable /G windowDataFolder:V_CCDPixelSize
	variable /G windowDataFolder:V_maxImageIndex
	
	variable /G windowDataFolder:V_trackingMaxJump
	variable /G windowDataFolder:V_trackingMaxBlinking
	variable /G windowDataFolder:V_trackingMaxLength
	variable /G windowDataFolder:V_trackingMinLength
	
	variable /G windowDataFolder:V_SOFInFramesToSkip
	variable /G windowDataFolder:V_SOFILastFrameInAnalysis
	variable /G windowDataFolder:V_SOFInFramesInMovieFrame
	variable /G windowDataFolder:V_SOFInDeconvolutionIterations
	
	variable /G windowDataFolder:V_bleachingMaxJump
	variable /G windowDataFolder:V_driftCorrectionMaxJump
	variable /G windowDataFolder:V_nFramesAverageSubtraction
	variable /G windowDataFolder:V_cameraMultiplication
	variable /G windowDataFolder:V_cameraOffset
	
	NVAR currentImage = windowDataFolder:V_currentImage
	NVAR nImages = windowDataFolder:V_nImages
	nImages = V_numberOfImages
	NVAR maxImageIndex = windowDataFolder:V_maxImageIndex
	maxImageIndex = nImages - 1
	
	NVAR smoothSigmaFactor = windowDataFolder:V_smoothSigmaFactor
	NVAR PFA = windowDataFolder:V_PFA
	NVAR currentThreshold = windowDataFolder:V_currentThreshold
	NVAR gaussianWidth = windowDataFolder:V_gaussianWidth
	NVAR pixelSize = windowDataFolder:V_CCDPixelSize
	NVAR particleVerifierOverlap = windowDataFolder:V_particleVerifierOverlap
	NVAR particleVerifierSymm = windowDataFolder:V_particleVerifierSymm
	NVAR particleVerifierEllipse = windowDataFolder:V_particleVerifierEllipse
	NVAR firstFrameInLocalization = windowDataFolder:V_firstFrameInLocalization
	NVAR lastFrameInLocalization = windowDataFolder:V_lastFrameInLocalization
	
	NVAR trackingMaxJump = windowDataFolder:V_trackingMaxJump
	NVAR trackingMaxBlinking = windowDataFolder:V_trackingMaxBlinking
	NVAR trackingMaxLength =  windowDataFolder:V_trackingMaxLength
	NVAR trackingMinLength =  windowDataFolder:V_trackingMinLength
	
	NVAR SOFInFramesToSkip = windowDataFolder:V_SOFInFramesToSkip
	NVAR SOFIlastFrameInAnalysis = windowDataFolder:V_SOFILastFrameInAnalysis
	NVAR SOFInFramesInMovieFrame = windowDataFolder:V_SOFInFramesInMovieFrame
	NVAR SOFInDeconvolutionIterations = windowDataFolder:V_SOFInDeconvolutionIterations
	
	NVAR bleachingMaxJump = windowDataFolder:V_bleachingMaxJump
	NVAR driftCorrectionMaxJump = windowDataFolder:V_driftCorrectionMaxJump
	NVAR nFramesAverageSubtraction = windowDataFolder:V_nFramesAverageSubtraction
	NVAR cameraMultiplication = windowDataFolder:V_cameraMultiplication
	NVAR cameraOffset = windowDataFolder:V_cameraOffset
	
	//  set some defaults
	pfa = 30
	smoothSigmaFactor = 5
	gaussianWidth = 1.6
	particleVerifierOverlap = 1
	lastFrameInLocalization = nImages - 1
	trackingMaxJump = 5
	trackingMaxLength = inf
	trackingMinLength = 2
	SOFInDeconvolutionIterations = 5
	bleachingMaxJump = 1
	driftCorrectionMaxJump = 1
	cameraMultiplication = 1
	SOFInFramesToSkip = 0
	SOFIlastFrameInAnalysis = nImages - 1
	
	string listParticleVerifiersFunc = GetIndependentModuleName() + "#PMParticleVerificationListItems()"
	string listPositionsFunc = GetIndependentModuleName() + "#GetPossiblePositionsWavesForPM()"
	string listTracksFunc = GetIndependentModuleName() + "#GetPossibleTrackingWaves()"
	
	// add the viewer graph
	DefineGuide /W=$windowName GViewerTop={FT,kViewerGraph_EdgeMargin},GViewerBottom={FB,-(kViewerControls_Margin + kViewerControls_Height + kViewerGraph_EdgeMargin)},GViewerLeft={FL,kViewerGraph_EdgeMargin}
	DefineGuide /W=$windowName GViewerRight={FR,-kViewerGraph_EdgeMargin}
	Display /N=CCDViewer /W=(182,133,546,401)/FG=(GViewerLeft,GViewerTop,GViewerRight,GViewerBottom)/HOST=$windowName
	
	// add the bottom controls in a subwindow
	DefineGuide /W=$windowName GLowerPanelBottom={FB,-kViewerControls_Margin}, GLowerPanelTop={GLowerPanelBottom, -kViewerControls_Height}
	DefineGuide /W=$windowName GLowerPanelLeft={GViewerLeft, 0}, GlowerPanelRight={GViewerRight, 0}
	
	NewPanel /N=CCDViewerControls /HOST=$windowName /FG=(GLowerPanelLeft, GLowerPanelTop, GLowerPanelRight, GLowerPanelBottom)
	SetDrawLayer UserBack
	Slider SLFrameSlider,win=$windowName#CCDViewerControls,pos={9,15},size={298,13},proc=SLViewerSliderProc
	Slider SLFrameSlider,win=$windowName#CCDViewerControls,limits={0,nImages - 1,1},value= 0,side= 0,vert= 0
	ValDisplay VDMaxFrames,pos={385,14},size={39,13},limits={0,0,0},barmisc={0,1000}
	ValDisplay VDMaxFrames,value= #("root:Packages:Localizer:" + windowName + ":V_maxImageIndex")
	Button BTShowAnalysisTools,win=$windowName#CCDViewerControls,pos={429,10},size={30,20},proc=BTToggleAnalysisTools,title=">>"
	SetVariable SVCurrentFrame,win=$windowName#CCDViewerControls,pos={316,13},size={66,15}
	SetVariable SVCurrentFrame,win=$windowName#CCDViewerControls,limits={0,nImages -1 ,0},value=currentImage, proc=SVSetImage,title=" "
	
	// add the panel with the tabcontrol for selecting the analysis type
	// the 'ControlFrame' guides are those that should be used by the actual controls to become available to the user (when the corresponding tab is selected)
	// they will be redefined appropriately by the tab procedure
	DefineGuide /W=$windowName GControlFrameLeft={FR,kAnalysisControls_Margin}, GControlFrameRight={FR,kAnalysisControls_Margin + kAnalysisControls_Width}
	DefineGuide /W=$windowName GControlFrameBottom={FB,-kAnalysisControls_Margin}, GControlFrameTop={GControlFrameBottom,-kAnalysisControls_Height}
	// the 'DisabledControl' guides are those that should be used to hide controls (when another tab is selected)
	DefineGuide /W=$windowName GDisabledControlsLeft={FR,kAnalysisControls_Margin}, GDisabledControlsRight={FR,kAnalysisControls_Margin + kAnalysisControls_Width}
	DefineGuide /W=$windowName GDisabledControlsBottom={FB,-kAnalysisControls_Margin}, GDisabledControlsTop={GControlFrameBottom,-kAnalysisControls_Height}
	DefineGuide /W=$windowName GTabBoxTop={GControlFrameBottom,-kAnalysisControls_Height - kTabHeight}
	NewPanel /N=TabBox /HOST=$windowName /FG=(GControlFrameLeft, GTabBoxTop, GControlFrameRight, GControlFrameBottom)
	TabControl TBSelectControls,win=$windowName#TabBox,pos={0,0},size={283,327},tabLabel(0)="Localization",tabLabel(1)="Tracking",tablabel(2)="SOFI",tabLabel(3)="Other",proc=TBSelectControlsProc
	
	// add the analysis controls (initially hidden)
	DefineGuide /W=$windowName GLocalizationControlsLeft={GControlFrameLeft,0}, GLocalizationControlsRight={GControlFrameRight,0}
	NewPanel /N=AnalysisControls /HOST=$windowName /FG=(GLocalizationControlsLeft, GControlFrameTop, GLocalizationControlsRight,GControlFrameBottom)
	SetDrawLayer UserBack
	SetDrawEnv fsize= 9
	DrawText 54,136,"Particle verification:"
	CheckBox CBThresholdCheck,win=$windowName#AnalysisControls,pos={9,11},size={62,14},proc=CBShowThresholdProc
	CheckBox CBThresholdCheck,win=$windowName#AnalysisControls,value= 0,title=""
	Button BTAnalyzePALM,win=$windowName#AnalysisControls,pos={199,275},size={70,20},proc=BTAnalyzePALMProc,title="Analyze!"
	CheckBox CBDisplayFittedPositions,win=$windowName#AnalysisControls,pos={9,173},size={120,14},proc=CBDisplayFittedPositionsProc,title="Display Fitted Positions"
	CheckBox CBDisplayFittedPositions,win=$windowName#AnalysisControls,value= 0
	PopupMenu PMSelectPositionsWave,win=$windowName#AnalysisControls,pos={142,171},size={125,20},bodyWidth=125,proc=PMUpdateViewerProc
	PopupMenu PMSelectPositionsWave,win=$windowName#AnalysisControls,mode=1,value=#listPositionsFunc
	SetVariable SVSetPixelSize,win=$windowName#AnalysisControls,pos={9,197},size={195,15},title="CCD pixel size (nm, optional):"
	SetVariable SVSetPixelSize,win=$windowName#AnalysisControls,value= pixelSize
	PopupMenu PMThresholdMethod,win=$windowName#AnalysisControls,pos={142,8},size={125,20},bodyWidth=125,proc=PMUpdateThresholdMethodProc
	PopupMenu PMThresholdMethod,win=$windowName#AnalysisControls,mode=1,value= #"\"GLRT;SmoothSigma;Absolute\"",title="Segmentation algorithm:"
	SetVariable SVSetGaussianWidth,win=$windowName#AnalysisControls,pos={9,214},size={230,15},title="Standard deviation of the PSF (pixels):"
	SetVariable SVSetGaussianWidth,win=$windowName#AnalysisControls,limits={0.1,inf,0.1},value= gaussianWidth,proc=SVUpdateThreshold
	SetVariable SVSetPFA,win=$windowName#AnalysisControls,pos={10,233},size={125,15},disable=0,title="GLRT Insensitivity:"
	SetVariable SVSetPFA,win=$windowName#AnalysisControls,limits={1,inf,1},value= pfa,proc=SVUpdateThreshold
	SetVariable SVSetSmoothSigmaFactor,win=$windowName#AnalysisControls,pos={10,233},size={175,15},disable=1,title="SmoothSigma factor:"
	SetVariable SVSetSmoothSigmaFactor,win=$windowName#AnalysisControls,limits={1e-3,inf,0.1},value= smoothSigmaFactor,proc=SVUpdateThreshold
	SetVariable SVSetThreshold,win=$windowName#AnalysisControls,pos={10,233},size={125,15},title="Threshold:"
	SetVariable SVSetThreshold,win=$windowName#AnalysisControls,value= currentThreshold,disable=1,proc=SVUpdateThreshold
	PopupMenu PMThresholdPreprocessing,win=$windowName#AnalysisControls,pos={27,49},size={240,20},bodyWidth=125,proc=PMUpdateViewerProc,title="Threshold preprocessing: "
	PopupMenu PMThresholdPreprocessing,win=$windowName#AnalysisControls,mode=1,popvalue="None",value= #"\"None;3x3 Median Filter;5x5 Median Filter;1x1 Gaussian Smoothing;2x2 Gaussian Smoothing;3x3 Mean Filter;5x5 Mean Filter\""
	PopupMenu PMThresholdPostProcessing,win=$windowName#AnalysisControls,pos={22,72},size={246,20},bodyWidth=125,proc=PMUpdateViewerProc,title="Threshold postprocessing: "
	PopupMenu PMThresholdPostProcessing,win=$windowName#AnalysisControls,mode=1,popvalue="None",value= #"\"None;Remove Isolated Pixels\""
	CheckBox CBShowParticles,win=$windowName#AnalysisControls,pos={27,31},size={128,14},proc=CBCheckShowParticlesProc,title="Show estimated particles"
	CheckBox CBShowParticles,win=$windowName#AnalysisControls,value= 0,disable=0
	PopupMenu PMParticleFinder,win=$windowName#AnalysisControls,pos={73,95},size={195,20},bodyWidth=125,proc=PMUpdateViewerProc,title="Particle finding:"
	PopupMenu PMParticleFinder,win=$windowName#AnalysisControls,mode=1,value= #"\"8-way adjacency;4-way adjacency\""
	PopupMenu PMParticleVerification,win=$windowName#AnalysisControls,pos={53,119},size={215,20},bodyWidth=125,proc=PMParticleVerificationProc,title="Choose..."
	PopupMenu PMParticleVerification,win=$windowName#AnalysisControls,mode=0,value=#listParticleVerifiersFunc
	SetVariable SVNFramesToSkip,win=$windowName#AnalysisControls,pos={10,254},size={130,15},bodyWidth=60,title="Frames between",value=firstFrameInLocalization
	SetVariable SVNFramesToSkip,win=$windowName#AnalysisControls,limits={0, nImages - 1, 1}
	SetVariable SVLastFrameToInclude,win=$windowName#AnalysisControls,pos={185,254},size={80,15},bodyWidth=60,title="and"
	SetVariable SVLastFrameToInclude,win=$windowName#AnalysisControls,value=lastFrameInLocalization,limits={0, nImages - 1, 1}
	PopupMenu PMLocalizationMethod,win=$windowName#AnalysisControls,pos={41,143},size={227,20},bodyWidth=125,title="Localization algorithm:",proc=PMUpdateViewerProc
	PopupMenu PMLocalizationMethod,win=$windowName#AnalysisControls,mode=1,popvalue="Gaussian Fitting",value= #"\"Gaussian Fitting;Gaussian Fitting (Fixed Width);Ellipsoidal Gaussian Fitting;Iterative Multiplication;Center-of-Mass;MLEwG\""
	Button BTCopyPositions,win=$windowName#AnalysisControls,pos={11,275},size={100,20},proc=BTCopyPositionsProc,title="Copy Positions"
	
	// add the particle tracking controls (initially inactive)
	DefineGuide /W=$windowName GTrackingControlsLeft={GDisabledControlsLeft,0}, GTrackingControlsRight={GDisabledControlsRight,0}
	NewPanel /N=TrackingControls /HOST=$windowName /FG=(GTrackingControlsLeft, GControlFrameTop, GTrackingControlsRight, GControlFrameBottom)
	GroupBox GBMakeTrack,win=$windowName#TrackingControls,pos={10,12},size={262,135},title="Particle tracking"
	GroupBox GBMakeTrack,frame=0
	SetVariable SVTrackingMaxJump,win=$windowName#TrackingControls,pos={20,57},size={178,15},bodyWidth=72,title="Max jump distance (px)"
	SetVariable SVTrackingMaxJump,limits={0,inf,0.1},value=trackingMaxJump
	Button BTDoTracking,win=$windowName#TrackingControls,pos={209,118},size={55,20},proc=BTRunParticleTrackingProc,title="Do it"
	SetVariable SVTrackingMaxBlinking,win=$windowName#TrackingControls,pos={29,80},size={169,15},bodyWidth=72,title="Max blinking (frames)"
	SetVariable SVTrackingMaxBlinking,limits={0,inf,1},value=trackingMaxBlinking
	PopupMenu PMSelectPosWaveForTracking,win=$windowName#TrackingControls,pos={28,31},size={235,20},bodyWidth=150,title="Positions to track: "
	PopupMenu PMSelectPosWaveForTracking,mode=1,value= #listPositionsFunc
	SetVariable SVTrackMinTrackLength,win=$windowName#TrackingControls,pos={52,101},size={146,15},bodyWidth=72,title="Min track length"
	SetVariable SVTrackMinTrackLength,limits={2,inf,1},value=trackingMinLength
	SetVariable SVTrackMaxTrackLength,win=$windowName#TrackingControls,pos={50,122},size={148,15},bodyWidth=72,title="Max track length"
	SetVariable SVTrackMaxTrackLength,limits={0,inf,1},value=trackingMaxLength
	CheckBox CBDisplayTracks,win=$windowName#TrackingControls,pos={19,159},size={109,14},proc=CBDisplayFittedPositionsProc,title="Display Fitted Tracks"
	CheckBox CBDisplayTracks,win=$windowName#TrackingControls,value= 0
	CheckBox CBDisplayTrackIdentifier,win=$windowName#TrackingControls,pos={37,179},size={115,14},proc=CBDisplayFittedPositionsProc,title="Display track identifier"
	CheckBox CBDisplayTrackIdentifier,win=$windowName#TrackingControls,value=0,proc=CBDisplayFittedPositionsProc
	CheckBox RBDisplayFullTracks,win=$windowName#TrackingControls,pos={37,198},size={126,14},title="Display tracks in entirety"
	CheckBox RBDisplayFullTracks,win=$windowName#TrackingControls,value= 0,mode=1,proc=RBParticleTrackDisplayProc
	CheckBox RBDisplayUpdatingTracks,win=$windowName#TrackingControls,pos={37,217},size={210,14},title="Tracks are constructed with movie playback"
	CheckBox RBDisplayUpdatingTracks,win=$windowName#TrackingControls,value= 1,mode=1,proc=RBParticleTrackDisplayProc
	PopupMenu PMSelectTrackWave,win=$windowName#TrackingControls,pos={142,157},size={125,20},bodyWidth=125,proc=PMUpdateViewerProc
	PopupMenu PMSelectTrackWave,win=$windowName#TrackingControls,mode=1,value=#listTracksFunc
	
	// add the SOFI controls (initially inactive)
	DefineGuide /W=$windowName GSOFIControlsLeft={GDisabledControlsLeft,0}, GSOFIControlsRight={GDisabledControlsRight,0}
	NewPanel /N=SOFIControls /HOST=$windowName /FG=(GSOFIControlsLeft, GControlFrameTop, GSOFIControlsRight, GControlFrameBottom)
	GroupBox GBSOFISettings,win=$windowName#SOFIControls,pos={10,12},size={262,167},title="SOFI Calculation"
	GroupBox GBSOFISettings,win=$windowName#SOFIControls,frame=0
	PopupMenu PMSOFIOrder,win=$windowName#SOFIControls,pos={21,31},size={120,20},title="Order: "
	PopupMenu PMSOFIOrder,win=$windowName#SOFIControls,mode=1,value="2;3;"
	CheckBox RBAutoSOFI,win=$windowName#SOFIControls,pos={169,33},size={86,14},title="Autocorrelation"
	CheckBox RBAutoSOFI,win=$windowName#SOFIControls,value= 0,mode=1,proc=RBSOFICorrelationModeProc
	CheckBox RBCrossSOFI,win=$windowName#SOFIControls,pos={169,55},size={88,14},title="Crosscorrelation"
	CheckBox RBCrossSOFI,win=$windowName#SOFIControls,value= 1,mode=1,proc=RBSOFICorrelationModeProc
	CheckBox CBSOFIDoAverageImage,win=$windowName#SOFIControls,pos={21,56},size={103,14},title="Also average image"
	CheckBox CBSOFIDoAverageImage,win=$windowName#SOFIControls,value= 0
	SetVariable SVSOFINFramesToSkip,win=$windowName#SOFIControls,pos={19,78},size={136,15},bodyWidth=60,title="Frames between"
	SetVariable SVSOFINFramesToSkip,win=$windowName#SOFIControls,value=SOFInFramesToSkip,limits={0, nImages - 1, 1}
	SetVariable SVSOFILastFrameToInclude,win=$windowName#SOFIControls,pos={185,78},size={80,15},bodyWidth=60,title="and"
	SetVariable SVSOFILastFrameToInclude,win=$windowName#SOFIControls,value=SOFIlastFrameInAnalysis,limits={0, nImages - 1, 1}
	CheckBox CBSOFIDropSaturatedFrames,win=$windowName#SOFIControls,pos={19,101},size={203,14},size={203,14},title="Drop frames with saturated pixels (spikes)"
	CheckBox CBSOFIDropSaturatedFrames,win=$windowName#SOFIControls,value= 1
	CheckBox CBSOFIMakeMovie,win=$windowName#SOFIControls,pos={19,123},size={72,14},title="Make movie:"
	CheckBox CBSOFIMakeMovie,win=$windowName#SOFIControls,value= 0
	SetVariable SVSOFInFramesInMovie,win=$windowName#SOFIControls,pos={97,123},size={168,15},bodyWidth=60,title=" frames for every image"
	SetVariable SVSOFInFramesInMovie,win=$windowName#SOFIControls,limits={2, nImages,1},value=SOFInFramesInMovieFrame
	Button BTDoSOFI,win=$windowName#SOFIControls,pos={209,145},size={55,20},proc=BTDoSOFIProc,title="Do it"
	GroupBox GBSOFIDeconvolution,win=$windowName#SOFIControls,pos={10,192},size={262,90},title="Richardson-Lucy Deconvolution"
	GroupBox GBSOFIDeconvolution,win=$windowName#SOFIControls,frame=0
	SetVariable SVSetGaussianWidth,win=$windowName#SOFIControls,pos={16,210},size={230,15},title="Standard deviation of the PSF (pixels):"
	SetVariable SVSetGaussianWidth,win=$windowName#SOFIControls,limits={0.1,inf,0.1},value=gaussianWidth
	SetVariable SVDeconvolutionIterations,win=$windowName#SOFIControls,pos={16,231},size={230,15},title="Number of iterations:"
	SetVariable SVDeconvolutionIterations,win=$windowName#SOFIControls,limits={1,inf,1},value=SOFInDeconvolutionIterations
	Button BTDoSOFIDeconvolution,win=$windowName#SOFIControls,pos={209,255},size={55,20},proc=BTDoSOFIDeconvolutionProc,title="Do it"
	
	// add the 'other' analysis controls (initially inactive)
	DefineGuide /W=$windowName GOtherControlsLeft={GDisabledControlsLeft,0}, GOtherControlsRight={GDisabledControlsRight,0}
	NewPanel /N=OtherControls /HOST=$windowName /FG=(GOtherControlsLeft, GControlFrameTop, GOtherControlsRight, GControlFrameBottom)
	SetDrawLayer UserBack
	SetDrawEnv fsize= 9,fstyle= 2
	DrawText 20,132,"Other settings will be taken from 'Localization' tab"
	GroupBox GBBasicAnalysis,win=$windowName#OtherControls, pos={10,12},size={262,54},title="Basic analysis",frame=0
	PopupMenu PMChooseCCDAnalysisMethod,win=$windowName#OtherControls,pos={20,36},size={175,20},bodyWidth=175
	PopupMenu PMChooseCCDAnalysisMethod,win=$windowName#OtherControls,mode=1,value= #"\"Average Intensity Trace;Summed Intensity Trace;Average Image;Variance Image;\""
	Button BTDoBasicAnalysis,win=$windowName#OtherControls,pos={207,36},size={55,20},title="Do it",proc=BTBasicAnalysisProc
	GroupBox GBBleaching,win=$windowName#OtherControls,pos={10,72},size={262,72},title="Bleaching analysis",frame=0
	SetVariable SVBleachingMaxJump,win=$windowName#OtherControls,pos={20,96},size={178,15},limits={0,100,0.1},bodyWidth=72,title="Max jump distance (px)",value=bleachingMaxJump
	Button BTDoBleaching,win=$windowName#OtherControls,pos={207,93},size={55,20},title="Do it",proc=BTBleachingAnalysisProc
	GroupBox GBDriftCorrection,win=$windowName#OtherControls,pos={10,152},size={262,72},title="Drift correction",frame=0
	PopupMenu PMDriftCorrectionMode,win=$windowName#OtherControls,pos={19,173},size={118,20},title="Mode:",mode=1,value= #"\"Automatic;Manual\""
	PopupMenu PMDriftCorrSelectPositionsWave,win=$windowName#OtherControls,pos={19,199},size={244,20},bodyWidth=150,title="Positions to correct: "
	PopupMenu PMDriftCorrSelectPositionsWave,win=$windowName#OtherControls,mode=1,value= #listPositionsFunc
	Button BTDoDriftCorrection,win=$windowName#OtherControls,pos={207,173},size={55,20},proc=BTCorrectDriftProc,title="Do it"
	GroupBox GBMovieProcessing,win=$windowName#OtherControls, pos={10,232},size={262,72},title="Process movie",frame=0
	PopupMenu PMChooseMovieProcessingMethod,win=$windowName#OtherControls,pos={20,254},size={175,20},bodyWidth=175,proc=PMChooseMovieProcessingProc
	PopupMenu PMChooseMovieProcessingMethod,win=$windowName#OtherControls,mode=1,value= #"\"Subtract Pixelwise Average;Difference Image;Convert File Format;Convert to Photons\""
	SetVariable SVFramesAverageSubtraction,win=$windowName#OtherControls,pos={19,281},size={179,15},bodyWidth=72,title="Frames to average over"
	SetVariable SVFramesAverageSubtraction,win=$windowName#OtherControls,limits={0,inf,1},value=nFramesAverageSubtraction
	SetVariable SVCameraMultiplication,win=$windowName#OtherControls,pos={146,285},size={121,15},bodyWidth=60,title="Multiplication"
	SetVariable SVCameraMultiplication,win=$windowName#OtherControls,limits={0,100,0.1},value=cameraMultiplication,disable=1
	SetVariable SVCameraOffset,win=$windowName#OtherControls,pos={20,285},size={91,15},bodyWidth=60,title="Offset"
	SetVariable SVCameraOffset,win=$windowName#OtherControls,limits={0,100,0.1},value=cameraOffset,disable=1
	Button BTDoMovieProcessing,win=$windowName#OtherControls,pos={207,254},size={55,20},title="Do it",proc=BTProcessMovieProc
	
	// add a graph for showing the histogram (initially hidden)
	DefineGuide /W=$windowName GHistogramLeft={GControlFrameLeft, 0}, GHistogramRight={GControlFrameRight, 0}
	DefineGuide /W=$windowName GHistogramTop={GViewerTop, 0}, GHistogramBottom={GTabBoxTop, -kHistogram_Margin}
	Display /HOST=$windowName /N=HistogramViewer /FG=(GHistogramLeft, GHistogramTop, GHistogramRight, GHistogramBottom)
	
	SetWindow $windowName,hook(keycontrol)=ViewerKeyHook
	
	// and append it to the graph
	AppendImage /W=$windowName#CCDViewer windowDataFolder:M_CCDImage
	AdjustGraphForImageDisplay(windowName + "#CCDViewer")
	
	// also set up the histogram
	wave hist = MakeImageHistogram(windowDataFolder:M_CCDImage, windowName)
	AppendToGraph /W=$windowName#HistogramViewer hist
	Label /W=$windowName#HistogramViewer Bottom, "Intensity"
	Label /W=$windowName#HistogramViewer Left, "Occurrence"
End

Function AdjustGraphForImageDisplay(graphNameStr)
	String graphNameStr
	
	Assert(strlen(graphNameStr) != 0)
	
	ModifyGraph /W=$graphNameStr margin(left)=14,margin(bottom)=14,margin(top)=14,margin(right)=14
	ModifyGraph /W=$graphNameStr mirror=2
	ModifyGraph /W=$graphNameStr nticks=4
	ModifyGraph /W=$graphNameStr minor=1
	ModifyGraph /W=$graphNameStr fSize=9
	ModifyGraph /W=$graphNameStr standoff=0
	ModifyGraph /W=$graphNameStr tkLblRot(left)=90
	ModifyGraph /W=$graphNameStr btLen=3
	ModifyGraph /W=$graphNameStr tlOffset=-2
	ModifyGraph /W=$graphNameStr tickUnit=1
End

Function CheckIfPanelFromSameFileExists(filePath)
	string filePath
	
	// called when a new panel is about to be created
	// check if the file requested by the user is not
	// already open in a different window
	
	// return 0 if there is no such window
	// otherwise bring it to the front and return 1
	variable i
	string windowName
	DFREF windowDataFolder
	for (i = 0; ; i+=1)
		windowName = WinName(i, 64, 1)
		if (strlen(windowName) == 0)
			return 0
		endif
		
		if (GrepString(windowName, "LocalizerViewer[0-9]*") == 1)
			// this is a window created by this package
			// check if the filePath is the same as the one
			// the user requests
			windowDataFolder = GetWindowDataFolder(windowName)
			SVAR windowFilePath = windowDataFolder:S_filePath
			
			if (StringMatch(filePath, windowFilePath) == 1)
				DoWindow /F $windowName
				return 1
			endif
		endif
	endfor
End

Function GetNewInterfacePanelCoordinates(left, top, right, bottom)
	variable &left, &top, &right, &bottom
	
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:Localizer
	
	variable /G root:Packages:Localizer:V_ViewerLeftCoordinate
	variable /G root:Packages:Localizer:V_ViewerTopCoordinate
	
	NVAR previousLeft = root:Packages:Localizer:V_ViewerLeftCoordinate
	NVAR previousTop = root:Packages:Localizer:V_ViewerTopCoordinate
	
	variable panelWidth = 505
	variable panelHeight = 535
	variable dx = 30, dy = 30	// offset new window
	
	string screenInfo = StringByKey("SCREEN1", IgorInfo(0))
	variable dummy, displayWidth, displayHeight
	sscanf screenInfo, "DEPTH=%g,RECT=0,0,%g,%g", dummy, displayWidth, displayHeight
	
	if (previousLeft == 0)
		left = 150
		top = 50
	else
		left = previousLeft + dx
		top = previousTop + dy
		if (left + panelWidth > displayWidth)
			// wrap around to the left side of the screen
			left = 150
			top = 50
		endif
		top = (top + panelHeight > displayHeight) ? 50 : top
	endif
	
	previousLeft = left
	previousTop = top
	right = left + panelWidth
	bottom = top + panelHeight
End

Function /DF GetWindowDataFolder(windowName)
	string windowName
	
	// window names can sometimes consist of paths to subwindows
	// e.g. when returned in a control procedure
	// this function takes care of all of that
	string baseWindowName = StringFromList(0, windowName, "#")
	return root:Packages:Localizer:$baseWindowName
End

Function /S GetBaseWindowName(windowName)
	string windowName
	
	// sometimes the window names returned can include subwindows
	// however, I take these into account explicitly
	// so I really want the base name
	return StringFromList(0, windowName, "#")
End

Function /S GetNameOfTopInterfacePanel()
	
	string windowName
	variable i
	
	for (i = 0; ; i+=1)
		windowName = WinName(i, 64, 1)
		if (strlen(windowName) == 0)
			return ""
		endif
		
		if (GrepString(windowName, "LocalizerViewer[0-9]*") == 1)
			return windowName
		endif
	endfor
End

Function ReadCCDFramesIntoMemory()
	
	variable RefNum, nFilesToLoad, i
	variable skipThisFile
	string filePath, filePaths, MacintoshFilePath, fileName
	string wName
	string errorMessage
	
	Open /D/R/MULT=1/T="????" RefNum

	if (strlen(S_Filename) == 0)	// The user cancelled the dialog
		return -1
	endif
	
	filePaths = S_FileName
	nFilesToLoad = ItemsInList(filePaths, "\r")
	
	// add a structure with a reference to the progress reporting function
	STRUCT LocalizerProgStruct progressStruct
	progressStruct.version = kLocalizerProgStructVersion
	FUNCREF ProgressFunctionPrototype progressStruct.func = ProgressReporterFunc
	
	for (i = 0; i < nFilesToLoad; i+=1)
		skipThisFile = 0
		filePath = StringFromList(i, filePaths, "\r")
		MacintoshFilePath = ParseFilePath(5, FilePath, ":", 0, 0)
		
		// get the number of images in the file
		ReadCCDImages /Z /H MacintoshFilePath
		if (V_flag != 0)		// did we fail to read the data?
			errorMessage = "Unable to read from  " + MacintoshFilePath
			Abort errorMessage
		endif
		
		wName = ParseFilePath(3, MacintoshFilePath, ":", 0, 0)
		wName = CleanupName(wName, 0)
		for ( ; ; )
			if (WaveExists(root:'Localizer Movies':$wName))
				wName = GetNewPositionsWaveName(wName, root:'Localizer Movies', "Name already exists, please provide a different one")
				if (strlen(wName) == 0)
					skipThisFile = 1
					nFilesToLoad -= 1
				endif
				break
			else
				break
			endif
		endfor
		
		if (skipThisFile == 1)
			continue
		endif
		
		NewDataFolder /O root:'Localizer Movies'
		ReadCCDImages /O /PROG=progressStruct /DEST=root:'Localizer Movies':$wName MacintoshFilePath
		
		NewInterfacePanel("root:'Localizer Movies':" + wName)
	endfor
End

Function SavePositionsToTextFile()
	
	// get the positions we want to save
	String posName
	Prompt posName, "Positions wave: ", popup, GetPossiblePositionsWaves()
	DoPrompt "Select the positions to be saved", posName
	if (V_flag == 1)
		return 0
	endif
	
	wave /Z pos = GetPositionsWaveReference(posName)
	if (WaveExists(pos) != 1)
		Abort "Internal error: cannot find the requested positions"
	endif
	
	// get the output filepath
	variable refNum
	string outputFilePath
	Open /D /F="Text File (*.txt):.txt;All Files:.*;" refNum
	if (strlen(S_fileName) == 0)
		return 0
	endif
	outputFilePath = S_fileName
	
	// construct a header that will identify the filetype and columns
	string header, columnLabels
	variable localizationMethod = NumberByKey("LOCALIZATION METHOD", note(pos))
	if (NumType(localizationMethod) == 2)
		localizationMethod = LOCALIZATION_GAUSS_FITTING
	endif
	
	switch (localizationMethod)
		case LOCALIZATION_GAUSS_FITTING:
			header = "Localized positions using symmetric 2D Gauss fitting\n"
			columnLabels = "First frame\tIntegrated intensity\tFitted PSF standard deviation\tX position (pixel)\tY position (pixel)\tBackground\tIntensity deviation\tPSF width deviation\tX position deviation\tY position deviation\tBackground deviation\tNumber of frames where this emitter is present\n"
			break
		case LOCALIZATION_GAUSS_FITTING_FIX:
			header = "Localized positions using symmetric 2D Gauss fitting with fixed PSF width\n"
			columnLabels = "First frame\tIntegrated intensity\tX position (pixel)\tY position (pixel)\tBackground\tIntensity deviation\tX position deviation\tY position deviation\tBackground deviation\tNumber of frames where this emitter is present\n"
			break
		case LOCALIZATION_MULTIPLICATION:
			header = "Localized positions using iterative multiplication\n"
			columnLabels = "First frame\tUsed PSF standard deviation\tX position (pixel)\tY position (pixel)\tNumber of frames where this emitter is present\n"
			break
		case LOCALIZATION_CENTROID:
			header = "Localized positions using centroid calculation\n"
			columnLabels = "First frame\tX position (pixel)\tY position (pixel)\tNumber of frames where this emitter is present\n"
			break
		case LOCALIZATION_ZEISSPALM:
			header = "Positions localized with the Zeiss PALM system\n"
			columnLabels = "First frame\tIntegrated number of photons\tX position (pixel)\tY position (pixel)\tLocalization precision (pixel)\tNumber of frames where this emitter is present\n"
			break
		case LOCALIZATION_ELLIPSOIDAL2DGAUSS:
			header = "Localized positions using ellipsoidal 2D Gauss fitting\n"
			columnLabels = "First frame\tIntegrated intensity\tFitted PSF standard deviation along x (pixel)\tFitted PSF standard deviation along y (pixel)\tX position (pixel)\tY position (pixel)\tCorrelation between x and y\t"
			columnLabels += "Background\tIntensity deviation\tPSF width deviation along x (pixel)\tPSF width deviation along y (pixel)\tX position deviation (pixel)\tY position deviation (pixel)\tCorrelation deviation\tBackground deviation\tNumber of frames where this emitter is present\n"
			break
		case LOCALIZATION_MLEWG:
			header = "Localized positions using MLEwG localization\n"
			columnLabels = "First frame\tIntegrated intensity\tFitted PSF standard deviation\tX position (pixel)\tY position (pixel)\tBackground\tPosition deviation\t\tNumber of frames where this emitter is present\n"
			break
		default:
			Abort "Unknown localization method"
	EndSwitch
	
	// open the file for writing
	Open refNum as outputFilePath
	
	// write the metadata
	fprintf refnum, "%s", header
	fprintf refnum, "%s\n", "IGOR WAVENOTE FOLLOWS"
	fprintf refnum, "%s\n", note(pos)
	fprintf refnum, "%s\n", "DATA FOLLOWS"
	fprintf refnum, "%s", columnLabels
	
	Close refNum
	
	// write the actual data using the Igor save operation
	Save /A=2 /J /M="\n" /U={0,0,0,0} pos as outputFilePath
End

Function LoadPositionsFromTextFile()
	// load PALM positions from a text file
	// currently only the Zeiss output format is supported
	
	string inputFilePaths
	variable refNum
	
	// allow the user to select a series of text files
	Open /D/R/F="Text Files:.txt; All Files:.*" /M="Select one or more text files" /MULT=1 refNum
	inputFilePaths = S_fileName
	if (strlen(inputFilePaths) == 0)	// cancelled
		return 0
	endif
	
	// for each of the files, check if it is a Zeiss file or a file created by this software
	string currentFilePath, singleLine
	string ZeissPALMFilepaths = ""
	string thisSoftwareFiles = ""
	variable i
	
	for (i = 0; i < ItemsInList(inputFilePaths, "\r"); i+=1)
		currentFilePath = StringFromList(i, inputFilePaths, "\r")
		Open /Z=1 /R refNum as currentFilePath
		if (V_flag != 0)
			// an error occurred opening this file
			Printf "An error occurred opening the file at %s, skipping\r", currentFilePath
			continue
		endif
		
		// check if it is a valid Zeiss PALM file by looking for the unusual [mm] units specified
		FReadLine refNum, singleLine
		Close refNum
		if (StringMatch(singleLine, "*[mm]*") == 1)
			// probably a Zeiss PALM file
			ZeissPALMFilepaths += currentFilePath + "\r"
			continue
		endif
		
		// check if it is a file created by this software
		if (StringMatch(singleLine, "*Localized positions using*") == 1)
			// probably created by this software
			thisSoftwareFiles += currentFilePath + "\r"
			continue
		endif
		
		// unknown or unsupported file type, ignore
		Printf "The file at %s does not seem to be a file containing PALM positions or is not supported, skipping\r", currentFilePath
		i -= 1
		inputFilePaths = RemoveFromList(inputFilePaths, currentFilePath, "\r")
	endfor
	
	// load the Zeiss files
	if (ItemsInList(ZeissPALMFilepaths, "\r") > 0)
		LoadZeissPALMPositions(ZeissPALMFilepaths)
	endif
	
	// load files created with this software
	if (ItemsInList(thisSoftwareFiles, "\r") > 0)
		LoadXOPPositionsFiles(thisSoftwareFiles)
	endif
End	
			
Function LoadZeissPALMPositions(ZeissPALMFilepaths)
	// load PALM positions from the text output format used by the Zeiss software
	// Zeiss works in units of nanometers, but since we work in pixels we need to convert
	
	// the zeiss positions have their own custom positions format with 6 columns
	// signaled by LOCALIZATION METHOD:LOCALIZATION_ZEISSPALM
	
	// the files to load should be specified as a list separated by carriage returns (\r)
	string ZeissPALMFilepaths
	
	Assert(strlen(ZeissPALMFilepaths) > 0)
	
	variable nDataFiles, pixelSize = 100, nPositions, i, j, k, nEntries, offset, nFrames
	variable xSize = 256, ySize = 256
	string currentMacintoshFilePath, waveOutputName, promptString
	string waveNote
	
	NewDataFolder /O root:'Localized Positions'
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:Localizer
	
	nDataFiles = ItemsInList(ZeissPALMFilepaths, "\r")
	
	for (i = 0; i < nDataFiles; i+=1)
		currentMacintoshFilePath = ParseFilePath(5, StringFromList(i, ZeissPALMFilepaths, "\r"), ":", 0, 0)
		waveOutputName = ParseFilePath(3, currentMacintoshFilePath, ":", 0, 0)
		
		sprintf promptString, "Loading positions from %s", currentMacintoshFilePath
		Prompt pixelSize, "Optical CCD pixel size (in nm)"
		Prompt xSize, "X-size (in pixels)"
		Prompt ySize, "Y-size (in pixels)"
		DoPrompt promptString, pixelSize, xSize, ySize
		if (V_flag == 1)	// cancel
			continue
		endif
		
		LoadWave /Q/A/G/O/B="N='_skip_';N=ZeissFirstFrame,T=96;N=ZeissNFrames,T=96;N=ZeissFramesMissing;N=ZeissXPosition,T=4; N=ZeissYPosition,T=4; N=ZeissPrecision,T=4;N=ZeissPhotons;T=96" currentMacintoshFilePath
		
		wave ZeissFirstFrame, ZeissNFrames, ZeissFramesMissing, ZeissXPosition, ZeissYPosition, ZeissPrecision, ZeissPhotons
		nEntries = DimSize(ZeissFirstFrame, 0)
		
		// The Zeiss output format does not sort the positions according to the ascending frame number in which a positions first appears
		// some routines here assume that the positions are ordered
		// do the sorting now
		Sort ZeissFirstFrame, ZeissFirstFrame, ZeissNFrames, ZeissFramesMissing, ZeissXPosition, ZeissYPosition, ZeissPrecision, ZeissPhotons
		
		nPositions = 0
		for (j = 0; j < nEntries; j += 1)
			if ((ZeissFramesMissing[j] != 0) || (ZeissPhotons[j] == 0))
							// the 'ZeissFramesMissing' column is unusual, so ignore it for now
			continue		// occasionally the Zeiss file contains strange data positions containing zero photons
			endif
			nPositions += 1
		endfor
		
		Make /D/O/N=(nPositions,6) root:'Localized Positions':$waveOutputName
		wave outputWave = root:'Localized Positions':$waveOutputName
		FastOP outputWave = 0
		
		offset = 0
		for (j = 0; j < nEntries; j+=1)
			if ((ZeissFramesMissing[j] != 0) || (ZeissPhotons[j] == 0))
				continue
			endif
			
			outputWave[offset][0] = ZeissFirstFrame[j]
			outputWave[offset][1] = ZeissPhotons[j]
			outputWave[offset][2] = ZeissXPosition[j] / pixelSize * 1e6	// positions are reported in mm by the Zeiss software
			outputWave[offset][3] = ZeissYPosition[j] / pixelSize * 1e6
			outputWave[offset][4] = ZeissPrecision[j] / pixelSize	// precision is reported in nm
			outputWave[offset][5] = ZeissNFrames[j]
			
			offset += 1
		endfor
		
		// append a note with as much additional information as possible
		waveNote = "LOCALIZATION METHOD:" + num2str(LOCALIZATION_ZEISSPALM) + ";"
		waveNote += "ORIGINAL FILE PATH:" + currentMacintoshFilePath + ";"
		waveNote += "X SIZE:" + num2str(xSize) + ";"
		waveNote += "Y SIZE:" + num2str(ySize) + ";"
		waveNote += "NUMBER OF IMAGES:" + num2str(nFrames) + ";"
		waveNote += "X PIXEL SIZE:" + num2str(pixelSize) + ";"
		waveNote += "Y PIXEL SIZE:" + num2str(pixelSize) + ";"
		
		Note /K outputWave	// remove a previous note (should not occur)
		Note /NOCR outputWave, waveNote
	endfor
		
		
	KillWaves /Z ZeissFirstFrame, ZeissNFrames, ZeissFramesMissing, ZeissXPosition, ZeissYPosition, ZeissPrecision, ZeissPhotons
End

Function LoadXOPPositionsFiles(PALMFilePaths)
	// load PALM positions from the text output format used by the Zeiss software
	// Zeiss works in units of nanometers, but since we work in pixels we need to convert
	
	// the zeiss positions have their own custom positions format with 6 columns
	// signaled by LOCALIZATION METHOD:LOCALIZATION_ZEISSPALM
	
	// the files to load should be specified as a list separated by carriage returns (\r)
	string PALMFilePaths
	
	Assert(strlen(PALMFilePaths) > 0)
	
	variable refNum, nDataFiles, pixelSize = 100, nPositions, i, j, k, offset, nFrames
	variable firstLineWithData
	variable localizationMethod
	string currentMacintoshFilePath, waveOutputName, promptString, singleLine, waveNote
	DFREF savDF = GetDataFolderDFR()
	
	NewDataFolder /O root:'Localized Positions'
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:Localizer
	
	nDataFiles = ItemsInList(PALMFilePaths, "\r")
	
	for (i = 0; i < nDataFiles; i+=1)
		currentMacintoshFilePath = ParseFilePath(5, StringFromList(i, PALMFilePaths, "\r"), ":", 0, 0)
		
		waveOutputName = CleanupName(ParseFilePath(3, currentMacintoshFilePath, ":", 0, 0), 0)
		if (strlen(waveOutputName) > 27)
			// loadwave easily complains of a name that's too long, so guard for that
			waveOutputName = waveOutputName[0, 26]
		endif
		
		waveNote = ""
		
		// get some information from the header
		Open /R refNum as currentMacintoshFilePath
		FReadLine refNum, singleLine
		
		// determine the type of positions from the first line
		if (StringMatch(singleLine, "*symmetric 2D Gauss fitting with fixed PSF width*") == 1)
			localizationMethod = LOCALIZATION_GAUSS_FITTING_FIX
			waveNote = "LOCALIZATION METHOD:" + num2str(localizationMethod) + ";"
		elseif (StringMatch(singleLine, "*symmetric 2D Gauss fitting*") == 1)
			localizationMethod = LOCALIZATION_GAUSS_FITTING
			waveNote = "LOCALIZATION METHOD:" + num2str(localizationMethod) + ";"
		elseif (StringMatch(singleLine, "*ellipsoidal 2D Gauss fitting*") == 1)
			localizationMethod = LOCALIZATION_ELLIPSOIDAL2DGAUSS
			waveNote = "LOCALIZATION METHOD:" + num2str(localizationMethod) + ";"
		elseif (StringMatch(singleLine, "*centroid calculation*") == 1)
			localizationMethod = LOCALIZATION_CENTROID
			waveNote = "LOCALIZATION METHOD:" + num2str(localizationMethod) + ";"
		elseif (StringMatch(singleLine, "*iterative multiplication*") == 1)
			localizationMethod = LOCALIZATION_MULTIPLICATION
			waveNote = "LOCALIZATION METHOD:" + num2str(localizationMethod) + ";"
		else
			Printf "The file at %s does not appear to be a valid positions file, skipping...\r", currentMacintoshFilePath
			Close refNum
			continue
		endif
		
		// parse the waveNote from the file, but stop when "DATA FOLLOWS" is encountered
		for (j = 1; ; j+=1)
			FReadLine refNum, singleLine
			if (strlen(singleLine) == 1)	// empty line
				continue
			endif
			if (stringmatch(singleLine, "DATA FOLLOWS*") == 1)
				firstLineWithData = j + 2
				break
			endif
			waveNote += singleLine + ";"
		endfor
		
		Close refNum
		
		// get rid of possible extra carriage returns produced by Igor
		waveNote = ReplaceString("\r", waveNote, "")
		waveNote += "POSITIONS FILE PATH:" + currentMacintoshFilePath + ";"
		
		SetDataFolder root:'Localized Positions'
		LoadWave /Q /A=$waveOutputName /G/ K=1 /M /L={firstLineWithData, firstLineWithData, 0, 0, 0} currentMacintoshFilePath
		SetDataFolder savDF
		
		if (ItemsInList(S_waveNames) != 1)
			Abort "An error occurred loading the data at " + currentMacintoshFilePath
		endif
		
		wave loadedWave = root:'Localized Positions':$StringFromList(0, S_waveNames)
		Note /K loadedWave, waveNote
	endfor
End

Function ChangeCurrentImage(n, windowName)
	variable n
	string windowName
	
	DFREF windowDataFolder = GetWindowDataFolder(windowName)
	NVAR nImages = windowDataFolder:V_nImages
	SVAR filePath = windowDataFolder:S_filePath
	NVAR currentImage = windowDataFolder:V_currentImage
	
	string baseWindowName = GetBaseWindowName(windowName)
	
	if (n < 0)
		n = 0
	endif
	
	if (n >= nImages)
		n = nImages - 1
	endif
	
	ReadCCDImages /O /C=1 /S=(n) /DEST=windowDataFolder:M_CCDImage filePath
	currentImage = n
	
	// the XOP returns a mxnx1 wave, which causes errors when some expect mxn
	// so to handle that, do a redimension here
	Redimension /N=(-1, -1) windowDataFolder:M_CCDImage
	
	GenerateGUIThreshold(baseWindowName)
	
	// update the slider and the histogram
	Slider SLFrameSlider win=$baseWindowName#CCDViewerControls, value=currentImage
	MakeImageHistogram(windowDataFolder:M_CCDImage, windowName)
End

Function ChangeToNextImageAndLoop(windowName)
	string windowName
	
	DFREF windowDataFolder = GetWindowDataFolder(windowName)
	NVAR nImages = windowDataFolder:V_nImages
	NVAR currentImage = windowDataFolder:V_currentImage
	
	string baseWindowName = GetBaseWindowName(windowName)
	
	variable nextImage = (currentImage + 1 >= nImages) ? 0 : currentImage + 1
	ChangeCurrentImage(nextImage, baseWindowName)
End

Function BTViewerProc(ba) : ButtonControl
	STRUCT WMButtonAction &ba
	
	string controlName = ba.ctrlName
	string windowName = ba.win
	
	DFREF windowDataFolder = GetWindowDataFolder(windowName)
	NVAR nImages = windowDataFolder:V_nImages
	NVAR currentImage = windowDataFolder:V_currentImage
	
	switch (ba.eventCode)
		case 1: 	// mouse down	
			strswitch(controlName)
			case "BTNext":
				ChangecurrentImage(currentImage + 1, windowName)
				break
			case "BTPrevious":
				ChangecurrentImage(currentImage - 1, windowName)
				break
			case "BTFirst":
				ChangecurrentImage(0, windowName)
				break
				
			case "BTLast":
				ChangecurrentImage(nImages - 1, windowName)
				break
			endswitch
		break
	EndSwitch
End

Function SLViewerSliderProc(sa) : SliderControl
	STRUCT WMSliderAction &sa
	
	variable curval
	string windowName = sa.win
	
	DFREF windowDataFolder = GetWindowDataFolder(windowName)
	NVAR currentImage = windowDataFolder:V_currentImage

	switch( sa.eventCode )
		case -1: // kill
			break
		default:
			if( sa.eventCode & 1 ) // value set
				curval = sa.curval
				ChangecurrentImage(curval, windowName)
			endif
			break
	endswitch

	return 0
End

Function ViewerKeyHook(s)
	STRUCT WMWinHookStruct &s
	
	string baseWindowName = GetBaseWindowName(s.winName)
	
	DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
	NVAR currentImage = windowDataFolder:V_currentImage
	
	switch (s.eventcode)
		case 11:	// keyboard event
			switch (s.keyCode)
				case 29:	// right arrow
					ChangeCurrentImage(currentImage + 1, baseWindowName)
					break
				case 28:	// left arrow
					ChangeCurrentImage(currentImage -1, baseWindowName)
					break
				case 32:	// space
					ToggleMoviePlayback(baseWindowName)
					break
			EndSwitch
			break
		case 22:	// scrollwheel event
			variable framesToScroll = - s.wheelDy
			ChangeCurrentImage(currentImage + framesToScroll, baseWindowName)
			break
	EndSwitch
End

Function BTToggleAnalysisTools(ba) : ButtonControl
	STRUCT WMButtonAction &ba
	
	string windowName = ba.win
	string controlName = ba.ctrlName
	
	DFREF windowDataFolder = GetWindowDataFolder(windowName)
	
	Variable /G windowDataFolder:V_ControlsAreShown
	NVAR controlsAreShown = windowDataFolder:V_ControlsAreShown
	
	switch (ba.eventCode)
		case 1: 	// mouse down
			if (!controlsAreShown)
				ShowAnalysisControls(windowName)
				Button $controlName, win=$windowName, title="<<"
				controlsAreShown = 1
			else
				HideAnalysisControls(windowName)
				Button $controlName, win=$windowName, title=">>"
				controlsAreShown = 0
			endif
			break
	EndSwitch
	
End

Function ShowAnalysisControls(wName)
	string wName
	
	string windowName = StringFromList(0, wName, "#")
	
	GetWindow $windowName, wsize
	
	variable widthOfControls = kAnalysisControls_Width + kAnalysisControls_Margin
	variable newWidth = (V_Right - V_Left + widthOfControls* 72 / ScreenResolution)
	
	// now redefine the guides
	// due a bug in the way Igor saves recreation macros, GViewerRight must be defined in terms of 
	// FR. The expression below was more clear, but triggers an error upon experiment load
	//DefineGuide /W=$windowName GViewerRight={GControlFrameLeft, -kViewerGraph_EdgeMargin}
	DefineGuide /W=$windowName GViewerRight={FR, - (kAnalysisControls_Margin + kAnalysisControls_Width + kViewerGraph_EdgeMargin)}
	DefineGuide /W=$windowName GControlFrameLeft={FR,- (kAnalysisControls_Margin + kAnalysisControls_Width)}, GControlFrameRight={FR,-kAnalysisControls_Margin}
	
	
	MoveWindow /W=$windowName V_Left, V_Top, V_Left + newWidth, V_Bottom
End

Function HideAnalysisControls(wName)
	string wName
	
	string windowName = StringFromList(0, wName, "#")
	
	GetWindow $windowName, wsize
	
	variable widthOfControls = kAnalysisControls_Width + kAnalysisControls_Margin
	variable newWidth = (V_Right - V_Left - widthOfControls * 72 / ScreenResolution)
	
	// now redefine the guides
	DefineGuide /W=$windowName GViewerRight={FR,-kViewerGraph_EdgeMargin}
	DefineGuide /W=$windowName GControlFrameLeft={FR,kAnalysisControls_Margin}, GControlFrameRight={FR,kAnalysisControls_Margin + kAnalysisControls_Width}
	
	MoveWindow /W=$windowName V_Left, V_Top, V_Left + newWidth, V_Bottom
End

Function TBSelectControlsProc(tca) : TabControl
	STRUCT WMTabControlAction &tca
	
	// handle the user selection of different controls using the tabbox
	// depending on the selection, move the unwanten controls outside the panel
	// and the wanted ones inside

	switch( tca.eventCode )
		case 2: // mouse up
			Variable tab = tca.tab
			string baseWindowName = GetBaseWindowName(tca.win)
			// get the text for the currently selected tab item
			ControlInfo /W=$baseWindowName#TabBox TBSelectControls
			string tabText = S_Value
			
			// start by deactivating all panels, then show the one that is requested
			DefineGuide /W=$baseWindowName GLocalizationControlsLeft={GDisabledControlsLeft,0}, GLocalizationControlsRight={GDisabledControlsRight,0}
			DefineGuide /W=$baseWindowName GTrackingControlsLeft={GDisabledControlsLeft,0}, GTrackingControlsRight={GDisabledControlsRight,0}
			DefineGuide /W=$baseWindowName GSOFIControlsLeft={GDisabledControlsLeft,0}, GSOFIControlsRight={GDisabledControlsRight,0}
			DefineGuide /W=$baseWindowName GOtherControlsLeft={GDisabledControlsLeft,0}, GOtherControlsRight={GDisabledControlsRight,0}
			
			StrSwitch(tabText)
				case "Localization":
					DefineGuide /W=$baseWindowName GLocalizationControlsLeft={GControlFrameLeft,0}, GLocalizationControlsRight={GControlFrameRight,0}
					break
				case "Tracking":
					DefineGuide /W=$baseWindowName GTrackingControlsLeft={GControlFrameLeft,0}, GTrackingControlsRight={GControlFrameRight,0}
					break
				case "SOFI":
					DefineGuide /W=$baseWindowName GSOFIControlsLeft={GControlFrameLeft,0}, GSOFIControlsRight={GControlFrameRight,0}
					break
				case "Other":
					DefineGuide /W=$baseWindowName GOtherControlsLeft={GControlFrameLeft,0}, GOtherControlsRight={GControlFrameRight,0}
					break
				default:
					Abort "Missing handler in TBSelectControlsProc, got " + tabText
					break
			EndSwitch
			break
	endswitch

	return 0
End

Function ToggleMoviePlayback(windowName)
	string windowName
	// see comments below
	
	string BaseWindowName = GetBaseWindowName(windowName)
	String /G root:Packages:Localizer:S_moviesToPlay
	SVAR moviesToPlay = root:Packages:Localizer:S_moviesToPlay
	
	if (FindListItem(baseWindowName, moviesToPlay) == -1)
		// start the movie
		StartPlayingMovie(baseWindowName)
	else
		//stop the movie
		StopPlayingMovie(baseWindowName)
	endif
End

Function StartPlayingMovie(windowName)
	string windowName
	
	// start playing the movie contained in windowName in a loop
	// the strategy used is to add a background task that updates the
	// frame each time it is run
	// a limitation is that the function executed by the task can not take
	// any parameters, so we have kludge our way around that so the
	// task knows which window needs to be updated
	// for now this is done by maintaining a global string variable
	// in the root:Packages:localizer folder, that contains a list
	// of the windows to be run in a movie
	// if this list is empty then the background task will terminate,
	// though manual termination is preferred
	
	string baseWindowName = GetBaseWindowName(windowName)
	// add this windowName to the string variable
	// if it already is in there then that is an error
	String /G root:Packages:Localizer:S_moviesToPlay
	SVAR moviesToPlay = root:Packages:Localizer:S_moviesToPlay
	if (FindListItem(baseWindowName, moviesToPlay) != -1)
		Abort "StopPlayingMovie error: the requested window is already present in S_moviesToPlay"
	endif
	
	moviesToPlay += baseWindowName + ";"
	
	// if there was a window already in the string then the task will be running
	// if not then start it
	if (ItemsInList(moviesToPlay) == 1)
		StartBackGroundMovieTask()
	endif
End

Function StopPlayingMovie(windowName)
	string windowName
	
	SVAR moviesToPlay = root:Packages:Localizer:S_moviesToPlay
	string baseWindowName = GetBaseWindowName(windowName)
	
	// windowName must be in the list, otherwise it's an error
	if (FindListItem(baseWindowName, moviesToPlay) == -1)
		Abort "StopPlayingMovie error: the requested window is not present in S_moviesToPlay"
	endif
	
	moviesToPlay = RemoveFromList(baseWindowName, moviesToPlay)
	
	// if there are no more windows left then stop the task
	if (ItemsInList(moviesToPlay) == 0)
		StopBackGroundMovieTask()
	endif
End

Function StartBackGroundMovieTask()
	SVAR moviesToPlay = root:Packages:Localizer:S_moviesToPlay
	
	Assert(ItemsInList(moviesToPlay) > 0)
	
	variable numTicks = 4 // period in ticks, or 1/60th of second
	CtrlNamedBackground MoviePlayer, period=numTicks, proc=RunBackgroundMovie
	CtrlNamedBackground MoviePlayer, start
End

Function StopBackGroundMovieTask()
	CtrlNamedBackground MoviePlayer, stop
End

Function RunBackgroundMovie(s)
	// this function is called as a background task
	STRUCT WMBackgroundStruct &s
	
	SVAR moviesToPlay = root:Packages:Localizer:S_moviesToPlay
	
	variable nMoviesToPlay = ItemsInList(moviesToPlay)
	if (nMoviesToPlay == 0)
		// no movies to play, stop the task
		return 1
	endif
	
	variable i
	string currentBaseWindowName
	DFREF windowDataFolder
	for (i = 0; i < nMoviesToPlay; i+=1)
		currentBaseWindowName = GetBaseWindowName(StringFromList(i, moviesToPlay))
		ChangeToNextImageAndLoop(currentBaseWindowName)
	endfor
	
	return 0
End

Function CBShowThresholdProc(cba) : CheckBoxControl
	STRUCT WMCheckboxAction &cba

	switch( cba.eventCode )
		case 2: // mouse up
			string baseWindowName = GetBaseWindowName(cba.win)
			GenerateGUIThreshold(baseWindowName)
			break
	endswitch

	return 0
End

Function CBCheckShowParticlesProc(cba) : CheckBoxControl
	STRUCT WMCheckboxAction &cba

	switch( cba.eventCode )
		case 2: // mouse up
			string BaseWindowName = GetBaseWindowName(cba.win)
			GenerateGUIThreshold(BaseWindowName)
			break
	endswitch

	return 0
End

Function SVUpdateThreshold(sva) : SetVariableControl
	STRUCT WMSetVariableAction &sva

	switch( sva.eventCode )
		case 1: // mouse up
		case 2: // Enter key
		case 3: // Live update
			Variable dval = sva.dval
			String sval = sva.sval
			string baseWindowName = GetBaseWindowName(sva.win)
			
			GenerateGUIthreshold(baseWindowName)
			
			break
	endswitch

	return 0
End

Function PMParticleVerificationProc(pa) : PopupMenuControl
	STRUCT WMPopupAction &pa

	switch( pa.eventCode )
		case 2: // mouse up
			Variable popNum = pa.popNum
			String popStr = pa.popStr
			string baseWindowName = GetBaseWindowName(pa.win)
			DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
			
			NVAR particleVerifierOverlap = windowDataFolder:V_particleVerifierOverlap
			NVAR particleVerifierSymm = windowDataFolder:V_particleVerifierSymm
			NVAR particleVerifierEllipse = windowDataFolder:V_particleVerifierEllipse
			
			StrSwitch (popStr)
				case "Remove Overlapping Particles":
					particleVerifierOverlap = !particleVerifierOverlap
					break
				case "Gaussian Fitting":
					particleVerifierSymm = !particleVerifierSymm
					break
				case "Ellipsoidal Gaussian Fitting":
					particleVerifierEllipse = !particleVerifierEllipse
					break
				default:
					Abort "Unknown particle verifier"
			EndSwitch
			
			GenerateGUIThreshold(baseWindowName)
			
			break
	endswitch

	return 0
End

Function /S PMParticleVerificationListItems()
	// this function needs to know which particle verifiers are enabled
	// the problem is that that depends on the particular window we're looking at
	// but this function doesn't have a clean way of finding that out
	// so we will use a bit of a hack, looking at what the top window is
	string windowName = GetNameOfTopInterfacePanel()
	DFREF windowDataFolder = GetWindowDataFolder(windowName)
	
	NVAR particleVerifierOverlap = windowDataFolder:V_particleVerifierOverlap
	NVAR particleVerifierSymm = windowDataFolder:V_particleVerifierSymm
	NVAR particleVerifierEllipse = windowDataFolder:V_particleVerifierEllipse
	
	string listString = ""
	string checkMark = "\\M0:!" + num2char(18)+":"
	
	if (particleVerifierOverlap != 0)
		listString += checkMark + "Remove Overlapping Particles" + ";"
	else
		listString += "Remove Overlapping Particles" + ";"
	endif
	
	if (particleVerifierSymm != 0)
		listString += checkMark + "Gaussian Fitting" + ";"
	else
		listString += "Gaussian Fitting" + ";"
	endif
	
	if (particleVerifierEllipse != 0)
		listString += checkMark + "Ellipsoidal Gaussian Fitting" + ";"
	else
		listString += "Ellipsoidal Gaussian Fitting" + ";"
	endif
	
	return listString
		
End

Function PMUpdateThresholdMethodProc(pa) : PopupMenuControl
	STRUCT WMPopupAction &pa

	switch( pa.eventCode )
		case 2: // mouse up
			string baseWindowName = GetBaseWindowName(pa.win)
			
			SetVariable SVSetThreshold, win=$baseWindowName#AnalysisControls, disable=1
			SetVariable SVSetPFA, win=$baseWindowName#AnalysisControls, disable=1
			SetVariable SVSetSmoothSigmaFactor, win=$baseWindowName#AnalysisControls, disable=1
			
			StrSwitch (pa.popStr)
				case "GLRT":
					SetVariable SVSetPFA, win=$baseWindowName#AnalysisControls, disable=0
					break
				case "Absolute":
					SetVariable SVSetThreshold, win=$baseWindowName#AnalysisControls, disable=0
					break
				case "SmoothSigma":
					SetVariable SVSetSmoothSigmaFactor, win=$baseWindowName#AnalysisControls, disable=0
					break
					break
				default:
					Abort "Unknown segmentation option in PMUpdateThresholdMethodProc"
					break
			EndSwitch
			
			GenerateGUIThreshold(baseWindowName)
			break
	endswitch
	return 0
End

Function PMUpdateViewerProc(pa) : PopupMenuControl
	STRUCT WMPopupAction &pa

	switch( pa.eventCode )
		case 2: // mouse up
			string baseWindowName = GetBaseWindowName(pa.win)
			GenerateGUIThreshold(baseWindowName)
			break
	endswitch
	return 0
End

Function CBDisplayFittedPositionsProc(cba) : CheckBoxControl
	STRUCT WMCheckboxAction &cba

	switch( cba.eventCode )
		case 2: // mouse up
			string baseWindowName = GetBaseWindowName(cba.win)
			GenerateGUIthreshold(baseWindowName)
				
			break
	endswitch

	return 0
End

Function RBParticleTrackDisplayProc(cba) : CheckBoxControl
	STRUCT WMCheckboxAction &cba

	switch( cba.eventCode )
		case 2: // mouse up
			Variable checked = cba.checked
			string controlName = cba.ctrlName
			string baseWindowName = GetBaseWindowName(cba.win)
			
			StrSwitch (controlName)
				case "RBDisplayFullTracks":
					CheckBox RBDisplayFullTracks, win=$baseWindowName#TrackingControls,value=1
					CheckBox RBDisplayUpdatingTracks, win=$baseWindowName#TrackingControls,value=0
					break
				case "RBDisplayUpdatingTracks":
					CheckBox RBDisplayFullTracks, win=$baseWindowName#TrackingControls,value=0
					CheckBox RBDisplayUpdatingTracks, win=$baseWindowName#TrackingControls,value=1
					break
				default:
					Abort "Unknown control name in RBParticleTrackDisplayProc"
					break
			EndSwitch
			
			GenerateGUIThreshold(baseWindowName)
			break
	endswitch

	return 0
End

Function RBSOFICorrelationModeProc(cba) : CheckBoxControl
	STRUCT WMCheckboxAction &cba

	switch( cba.eventCode )
		case 2: // mouse up
			Variable checked = cba.checked
			string controlName = cba.ctrlName
			string baseWindowName = GetBaseWindowName(cba.win)
			
			CheckBox RBAutoSOFI, win=$baseWindowName#SOFIControls, value=0
			CheckBox RBCrossSOFI, win=$baseWindowName#SOFIControls, value=0
			
			StrSwitch (controlName)
				case "RBAutoSOFI":
					CheckBox RBAutoSOFI, win=$baseWindowName#SOFIControls, value=1
					break
				case "RBCrossSOFI":
					CheckBox RBCrossSOFI, win=$baseWindowName#SOFIControls, value=1
					break
				default:
					Abort "Unknown control name in RBSOFICorrelationModeProc"
					break
			EndSwitch
			break
	endswitch

	return 0
End

Function SVSetImage(sva) : SetVariableControl
	STRUCT WMSetVariableAction &sva

	switch( sva.eventCode )
		case 1: // mouse up
		case 2: // Enter key
		case 3: // Live update
			Variable dval = sva.dval
			string baseWindowName = GetBaseWindowName(sva.win)
			
			ChangeCurrentImage(dval, baseWindowName)
			
			break
	endswitch

	return 0
End

Function BTAnalyzePALMProc(ba) : ButtonControl
	STRUCT WMButtonAction &ba
	
	variable msTimer, sec, err

	switch( ba.eventCode )
		case 2: // mouse up
				// click code here
				Struct FitData fitParams
				string baseWindowName = GetBaseWindowName(ba.win)
				
				// get all the fit options and parameters the user defined in the GUI
				GetGUIFitSettingsAndOptions(fitParams, baseWindowName)
				
				msTimer = startMSTimer
				
				string posName
				err = DoPALMFitting(fitParams, baseWindowName, posName, 1)
				if (err != 0)	// an error occured
					Abort "An error occurred during the analysis"
				endif
				
				sec = stopMSTimer(msTimer)
				sec /= 1e6
				Printf "Elapsed time is %s\r", PrettyPrintTimeElapsed(sec)
				
				ControlUpdate /W=$baseWindowName#AnalysisControls PMSelectPositionsWave
				ControlUpdate /W=$baseWindowName#TrackingControls PMSelectPosWaveForTracking
				ControlUpdate /W=$baseWindowName#OtherControls PMDriftCorrSelectPositionsWave
				PopupMenu PMSelectPositionsWave win=$baseWindowName#AnalysisControls, mode=1, popMatch=posName
				break

	endswitch
	return 0
End

Function BTCopyPositionsProc(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	switch( ba.eventCode )
		case 2: // mouse up
			string baseWindowName = GetBaseWindowName(ba.win)
			DFREF windowDataFolder = GetWindowDataFolder(baseWIndowName)
			NewDataFolder /O root:'Localized Positions'
			DFREF outputFolder = root:'Localized Positions'
			string posName
			
			SVAR filePath = windowDataFolder:S_filePath
			
			ControlInfo /W=$baseWindowName#AnalysisControls PMSelectPositionsWave
			posName = S_Value
			
			if (StringMatch(posName, "* No Positions *") == 1)
				return 0
			endif
			
			wave /Z posWave = GetPositionsWaveReference(posName)
			
			if (WaveExists(posWave) == 0)
				SetDataFolder savDF
				Abort "The specified positions wave doesn't exist"
			endif
			
			// propose a name based on the filename
			string fileName = ParseFilePath(3, filePath, ":", 0, 1)
			string proposedName = CleanupName(fileName, 0)
			
			posName = GetNewPositionsWaveName(proposedName, outputFolder, "Enter the new name of the positions wave")
			if (strlen(posName) == 0)	// the user canceled the dialog
				return -1
			endif
			
			Duplicate /O posWave, outputFolder:$posName
			break
	endswitch

	return 0
End

Function BTRunParticleTrackingProc(ba) : ButtonControl
	STRUCT WMButtonAction &ba
	
	variable msTimer, sec, err

	Switch( ba.eventCode )
		case 2: // mouse up
				// click code here
			string baseWindowName = GetBaseWindowName(ba.win)
			DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
			DFREF trackDataFolder = root:'Particle Tracks'
			
			NVAR trackingMaxJump = windowDataFolder:V_trackingMaxJump
			NVAR trackingMaxBlinking = windowDataFolder:V_trackingMaxBlinking
			NVAR trackingMaxLength =  windowDataFolder:V_trackingMaxLength
			NVAR trackingMinLength =  windowDataFolder:V_trackingMinLength
			
			// get the name of the selected positions wave
			ControlInfo /W=$baseWindowName#TrackingControls PMSelectPosWaveForTracking
			string posName = S_Value
			
			if (StringMatch(posName, "* No Positions *") == 1)
				return 0
			endif
			
			wave /Z pos = GetPositionsWaveReference(posName)
			if (WaveExists(pos) != 1)
				Abort "The selected positions wave was not found"
			endif
			
			string outputWaveName = GetNewPositionsWaveName("", trackDataFolder, "")
			if (strlen(outputWaveName) == 0)
				return 0
			endif
			
			wave /WAVE W_GroupedEmitters = GroupEmitters(pos, CombinedPosEstimator_LastPos, trackingMaxJump, trackingMaxBlinking)
			
			variable nGroups = DimSize(W_GroupedEmitters, 0)
			variable nGroupsRemaining = nGroups
			variable i
			
			// eliminate all entries that do not fit between the requested lengths
			for (i = 0; i < nGroups; i+=1)
				wave currentEmitter = W_GroupedEmitters[i]
				if ((DimSize(currentEmitter, 0) < trackingMinLength) || (DimSize(currentEmitter, 0) > trackingMaxLength))
					nGroupsRemaining -= 1
				endif
			endfor
			
			Make /O/N=(nGroupsRemaining) /WAVE root:'Particle Tracks':$outputWaveName
			wave /WAVE outputWave = root:'Particle Tracks':$outputWaveName
			
			variable offset = 0
			for (i = 0; i < nGroups; i+=1)
				wave currentEmitter = W_GroupedEmitters[i]
				if ((DimSize(currentEmitter, 0) < trackingMinLength) || (DimSize(currentEmitter, 0) > trackingMaxLength))
					continue
				endif
				outputWave[offset] = currentEmitter
				offset += 1
			endfor
			
			// add a wavenote to the output wave
			string waveNote = Note(pos) + "TRACKING MAX SHIFT:" + num2str(trackingMaxJump) + ";" + "TRACKING MAX BLINKING:" + num2str(trackingMaxBlinking) + ";"
			waveNote += "TRACKING MIN LENGTH:" + num2str(trackingMinLength) + ";" + "TRACKING MAX LENGTH:" + num2str(trackingMaxLength) + ";"
			Note /K outputWave, waveNote
			
			// update the popup menu showing the tracks
			ControlUpdate /W=$baseWindowName#TrackingControls PMSelectTrackWave
			PopupMenu PMSelectTrackWave, win=$baseWindowName#TrackingControls, popMatch=outputWaveName
			break
	EndSwitch
End

Function BTDoSOFIProc(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	Switch( ba.eventCode )
		case 2: // mouse up
				// click code here
			string baseWindowName = GetBaseWindowName(ba.win)
			DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
			
			SVAR dataFilePath = windowDataFolder:S_filePath
			NVAR framesToSkip = windowDataFolder:V_SOFInFramesToSkip
			NVAR lastFrameInAnalysis = windowDataFolder:V_SOFILastFrameInAnalysis
			
			if (lastFrameInAnalysis <= framesToSkip)
				Abort "Invalid frame range"
			endif
			variable nImagesToInclude = lastFrameInAnalysis - framesToSkip
			
			variable sofiOrder
			ControlInfo /W=$baseWindowName#SOFIControls PMSOFIOrder
			sofiOrder = str2num(S_value)
			
			variable doCrossCorrelation
			ControlInfo /W=$baseWindowName#SOFIControls RBCrossSOFI
			doCrossCorrelation = V_value
			
			variable doAverage
			ControlInfo /W=$baseWindowName#SOFIControls CBSOFIDoAverageImage
			doAverage = V_value
			
			variable makeMovie
			ControlInfo /W=$baseWindowName#SOFIControls CBSOFIMakeMovie
			makeMovie = V_value
			
			variable nFramesToGroup
			if (makeMovie != 0)
				ControlInfo /W=$baseWindowName#SOFIControls SVSOFInFramesInMovie
				nFramesToGroup = V_value
			else
				nFramesToGroup = 0
			endif
			
			variable dropSaturation = 0
			ControlInfo /W=$baseWindowName#SOFIControls CBSOFIDropSaturatedFrames
			dropSaturation = V_value
			
			// add a structure with a reference to the progress reporting function
			STRUCT LocalizerProgStruct progressStruct
			progressStruct.version = kLocalizerProgStructVersion
			FUNCREF ProgressFunctionPrototype progressStruct.func = ProgressReporterFunc
			
			SOFIAnalysis /XC=(doCrossCorrelation) /ORDR=(sofiOrder) /GRP=(nFramesToGroup) /LAG=0 /SUB={framesToSkip, nImagesToInclude} /AVG=(doAverage) /NSAT=(dropSaturation) /PROG=progressStruct /DEST=windowDataFolder:M_SOFI dataFilePath
			wave M_SOFI = windowDataFolder:M_SOFI
			
			// adjust the wave scaling to match the pixels of the original data
			// if crosscorrelation is used
			if (doCrossCorrelation)
				SetScale /P x, 1, 1 / sofiOrder, M_SOFI
				SetScale /P y, 1, 1 / sofiOrder, M_SOFI
			endif
			
			// provide a wave note that encapsulates the relevant information
			string waveNote = "KIND:SOFI;ORDER:" + num2istr(sofiOrder) + ";CROSSCORRELATION:" + num2istr(doCrossCorrelation) + ";GROUPING:" + num2istr(nFramesToGroup) + ";SKIP:" + num2istr(framesToSkip) + ";DROPSATURATION:" + num2istr(dropSaturation) + ";ORIGINAL FILE PATH:" + dataFilePath + ";"
			Note /K M_SOFI, waveNote
			
			string SOFIWindowTitle = "SOFI image from " + ParseFilePath(3, dataFilePath, ":", 0, 1)
			string SOFIWindowName = ReplaceString("LocalizerViewer", baseWindowName, "SOFIViewer")
			
			DoWindow /F $SOFIWindowName
			if (V_flag == 0)
				Display /K=1 /N=$SOFIWindowName as SOFIWindowTitle
				AppendImage /G=1 /W=$SOFIWindowName M_SOFI
				AdjustGraphForImageDisplay(SOFIWindowName)
				
				ModifyGraph /W=$SOFIWindowName width={Aspect, DimSize(M_SOFI, 0) / DimSize(M_SOFI, 1)}
				
				// if this is a movie (more than one image)
				// then show a slider automatically
				if (DimSize(M_SOFI, 2) > 1)
					WMAppend3DImageSlider()
				else
					// this is really a 2D image, and some procedures (e.g. ImageSave)
					// choke on mxmx1 matrices. So turn it into one explictly
					Redimension /N=(-1, -1) M_SOFI
				endif
			endif
			
			if (doAverage)
				string averageWindowTitle = "Average/SOFI image from " + ParseFilePath(3, dataFilePath, ":", 0, 1)
				string averageWindowName = ReplaceString("LocalizerViewer", baseWindowName, "AverageSOFIViewer")
			
				DoWindow /F $averageWindowName
				if (V_flag == 0)
					wave M_SOFI_avg = windowDataFolder:M_SOFI_avg
					
					Display /K=1 /N=$averageWindowName as averageWindowTitle
					AppendImage /G=1 /W=$averageWindowName M_SOFI_avg
					AdjustGraphForImageDisplay(averageWindowName)
					
					ModifyGraph /W=$averageWindowName width={Aspect, DimSize(M_SOFI_avg, 0) / DimSize(M_SOFI_avg, 1)}
					
					// if this is a movie (more than one image)
					// then show a slider automatically
					if (DimSize(M_SOFI_avg, 2) > 1)
						WMAppend3DImageSlider()
					else
						// this is really a 2D image, and some procedures (e.g. ImageSave)
						// choke on mxmx1 matrices. So turn it into one explictly
						Redimension /N=(-1, -1) M_SOFI_avg
					endif
					
					AutoPositionWindow /M=0 /R=$SOFIWindowName $averageWindowName
				endif
			endif
			break
	EndSwitch
End

Function BTDoSOFIDeconvolutionProc(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	Switch( ba.eventCode )
		case 2:	// mouse up
			string baseWindowName = GetBaseWindowName(ba.win)
			DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
			
			NVAR gaussianWidth = windowDataFolder:V_gaussianWidth
			NVAR SOFInDeconvolutionIterations = windowDataFolder:V_SOFInDeconvolutionIterations
			
			SVAR dataFilePath = windowDataFolder:S_filePath
			
			wave /Z M_SOFI = windowDataFolder:M_SOFI
			if (WaveExists(M_SOFI) != 1)
				Abort "No SOFI image was found"
			endif
			
			// get the calculation settings from the wavenote
			string waveNote = Note(M_SOFI)
			variable order = NumberByKey("ORDER", waveNote)
			variable isCross = NumberByKey("CROSSCORRELATION", waveNote)
			
			variable effectivePSFWidth
			if (isCross != 0)
				effectivePSFWidth = gaussianWidth
			else
				// generate more pixels in XCSOFI, so the effective psfWidth is larger
				effectivePSFWidth = gaussianWidth * order
			endif
			
			// make an image of the PSF
			variable psfDimensions = ceil(2 * 4 * effectivePSFWidth)
			if (mod(psfDimensions, 2) != 1)
				psfDimensions += 1
			endif
			
			Make /D/FREE/N=(psfDimensions, psfDimensions) M_PSF
			SetScale /P x, - (psfDimensions - 1) / 2, 1, M_PSF
			SetScale /P y, - (psfDimensions - 1) / 2, 1, M_PSF
			
			M_PSF = StatsNormalPDF(sqrt(x^2 + y^2), 0, effectivePSFWidth)^order
			
			// if the SOFI image really is a movie then multiple frames need to be processed
			variable nFrames = DimSize(M_SOFI, 2)
			if (nFrames == 0)
				nFrames = 1
			endif
			
			Make /O/N=(DimSize(M_SOFI, 0), DimSize(M_SOFI, 1), nFrames) /D windowDataFolder:M_DeconvolvedSOFI
			wave M_DeconvolvedSOFI = windowDataFolder:M_DeconvolvedSOFI
			
			variable i
			for (i = 0; i < nFrames; i+=1)
				MatrixOP /O /FREE M_ExtractedPlane = M_SOFI[][][i]
				WaveStats /Q/M=1 M_ExtractedPlane
				if (V_min < 0)	// contains negative values, these appear to produce noise in the deconvolution
					MatrixOP /FREE /O M_ExtractedPlane = M_ExtractedPlane - V_min
				endif
				ImageRestore /ITER=(SOFInDeconvolutionIterations) /DEST=windowDataFolder:M_DeconvolvedSOFI_Temp srcWave=M_ExtractedPlane, psfWave=M_PSF
				ImageTransform /D=windowDataFolder:M_DeconvolvedSOFI_Temp /P=(i) setPlane, M_DeconvolvedSOFI
			endfor
			
			KillWaves /Z windowDataFolder:M_DeconvolvedSOFI_Temp
			
			// copy the original wave scaling as well
			CopyScales M_SOFI, M_DeconvolvedSOFI
			
			string windowTitle = "Deconvolved SOFI image from " + ParseFilePath(3, dataFilePath, ":", 0, 1)
			string windowName = ReplaceString("LocalizerViewer", baseWindowName, "DeconvolvedSOFIViewer")
			
			DoWindow /F $windowName
			if (V_flag == 0)
				Display /K=1 /N=$windowName as windowTitle
				AppendImage /G=1 /W=$windowName M_DeconvolvedSOFI
				AdjustGraphForImageDisplay(windowName)
				ModifyGraph /W=$windowName width={Aspect, DimSize(M_DeconvolvedSOFI, 0) / DimSize(M_DeconvolvedSOFI, 1)}
				
				// if this is a movie (more than one image)
				// then show a slider automatically
				if (DimSize(M_DeconvolvedSOFI, 2) > 1)
					WMAppend3DImageSlider()
				else
					// this is really a 2D image, and some procedures (e.g. ImageSave)
					// choke on mxmx1 matrices. So turn it into one explictly
					Redimension /N=(-1, -1) M_DeconvolvedSOFI
				endif
			endif
			break
	EndSwitch
End

Function BTBasicAnalysisProc(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	switch( ba.eventCode )
		case 2: // mouse up
			// click code here
			string baseWindowName = GetBaseWindowName(ba.win)
			DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
			
			variable analysisMethod
			SVAR CCDFilePath = windowDataFolder:S_filePath
			
			string fileName = ParseFilePath(5, CCDFilePath, ":", 0, 0)
			fileName = ParseFilePath(3, fileName, ":", 0, 0)
			
			ControlInfo /W=$baseWindowName#OtherControls PMChooseCCDAnalysisMethod
			StrSwitch (S_Value)
				case "Average Intensity Trace":
					analysisMethod = ANALYZECCD_AVERAGETRACE
					break
				case "Summed Intensity Trace":
					analysisMethod = ANALYZECCD_SUMMEDINTENSITYTRACE
					break
				case "Average Image":
					analysisMethod = ANALYZECCD_AVERAGE_IMAGE
					break
				case "Variance Image":
					analysisMethod = ANALYZECCD_VARIANCE_IMAGE
					break
				default:
					Abort "Unknown CCD analysis method"
			EndSwitch
			
			// add a structure with a reference to the progress reporting function
			STRUCT LocalizerProgStruct progressStruct
			progressStruct.version = kLocalizerProgStructVersion
			FUNCREF ProgressFunctionPrototype progressStruct.func = ProgressReporterFunc
			
			string graphWindowName
			string graphTitle
			GetWindow $baseWindowName, title
			string baseWindowTitle = S_Value
			
			switch (analysisMethod)
				case ANALYZECCD_SUMMEDINTENSITYTRACE:
					AnalyzeCCDImages /M=(analysisMethod) /PROG=progressStruct /DEST=windowDataFolder:W_SummedIntensityTrace CCDFilePath
					wave W_SummedIntensityTrace = windowDataFolder:W_SummedIntensityTrace
					graphWindowName = ReplaceString("LocalizerViewer", baseWindowname, "SummedIntensityTrace")
					graphTitle = "Summed intensity trace from " + baseWindowTitle
					DoWindow /F $graphWindowName
					if (V_flag != 1)
						Display /K=1 /N=$graphWindowName
						AppendToGraph /W=$S_name W_SummedIntensityTrace
						Label /W=$S_name Left, "Summed Intensity"
						Label /W=$S_name Bottom, "Frame Number"
					endif
					DoWindow /T $graphWindowName, graphTitle
					break
				case ANALYZECCD_AVERAGETRACE:
					AnalyzeCCDImages /M=(analysisMethod) /PROG=progressStruct /DEST=windowDataFolder:W_AverageIntensityTrace CCDFilePath
					wave W_AverageIntensityTrace = windowDataFolder:W_AverageIntensityTrace
					graphWindowName = ReplaceString("LocalizerViewer", baseWindowname, "AverageIntensityTrace")
					graphTitle = "Average intensity trace from " + baseWindowTitle
					DoWindow /F $graphWindowName
					if (V_flag != 1)
						Display /K=1 /N=$graphWindowName
						AppendToGraph /W=$S_name W_AverageIntensityTrace
						Label /W=$S_name Left, "Average Intensity"
						Label /W=$S_name Bottom, "Frame Number"
					endif
					DoWindow /T $graphWindowName, graphTitle
					break
				case ANALYZECCD_AVERAGE_IMAGE:	// calculate a image containing an average of the entire trace
					AnalyzeCCDImages /M=(analysisMethod) /PROG=progressStruct /DEST=windowDataFolder:M_AverageImage CCDFilePath
					wave M_AverageImage = windowDataFolder:M_AverageImage
					graphWindowName = ReplaceString("LocalizerViewer", baseWindowname, "AverageImage")
					graphTitle = "Average image from " + baseWindowTitle
					DoWindow /F $graphWindowName
					if (V_flag != 1)
						Display /K=1 /N=$graphWindowName
						AppendImage /W=$S_name M_AverageImage
						AdjustGraphForImageDisplay(S_name)
						ModifyGraph /W=$S_name width={Aspect, DimSize(M_AverageImage, 0) / DimSize(M_AverageImage, 1)}, height=0
					endif
					DoWindow /T $graphWindowName, graphTitle
					break
				case ANALYZECCD_VARIANCE_IMAGE:	// calculate an image with the variance of each pixel
					AnalyzeCCDImages /M=(analysisMethod) /PROG=progressStruct /DEST=windowDataFolder:M_Variance CCDFilePath
					wave M_Variance = windowDataFolder:M_Variance
					graphWindowName = ReplaceString("LocalizerViewer", baseWindowname, "VarianceImage")
					graphTitle = "Variance image from " + baseWindowTitle
					DoWindow /F $graphWindowName
					if (V_flag != 1)
						Display /K=1 /N=$graphWindowName
						AppendImage /W=$S_name M_Variance
						AdjustGraphForImageDisplay(S_name)
						ModifyGraph /W=$S_name width={Aspect, DimSize(M_Variance, 0) / DimSize(M_Variance, 1)}, height=0
					endif
					break
				default:
					Abort "Unknown CCD analysis method"
					break
			endswitch
			
			break
	endswitch
	
	return 0
End

Function BTBleachingAnalysisProc(ba) : ButtonControl
	STRUCT WMButtonAction &ba
	
	switch( ba.eventCode )
		case 2: // mouse up
			// click code here
			string baseWindowName = GetBaseWindowName(ba.win)
			NVAR bleachingMaxJump = windowDataFolder:V_bleachingMaxJump
			
			RecoverEmittersByBleachingSteps(bleachingMaxJump, baseWindowName)
			break
	EndSwitch
End

Function PMChooseMovieProcessingProc(pa) : PopupMenuControl
	STRUCT WMPopupAction &pa

	switch( pa.eventCode )
		case 2: // mouse up
			string baseWindowName = GetBaseWindowName(pa.win)
			string selection = pa.popStr
			
			StrSwitch(selection)
				case "Subtract Pixelwise Average":
					SetVariable SVFramesAverageSubtraction,win=$baseWindowName#OtherControls,disable=0
					SetVariable SVCameraMultiplication,win=$baseWindowName#OtherControls,disable=1
					SetVariable SVCameraOffset,win=$baseWindowName#OtherControls,disable=1
					break
				case "Convert to Photons":
					SetVariable SVFramesAverageSubtraction,win=$baseWindowName#OtherControls,disable=1
					SetVariable SVCameraMultiplication,win=$baseWindowName#OtherControls,disable=0
					SetVariable SVCameraOffset,win=$baseWindowName#OtherControls,disable=0
					break
				default:
					SetVariable SVFramesAverageSubtraction,win=$baseWindowName#OtherControls,disable=1
					SetVariable SVCameraMultiplication,win=$baseWindowName#OtherControls,disable=1
					SetVariable SVCameraOffset,win=$baseWindowName#OtherControls,disable=1
					break
			EndSwitch
			break
	endswitch
	return 0
End

Function BTProcessMovieProc(ba) : ButtonControl
	STRUCT WMButtonAction &ba
	
	switch( ba.eventCode )
		case 2: // mouse up
			// click code here
			string baseWindowName = GetBaseWindowName(ba.win)
			DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
			NewDataFolder /O root:'Localizer Movies'
			DFREF movieFolder = root:'Localizer Movies'
			
			NVAR nFramesAverageSubtraction = windowDataFolder:V_nFramesAverageSubtraction
			NVAR cameraMultiplication = windowDataFolder:V_cameraMultiplication
			NVAR cameraOffset = windowDataFolder:V_cameraOffset
			
			SVAR filePath = windowDataFolder:S_filePath
			
			// ask how the user wants to save the data  and get the output path
			variable outputFormat
			string outputPath
			GetOutputStorageTypeAndPath(filePath, outputFormat, outputPath)
			if (strlen(outputPath) == 0)
				return 0
			endif
			
			// determine the processing to be done
			variable processMethod
			ControlInfo /W=$baseWindowName#OtherControls PMChooseMovieProcessingMethod
			StrSwitch (S_Value)
				case "Subtract Pixelwise Average":
					processMethod = PROCESSCCDIMAGE_SUBAVG
					break
				case "Difference Image":
					processMethod = PROCESSCCDIMAGE_DIFFIMAGE
					break
				case "Convert File Format":
					processMethod = PROCESSCCDIMAGE_CONVERTFORMAT
					break
				case "Convert to photons":
					processMethod = PROCESSCCDIMAGE_CONVERTPHOTONS
					break
				default:
					Abort "Unknown CCD processing method in StrSwitch in BTProcessMovieProc"
			EndSwitch
			
			// add a structure with a reference to the progress reporting function
			STRUCT LocalizerProgStruct progressStruct
			progressStruct.version = kLocalizerProgStructVersion
			FUNCREF ProgressFunctionPrototype progressStruct.func = ProgressReporterFunc
			
			// do the actual work
			ProcessCCDImages /M=(processMethod) /CAL={cameraOffset, cameraMultiplication} /AVG=(nFramesAverageSubtraction) /OUT=(outputFormat) /PROG=progressStruct /O filePath, outputPath
			
			// open the newly created movie
			NewInterfacePanel(outputPath)
			break
	EndSwitch
End

Function MakeAverageTraceFromMarquee()
	
	GetMarquee /Z left, bottom
	if (V_flag == 0)
		return 0
	endif
	
	string baseWindowName = GetBaseWindowName(S_MarqueeWin)
	DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
	
	if (GrepString(baseWindowName, "LocalizerViewer[0-9]*") != 1)
		Abort "This procedure can only be called on viewer windows created with the Localizer package"
	endif
	
	wave image = windowDataFolder:M_CCDImage
	SVAR filePath = windowDataFolder:S_FilePath
	
	variable xSize = DimSize(image, 0)
	variable ySize = DimSize(image, 1)
	Variable bottom, left, right, top
	
	bottom = round(min(V_bottom, V_top))
	top = round(max(V_bottom, V_top))
	left = round(min(V_left, V_right))
	right = round(max(V_left, V_right))
	
	// crop the marquee to the dimensions of the image
	bottom = (bottom < 0) ? 0 : bottom
	left = (left < 0) ? 0 : left
	top = (top >= ySize) ? ySize - 1 : top
	right = (right >= xSize) ? xSize - 1 : right
	
	if ((top == bottom) || (left == right))	// it's not a real box
		return -1
	endif
	
	// add a structure with a reference to the progress reporting function
	STRUCT LocalizerProgStruct progressStruct
	progressStruct.version = kLocalizerProgStructVersion
	FUNCREF ProgressFunctionPrototype progressStruct.func = ProgressReporterFunc
	
	// run the operation that will generate the trace
	AnalyzeCCDImages /M=(ANALYZECCD_AVERAGETRACE) /ROI={left, right, bottom, top} /PROG=progressStruct /DEST=windowDataFolder:W_AverageIntensityTrace filePath
	
	// now display the resulting image
	string graphWindowName = ReplaceString("LocalizerViewer", baseWindowName, "AverageIntensityTrace")
	
	GetWindow $baseWindowName, title
	string graphWindowTitle = "Average trace from " + S_value
	
	DoWindow /F $graphWindowName
	if (V_flag == 0)
		Display /K=1 /N=$graphWindowName windowDataFolder:W_AverageIntensityTrace as "Average intensity trace"
		Label /W=$S_name left, "Average intensity"
		Label /W=$S_name bottom, "Frame number"
	endif
	
	DoWindow /T $graphWindowName, graphWindowTitle
End

Function MakeCroppedStackFromMarquee()
	GetMarquee /Z left, bottom
	if (V_flag == 0)
		return 0
	endif
	
	string baseWindowName = GetBaseWindowName(S_MarqueeWin)
	DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
	
	if (GrepString(baseWindowName, "LocalizerViewer[0-9]*") != 1)
		Abort "This procedure can only be called on viewer windows created with the Localizer package"
	endif
	
	wave image = windowDataFolder:M_CCDImage
	SVAR filePath = windowDataFolder:S_FilePath
	
	variable xSize = DimSize(image, 0)
	variable ySize = DimSize(image, 1)
	Variable bottom, left, right, top
	
	bottom = round(min(V_bottom, V_top))
	top = round(max(V_bottom, V_top))
	left = round(min(V_left, V_right))
	right = round(max(V_left, V_right))
	
	// crop the marquee to the dimensions of the image
	bottom = (bottom < 0) ? 0 : bottom
	left = (left < 0) ? 0 : left
	top = (top >= ySize) ? ySize - 1 : top
	right = (right >= xSize) ? xSize - 1 : right
	
	if ((top == bottom) || (left == right))	// it's not a real box
		return -1
	endif
	
	variable outputFormat
	string outputPath
	GetOutputStorageTypeAndPath(filePath, outputFormat, outputPath)
	if (strlen(outputPath) == 0)
		return 0
	endif
	
	// add a structure with a reference to the progress reporting function
	STRUCT LocalizerProgStruct progressStruct
	progressStruct.version = kLocalizerProgStructVersion
	FUNCREF ProgressFunctionPrototype progressStruct.func = ProgressReporterFunc
	
	// do the actual processing
	ProcessCCDImages /O /ROI={left, right, bottom, top} /M=(PROCESSCCDIMAGE_CROP) /OUT=(outputFormat) /PROG=progressStruct filePath, outputPath
	
	// open the newly created movie
	NewInterfacePanel(outputPath)
End

Function Clustering_Marquee(procedureName)
	string procedureName	// either "LClustering" or "Direct"
	
	variable xPixelSize, yPixelSize, xStart, yStart, xEnd, yEnd
	string marqueeWin
	
	wave /Z positions = GetCoordinatesFromMarquee(marqueeWin, xPixelSize, yPixelSize, xStart, xEnd, yStart, yEnd)
	if (WaveExists(positions) == 0)
		// signals a user cancel
		return 0
	endif
	
	variable nPos = DimSize(positions, 0), xCol, yCol, zCol
	variable nPositionsWithinLimits = 0, i, offset
	
	// get the indices of the columns containing the x and y coordinates
	getColumnsForEmitterPositions(positions, xCol, yCol, zCol)
	if ((xCol == -1) || (yCol == -1))
		Abort "The positions passed to ExtractSubsetOfPositions_menu() do not appear to contain any (x,y) information"
	endif
	
	
	wave M_SelectedPositions = GetPositionsWithinLimits(positions, xStart, xEnd, yStart, yEnd)
	nPositionsWithinLimits = DimSize(M_SelectedPositions, 0)
	
	// if there are no positions within the box then don't do anything
	if (nPositionsWithinLimits == 0)
		return 0
	endif
	
	StrSwitch (procedureName)
		case "LClustering":
			DFREF savDF = GetDataFolderDFR()
			SetDataFolder root:Packages:Localizer
			variable calculationRange = 0.45 * sqrt((xEnd - xStart)^2 + (yEnd - yStart)^2)
			RipleyLFunctionClustering /RNGE={calculationRange, round(xEnd - xStart) / 0.01} /REGN={xStart, xEnd, yStart, yEnd} M_SelectedPositions
			SetDataFolder savDF
			
			wave W_lFunction = root:Packages:Localizer:W_lFunction
			Duplicate /O W_lFunction, root:Packages:Localizer:W_LFunctionSubX
			wave W_LFunctionSubX = root:Packages:Localizer:W_LFunctionSubX
			
			W_LFunctionSubX = W_LFunction - x
			
			if (xPixelSize != 0)
				SetScale /I x, 0, calculationRange * xPixelSize, W_LFunctionSubX
			endif
			
			DoWindow /F LFunction
			if (V_flag != 1)
				Display /K=1/N=LFunction W_LFunctionSubX
				Label /W=LFunction Left, "L Function - x"
			endif
			if (xPixelSize != 0)
				Label /W=LFunction Bottom, "Distance (nm)"
			else
				Label /W=LFunction Bottom, "Distance (pixel)"
			endif
			break
		case "Direct":
			// get the center of the cluster
			variable clusterCenterX = 0, clusterCenterY = 0
			for (i = 0; i < nPositionsWithinLimits; i+=1)
				clusterCenterX += M_SelectedPositions[i][xCol]
				clusterCenterY += M_SelectedPositions[i][yCol]
			endfor
			clusterCenterX /= nPositionsWithinLimits
			clusterCenterY /= nPositionsWithinLimits
			
			// calculate the distances of all the positions from the center
			Make /D/O/N=(nPositionsWithinLimits) /FREE W_CenterDistances
			for (i = 0; i < nPositionsWithinLimits; i+=1)
				W_CenterDistances[i] = sqrt((M_SelectedPositions[i][xCol] - clusterCenterX)^2 + (M_SelectedPositions[i][yCol] - clusterCenterY)^2)
			endfor
			
			WaveStats /Q W_CenterDistances
			variable stdDev = V_sdev
			variable maxDistance = V_max
			variable avgDistance = V_avg
			Printf "In pixels: %u points, center is at (%g, %g), average distance from the center is %g, standard deviation is %g, radius of enclosing circle is %g\r%u,%g,%g,%g,%g,%g\r", nPositionsWithinLimits, clusterCenterX, clusterCenterY, avgDistance, stdDev, maxDistance, nPositionsWithinLimits, clusterCenterX, clusterCenterY, avgDistance, stdDev, maxDistance
			
			// also print the information in pixels if available
			if (xPixelSize != 0)
				string formatString = "In nanometer: %u points, center is at (%g nm, %g nm), average distance from the center is %g nm, standard deviation is %g nm, radius of enclosing circle is %g nm\r%u,%g,%g,%g,%g,%g\r" // work around the stupid 400-character limitation in Igor
				Printf formatString, nPositionsWithinLimits, clusterCenterX * xPixelSize, clusterCenterY * xPixelSize, avgDistance * xPixelSize, stdDev * xPixelSize, maxDistance * xPixelSize, nPositionsWithinLimits, clusterCenterX * xPixelSize, clusterCenterY * xPixelSize, avgDistance * xPixelSize, stdDev * xPixelSize, maxDistance * xPixelSize
			endif
			break
		default:
			Abort "Unknown analysis method in Clustering_Marquee"
			break
	EndSwitch
End

Function /WAVE GetCoordinatesFromMarquee(marqueeWin, xPixelSize, yPixelSize, xStart, xEnd, yStart, yEnd)
	string &marqueeWin
	variable &xPixelSize, &yPixelSize, &xStart, &xEnd, &yStart, &yEnd
	
	// returns the coordinates of the marquee in units of nm (if the pixelsizes != 1) or in pixels otherwise
	
	GetMarquee /Z left, bottom
	if (V_flag == 0)
		return $""
	endif
	
	marqueeWin = S_MarqueeWin
	variable bottom, top, left, right
	bottom = min(V_bottom, V_top)
	top = max(V_bottom, V_top)
	left = min(V_left, V_right)
	right = max(V_left, V_right)
	
	if ((top == bottom) || (left == right))	// it's not a real box
		return $""
	endif
	
	// try to get the positions wave from the graph automatically
	string posName
	wave /Z pos = GetPositionsWaveFromGraph(marqueeWin)
	// if this failed then prompt the user
	if (WaveExists(pos) != 1)
		Prompt posName, "Positions wave:", popup, GetPossiblePositionsWaves()
		DoPrompt "Choose a positions wave", posName
		if (V_flag == 1)
			return $""
		endif
	
		wave /Z pos = GetPositionsWaveReference(posName)	
		if (WaveExists(pos) == 0)
			Abort "The selected positions wave doesn't seem to exist!"
		endif
	endif
	
	variable err
	string units
	
	// the main localizerviewer panel is an exception, since its coordinates
	// are always in units of pixels, so it doesn't bother with unit strings
	// and so forth. Handle this case separately
	if (GrepString(marqueeWin, "^LocalizerViewer[0-9]*#CCDViewer$"))
		xPixelSize = NumberByKey("X PIXEL SIZE", note(pos))
		yPixelSize = NumberByKey("Y PIXEL SIZE", note(pos))
		if ((NumType(xPixelSize) == 2) || (NumType(yPixelSize) == 2))
			xPixelSize = 0
			yPixelSize = 0
		endif
		xStart = left
		yStart = bottom
		xEnd = right
		yEnd = top
		return pos
	endif
	
	// determine if the plot is in units of pixels or um
	string traceNames, imageNames, unitsString
	traceNames = TraceNameList("", ";", 1)
	imageNames = ImageNameList("", ";")
	if (ItemsInList(traceNames) > 0)	// the plot contains some traces, use these to get the scale
		wave traceWave = TraceNameToWaveRef("", StringFromList(0, traceNames))
		unitsString = WaveUnits(traceWave, 0)
	elseif (ItemsInList(imageNames) > 0) // the plot does not contain any traces, try using an image instead
		wave imageWave = ImageNameToWaveRef("", StringFromList(0, imageNames))
		unitsString = WaveUnits(imageWave, 0)
	else	// no traces or images on the plot
		Abort "The graph does not seem to contain any traces or images"
	endif
	
	// adjust the marquee coordinates depending on whether the graph
	// is in units of pixels or um
	strswitch (unitsString)
		case "pixel":
		case "":	// treat no units as pixels
			xPixelSize = 0
			yPixelSize = 0
			break
		case "um":
			// get the wave units
			xPixelSize = NumberByKey("X PIXEL SIZE", note(pos))
			yPixelSize = NumberByKey("Y PIXEL SIZE", note(pos))
			if ((NumType(xPixelSize) == 2) || (NumType(yPixelSize) == 2))	// NaN
				Abort "Unable to recognize the units of the plot."
			endif
			break
		default:
			Abort "Unable to recognize the units of the plot."
	endswitch
	
	if (xPixelSize != 0)
		// convert the box to pixels
		xStart = left / (xPixelSize / 1000)
		yStart = bottom / (yPixelSize / 1000)
		XEnd = right / (xPixelSize / 1000)
		yEnd = top / (yPixelSize / 1000)
	else
		// the box already is in units of pixels
		xStart = left
		yStart = bottom
		XEnd = right
		yEnd = top
	endif
	
	return pos
End

Function /S GetNewPositionsWaveName(suggestedName, outputDataFolder, promptString)
	string suggestedName
	DFREF outputDataFolder
	string promptString
	// prompt the user for a wave name in dataFolder
	// check that the name is a valid wave name and return it
	// if the user canceled then return an empty string
	// the promptstring is used in the dialogs to let the user know what happens
	
	string outputWaveName = suggestedName
	variable outputWaveFound = 0
	
	do
		Prompt outputWaveName, "Wave name:"
		DoPrompt promptString, outputWaveName
		if (V_flag == 1)	// the user canceled
			return ""
		endif
		
		// check if the name is not a blank string
		if (strlen(outputWaveName) == 0)
			continue
		endif
		
		// check if the name is not reserved
		if ((StringMatch(outputWaveName, "* No Positions *") == 1) || (StringMatch(outputWaveName, "* No Tracks *") == 1))
			continue
		endif
		
		// check if the name is acceptable to Igor
		// do not allow liberal names
		if (stringmatch(outputWaveName, CleanupName(outputWaveName, 0)) != 1)
			outputWaveName = CleanupName(outputWaveName, 0)
			continue
		endif
		
		// check if the wave already exists
		wave /Z outputWave = outputDataFolder:$outputWaveName
		if (WaveExists(outputWave) == 1)
			// the requested wave already exists
			// if this is the pre-existing wave is a wave create by this package, then offer the user to overwrite it
			// otherwise require that the user specifies a new name
			if (IsPositionsWave(outputWave) || IsTrackingWave(outputWave) || IsRegistrationMap(outputWave))
				DoAlert 2, "A set of data called \"" + outputWaveName + "\" already exists! Do you want to replace it with the new wave?"
				switch (V_flag)
					case 1:	// OK to replace
							// we're all set
						outputWaveFound = 1
						break
					case 2:	// Don't replace, use a different name instead
						outputWaveName = UniqueName(outputWaveName, 1, 0)
						continue
						break
					case 3:	// user cancelled, escape
						return ""
						break
					default:	// shouldn't happen
						Abort "Unknown value for V_flag in DoAlert"
						break
				endswitch
			else
				// don't overwrite a wave that is not a positions wave
				outputWaveName = UniqueName(outputWaveName, 1, 0)
				continue
			endif
		else
			// the output wave does not exist within the requested data folder yet
			// however, it could still be name of a positions wave stored somewhere else
			if (IsUniquePositionsWaveName(outputWaveName) == 0)
				outputWaveName = UniqueName(outputWaveName, 1, 0)
				continue
			endif
		endif
		
		// if we arrive here then the waveName looks okay
		outputWaveFound = 1
	while (outputWaveFound == 0)
	
	return outputWaveName
End

Function IsUniquePositionsWaveName(suggestedName)
	string suggestedName
	
	string listOfPositionsWaves = GetPossiblePositionsWaves()
	variable nWaves = ItemsInList(listOfPositionsWaves)
	
	string thisWaveName
	variable i
	for (i = 0; i < nWaves; i+=1)
		thisWaveName = StringFromList(i, listOfPositionsWaves)
		if (StringMatch(thisWaveName, suggestedName) == 1)
			return 0
		endif
	endfor
	
	return 1
End

Function /WAVE MakeImageHistogram(image, windowName)
	wave image
	string windowName
	
	DFREF savDF = GetDataFolderDFR()
	DFREF windowDataFolder = GetWindowDataFolder(windowName)
	SetDataFolder windowDataFolder
	ImageHistogram image
	SetDataFolder savDF
	
	wave hist = windowDataFolder:W_ImageHist
	
	return hist
End

Function GetGUIFitSettingsAndOptions(fitParams, windowName)
	STRUCT FitData &fitParams
	string windowName
	
	DFREF windowDataFolder = GetWindowDataFolder(windowName)
	string baseWindowName = GetBaseWindowName(windowName)
	
	NVAR currentThreshold = windowDataFolder:V_currentThreshold
	NVAR currentImage = windowDataFolder:V_currentImage
	NVAR PFA = windowDataFolder:V_PFA
	NVAR smoothSigmaFactor = windowDataFolder:V_smoothSigmaFactor
	NVAR gaussianWidth = windowDataFolder:V_gaussianWidth
	NVAR nImages = windowDataFolder:V_nImages
	NVAR firstFrameInLocalization = windowDataFolder:V_firstFrameInLocalization
	NVAR lastFrameInLocalization = windowDataFolder:V_lastFrameInLocalization
	SVAR CCDFilePath = windowDataFolder:S_filePath
	
	NVAR particleVerifierOverlap = windowDataFolder:V_particleVerifierOverlap
	NVAR particleVerifierSymm = windowDataFolder:V_particleVerifierSymm
	NVAR particleVerifierEllipse = windowDataFolder:V_particleVerifierEllipse
	
	variable thresholdMethod, preprocessing, postprocessing, particlefinder, localizationMethod
	
	ControlInfo /W=$baseWindowName#AnalysisControls PMThresholdMethod
	StrSwitch (S_Value)
		case "GLRT":
			thresholdMethod = THRESHOLD_METHOD_GLRT
			break
		case "Isodata":
			thresholdMethod = THRESHOLD_METHOD_ISODATA
			break
		case "Triangle":
			thresholdMethod = THRESHOLD_METHOD_TRIANGLE
			break
		case "Absolute":
			thresholdMethod = THRESHOLD_METHOD_DIRECT
			break
		case "SmoothSigma":
			thresholdMethod = THRESHOLD_METHOD_SMOOTHSIGMA
			break
		default:
			Abort "Unknown segmentation method"
			break
	EndSwitch
	
	ControlInfo /W=$baseWindowName#AnalysisControls PMThresholdPreprocessing
	StrSwitch (S_Value)
		case "None":
			preprocessing = PREPROCESSOR_NONE
			break
		case "3x3 Median Filter":
			preprocessing = PREPROCESSOR_3X3MEDIAN
			break
		case "5x5 Median Filter":
			preprocessing = PREPROCESSOR_5X5MEDIAN
			break
		case "1x1 Gaussian Smoothing":
			preprocessing = PREPROCESSOR_1X1GAUSSIAN
			break
		case "2x2 Gaussian Smoothing":
			preprocessing = PREPROCESSOR_2X2GAUSSIAN
			break
		case "3x3 Mean Filter":
			preprocessing = PREPROCESSOR_3X3MEAN
			break
		case "5x5 Mean Filter":
			preprocessing = PREPROCESSOR_5X5MEAN
			break
		default:
			Abort "Unknown preprocessing method"
			break
	EndSwitch
	
	ControlInfo  /W=$baseWindowName#AnalysisControls PMThresholdPostProcessing
	StrSwitch (S_Value)
		case "None":
			postprocessing =  POSTPROCESSOR_NONE
			break
		case "Remove Isolated Pixels":
			postprocessing =  POSTPROCESSOR_REMOVE_ISOLATED
			break
		default:
			Abort "Unknown postprocessing method"
			break
	EndSwitch
	
	ControlInfo /W=$baseWindowName#AnalysisControls PMParticleFinder
	StrSwitch (S_Value)
		case "4-way adjacency":
			particlefinder = PARTICLEFINDER_ADJACENT4
			break
		case "8-way adjacency":
			particlefinder = PARTICLEFINDER_ADJACENT8
			break
		default:
			Abort "Unknown particle verifier method"
			break
	EndSwitch
	
	ControlInfo /W=$baseWindowName#AnalysisControls PMLocalizationMethod
	StrSwitch (S_Value)
		case "Gaussian Fitting":
			localizationMethod = LOCALIZATION_GAUSS_FITTING
			break
		case "Gaussian Fitting (Fixed Width)":
			localizationMethod = LOCALIZATION_GAUSS_FITTING_FIX
			break
		case "Iterative Multiplication":
			localizationMethod = LOCALIZATION_MULTIPLICATION
			break
		case "Center-of-Mass":
			localizationMethod = LOCALIZATION_CENTROID
			break
		case "Ellipsoidal Gaussian Fitting":
			localizationMethod = LOCALIZATION_ELLIPSOIDAL2DGAUSS
			break
		case "MLEwG":
			localizationMethod = LOCALIZATION_MLEWG
			break
		default:
			Abort "Unknown localization method in GetGUIFitSettingsAndOptions()"
	EndSwitch
	
	// do a bunch of ugly crap with the particle verification since it's pretty hard to procedurally generate a multi-keyword flag
	if (particleVerifierSymm != 0)
		fitParams.PVerSymm = PARTICLEVERIFIER_2DGAUSS
	else
		fitParams.PVerSymm = PARTICLEVERIFIER_NONE
	endif
	
	if (particleVerifierEllipse != 0)
		fitParams.PVerEllipse = PARTICLEVERIFIER_ELLIPSEGAUSS
	else
		fitParams.PVerEllipse = PARTICLEVERIFIER_NONE
	endif
	
	if (particleVerifierOverlap != 0)
		fitParams.PVerOverlap = PARTICLEVERIFIER_OVERLAP
	else
		fitParams.PVerOverlap = PARTICLEVERIFIER_NONE
	endif
	
	fitParams.directThresholdLevel = currentThreshold
	fitParams.currentImage = currentImage
	fitParams.PFA = PFA
	fitParams.smoothSigmaFactor = smoothSigmaFactor
	fitParams.PSFWidth = gaussianWidth
	fitParams.firstFrameToAnalyze = firstFrameInLocalization
	fitParams.lastFrameToAnalyze = lastFrameInLocalization
	fitParams.numberOfImages = nImages
	fitParams.thresholdMethod = thresholdMethod
	fitParams.preprocessing = preprocessing
	fitParams.postprocessing = postprocessing
	fitParams.particlefinder = particlefinder
	fitParams.localizationMethod = localizationMethod
	fitParams.CCDFilePath = CCDFilePath
End

Function GenerateGUIThreshold(windowName)
	string windowName
	
	DFREF windowDataFolder = GetWindowDataFolder(windowName)
	DFREF savDF = GetDataFolderDFR()
	string baseWindowName = GetBaseWindowName(windowName)
	
	NVAR currentImage = windowDataFolder:V_currentImage
	variable showThreshold, showPositions, showFittedPositions, showParticleTracks, showTrackIdentifiers
	wave image = windowDataFolder:M_CCDImage
	string tracesInGraph, imagesInGraph
	
	// check if we need to display the segmentation
	ControlInfo /W=$baseWindowName#AnalysisControls CBThresholdCheck
	showThreshold = V_value
	
	// check if we need to display the particles
	ControlInfo /W=$baseWindowName#AnalysisControls CBShowParticles
	showPositions = V_value
	
	// check if we need to display fitted positions
	ControlInfo /W=$baseWindowName#AnalysisControls CBDisplayFittedPositions
	if (V_value == 1)
		// does the wave with the fitted positions exist?
		// we can only show the positions if they exist
		ControlInfo /W=$baseWindowName#AnalysisControls PMSelectPositionsWave
		
		wave /Z fittedPositions = GetPositionsWaveReference(S_Value)
		
		if (WaveExists(fittedPositions) == 0)
			showFittedPositions = 0
		else
			showFittedPositions = 1
		endif
	else
		showFittedPositions = 0
	endif
	
	// check if we need to display particle tracks
	ControlInfo /W=$baseWindowName#TrackingControls CBDisplayTracks
	showParticleTracks = V_Value
	variable showFullTrack
	if (showParticleTracks == 1)
		// does the wave with the particles exists?
		ControlInfo /W=$baseWindowName#TrackingControls PMSelectTrackWave
		wave /Z trackingWave = GetTrackingWaveReference(S_Value)
		if (WaveExists(trackingWave) == 0)
			showParticleTracks = 0
		endif
		
		// do we need the full track or not?
		ControlInfo /W=$baseWindowName#TrackingControls RBDisplayFullTracks
		showFullTrack = V_Value
		
		// and the track identifiers?
		ControlInfo /W=$baseWindowName#TrackingControls CBDisplayTrackIdentifier
		showTrackIdentifiers = V_Value
	endif
	
	// do we need to calculate the threshold or the positions?
	// if not then there is no reason to run the calculation
	if ((showThreshold == 1) || (showPositions == 1) || (showFittedPositions == 1))
		SetDataFolder windowDataFolder
		Struct FitData fitParams 
		GetGUIFitSettingsAndOptions(fitParams, baseWindowName)
		Generatethreshold(image, fittedPositions, fitParams, showThreshold, showPositions, showFittedPositions, currentImage)
		SetDataFolder savDF
	endif
	
	// check if we need to show tracks
	if (showParticleTracks == 1)
		SetDataFolder windowDataFolder
		wave /WAVE W_TrackingWaves = GenerateTrackingDisplayWaves(trackingWave, showFullTrack, showTrackIdentifiers, currentImage)
		SetDataFolder savDF
		wave M_Tracks = W_TrackingWaves[0]
		wave M_TrackingColors = W_TrackingWaves[1]
		wave W_TrackIndices = W_TrackingWaves[2]
		wave M_TrackIndicesLocation = W_TrackingWaves[3]
		wave /Z M_TrackIndicesColor = W_TrackingWaves[4]
	endif
	
	// what traces are currently displayed?
	tracesInGraph = TraceNameList(baseWindowName + "#CCDViewer",";",1)
	imagesInGraph = ImageNameList(baseWindowName + "#CCDViewer",";")
	
	// check if we need to show threshold traces
	if (showThreshold == 1)	// show the threshold
		// is the threshold already on the graph? If they are not on the graph, then show them
		if  (stringmatch(imagesInGraph,"*M_SegmentedImage*") != 1)
			AppendImage /W=$baseWindowName#CCDViewer windowDataFolder:M_SegmentedImage
			ModifyImage  /W=$baseWindowName#CCDViewer M_SegmentedImage  ctab= {0.5,1,RedWhiteGreen,1}, minRGB=NaN,maxRGB=0
		endif
		
	else		// don't show the threshold
		if  (stringmatch(imagesInGraph,"*M_SegmentedImage*") == 1)
			RemoveImage /W=$baseWindowName#CCDViewer M_SegmentedImage
		endif
	endif
	
	// check if we need to show positions
	if (showPositions == 1)	// show the particles
		
		// are the particles already on the graph? If not, show them
		if  (stringmatch(tracesInGraph,"*W_particles*") != 1)
			AppendToGraph /W=$baseWindowName#CCDViewer windowDataFolder:W_particles_Y vs windowDataFolder:W_particles_X
			ModifyGraph /W=$baseWindowName#CCDViewer mode(W_particles_Y)=3,msize(W_particles_Y)=4,rgb(W_particles_Y)=(0,0,65535)
			ModifyGraph /W=$baseWindowName#CCDViewer textMarker(W_particles_Y)=0
		endif
		
	else		// don't show the particles. If they are on the graph then remove them
		if  (stringmatch(tracesInGraph,"*W_particles*") == 1)
			RemoveFromGraph /W=$baseWindowName#CCDViewer W_particles_Y
		endif
	endif
	
	// check if we need to show the fitted positions
	if (showFittedPositions == 1)
		// are the fitted positions already on the graph? If not, show them
		if  (stringmatch(tracesInGraph,"*ps_temp*") != 1)
			AppendToGraph /W=$baseWindowName#CCDViewer windowDataFolder:ps_temp_Y vs windowDataFolder:ps_temp_X
			ModifyGraph /W=$baseWindowName#CCDViewer mode(ps_temp_Y)=3,marker(ps_temp_Y)=8, rgb(ps_temp_Y)=(65535,65535,0)
		endif
			
	else		// we need to remove the positions plot if it's on the graph
		if  (stringmatch(tracesInGraph,"*ps_temp*") == 1)
			RemoveFromGraph /W=$baseWindowName#CCDViewer $"ps_temp_Y"
		endif
	endif
	
	// check if we need to show the tracks
	if (showParticleTracks == 1)
		// are the fitted positions already on the graph? If not, show them
		if  (stringmatch(tracesInGraph,"*M_ActiveTracks*") != 1)
			AppendToGraph /W=$baseWindowName#CCDViewer M_Tracks[][1] vs M_Tracks[][0]
			ModifyGraph /W=$baseWindowName#CCDViewer zColor(M_ActiveTracks)={M_TrackingColors,*,*,directRGB}
		endif
		if (showTrackIdentifiers == 1)
			if (stringMatch(tracesInGraph, "*M_TrackIndicesLocation*") == 0)
				AppendToGraph /W=$baseWindowName#CCDViewer M_TrackIndicesLocation[][1] vs M_TrackIndicesLocation[][0]
				ModifyGraph /W=$baseWindowName#CCDViewer zColor(M_TrackIndicesLocation)={M_TrackIndicesColor,*,*,directRGB}, mode(M_TrackIndicesLocation)=3
				ModifyGraph /W=$baseWindowName#CCDViewer msize(M_TrackIndicesLocation)=3,textMarker(M_TrackIndicesLocation)={W_TrackIndices,"default",0,0,5,5,5}
			endif
		else
			if (stringMatch(tracesInGraph, "*M_TrackIndicesLocation*") == 1)
				RemoveFromGraph /W=$baseWindowName#CCDViewer /Z M_TrackIndicesLocation
			endif
		endif
	else		// remove the tracking wave if its on the graph
		if  (stringmatch(tracesInGraph,"*M_ActiveTracks*") == 1)
			RemoveFromGraph /W=$baseWindowName#CCDViewer M_ActiveTracks
			RemoveFromGraph /W=$baseWindowName#CCDViewer /Z M_TrackIndicesLocation
		endif
	endif
		
	// explicitly update the display if we're running on Windows
	#if (StringMatch(IgorInfo(2), "Windows") == 1)
		DoUpdate
	#endif
	
End

Function Generatethreshold(image, fittedPositions, fitParams, showThreshold, showPositions, showFittedPositions, frameNumber)
	wave /Z image, fittedPositions
	struct FitData &fitParams
	variable showThreshold, showPositions, showFittedPositions, frameNumber
	
	// generate all waves corresponding to the emitter segmentation
	// right now this function requires that the active data folder is set to
	// one of the package subfolders
	
	Assert(frameNumber >= 0)
	Assert(WaveExists(image))
	
	// take the filtering of the localization into account:
	// if the localization method is 2D Gaussian fitting, then enable the corresponding particle verifier
	// and similar for ellipsoidal Gaussian fitting
	if (fitParams.localizationMethod == LOCALIZATION_GAUSS_FITTING)
		fitParams.PVerSymm = PARTICLEVERIFIER_2DGAUSS
	elseif (fitParams.localizationMethod == LOCALIZATION_ELLIPSOIDAL2DGAUSS)
		fitParams.PVerEllipse = PARTICLEVERIFIER_ELLIPSEGAUSS
	endif
	
	EmitterSegmentation /M=(fitParams.thresholdMethod) /G={fitParams.preprocessing, fitParams.postprocessing} /PFA=(fitParams.PFA) /SSG=(fitParams.smoothSigmaFactor) /ABS=(fitParams.directThresholdLevel) /WDTH=(fitParams.PSFWidth) /S=(showPositions) /F=(fitParams.particlefinder) /PVER={fitParams.PVerSymm, fitParams.PVerEllipse, fitParams.PVerOverlap} /R=(fitParams.minDistanceBetweenParticles) image
	
	if (showFittedPositions != 0)
		GenerateFittedPositionsWaves(fittedPositions, frameNumber)
	endif
	
	if (showPositions != 0)
		MakeParticleWavesAndLabel()
	endif
End

Function /WAVE GenerateTrackingDisplayWaves(trackWave, showFullTrack, makeTrackIndices, frameNumber)
	wave /WAVE trackWave
	variable showFullTrack, makeTrackIndices, frameNumber
	
	NewDataFolder /O :TrackingDisplayWaves
	DFREF TrackingDisplayWavesFolder = :TrackingDisplayWaves
	
	// make a wave that will display the particle tracks active in the current frame
	// if showFullTrack != 0 then show the complete track for any track that includes
	// this frameNumber in its run
	
	// if makeTrackIndices != 0 then also make another set of waves, containing the
	// an additional 2D wave with x,y coordinates, a  numeric wave providing markers,
	// and a wave providing colors for each trace
	
	// the return from this function is a wave containing two wave references:
	// the first is a 2D wave containing the x and y coordinates of the tracks
	// each track is separated from the next by a row of NaN's
	// the second is a 2D wave suitable for the color as f(z) display, to color the tracks
	
	// we need to know how to get the x and y data out of the positions
	variable xCol, yCol, zCol
	GetColumnsForEmitterPositions(trackWave, xCol, yCol, zCol)
	
	// this function is assumed to be called within the the relevant root:Packages:Localizer:<windowFolder>
	// for the requested window
	variable nTracks = DimSize(trackWave, 0)
	variable i, j
	
	Make /O/N=(0, 2) /D trackingDisplayWavesFolder:M_ActiveTracks
	wave M_ActiveTracks = trackingDisplayWavesFolder:M_ActiveTracks
	
	Make /O/N=0 /I/U trackingDisplayWavesFolder:W_TrackIndices
	Make /O/N=(0, 2) /D trackingDisplayWavesFolder:M_TrackIndicesLocation
	wave W_TrackIndices = trackingDisplayWavesFolder:W_TrackIndices
	wave M_TrackIndicesLocation = trackingDisplayWavesFolder:M_TrackIndicesLocation
	
	// get a wave containing colors for the tracks
	ColorTab2Wave Rainbow
	wave M_Colors = M_Colors
	variable nColors = DimSize(M_Colors, 0)
	
	// loop over all the tracks to find those that contain the present frame
	variable nTracksInWindow
	variable nPositionsInTrack
	variable firstOccurrence, lastOccurrence
	variable nValidTracks = 0, colorIndex
	for (i = 0; i < nTracks; i+=1)
		wave currentEmitter = trackWave[i]
		nPositionsInTrack = DimSize(currentEmitter, 0)
		firstOccurrence = currentEmitter[0][0]
		lastOccurrence = currentEmitter[nPositionsInTrack - 1][0]
		
		if (lastOccurrence < frameNumber)
			continue
		endif
		
		// make use of the fact that the trackWave is sorted according to first occurrence
		// to avoid doing unnecessary work
		if (firstOccurrence > frameNumber)
			break
		endif
		
		// if we are here then currentEmitter contains a track that is active in this frame
		nValidTracks += 1
		
		// set up an entry in the indices wave if required
		if (makeTrackIndices != 0)
			Redimension /N=(nValidTracks) W_TrackIndices
			Redimension /N=(nValidTracks, -1) M_TrackIndicesLocation
			W_TrackIndices[nValidTracks - 1] = i
			M_TrackIndicesLocation[nValidTracks - 1][0] = currentEmitter[0][xCol]
			M_TrackIndicesLocation[nValidTracks - 1][1] = currentEmitter[0][yCol]
		endif
		
		// the number of points that we have to include in the plot depends on whether the full track
		// should be shown, or whether it should be constructed as we step through the movie
		for (j = 0; j < nPositionsInTrack; j+=1)
			if ((showFullTrack != 0) || (currentEmitter[j][0] <= frameNumber))
				Redimension /N=(DimSize(M_ActiveTracks, 0) + 1, -1) M_ActiveTracks
				M_ActiveTracks[DimSize(M_ActiveTracks, 0) - 1][0] = currentEmitter[j][xCol]
				M_ActiveTracks[DimSize(M_ActiveTracks, 0) - 1][1] = currentEmitter[j][yCol]
			endif
		endfor
		
		// finish up each entry with a row of NaN's
		Redimension /N=(DimSize(M_ActiveTracks, 0) + 1, -1) M_ActiveTracks
		M_ActiveTracks[DimSize(M_ActiveTracks, 0) - 1][] = NaN
	endfor
	
	// now create the colors for the tracks
	variable nPoints = DimSize(M_ActiveTracks, 0)
	Make /O/N=(nPoints, 3) TrackingDisplayWavesFolder:M_TrackColors
	wave M_TrackColors = TrackingDisplayWavesFolder:M_TrackColors
	
	// if there are no tracks then we don't need to worry about assigning
	// colors
	if (nPoints != 0)
		variable currentTrackIndex = 0
		for (i = 0; i < nPoints - 1; i+=1)
			if (numType(M_ActiveTracks[i][0]) == 2)
				// this marks the shift to the next track
				M_TrackColors[i][] = NaN
				currentTrackIndex += 1
			endif
			
			colorIndex = floor(currentTrackIndex / nValidTracks * nColors)
			M_TrackColors[i][] = M_Colors[colorIndex][q]
		endfor
		
		// and create the colors for the track indices
		if (makeTrackIndices != 0)
			Make /O/N=(nValidTracks, 3) trackingDisplayWavesFolder:M_TrackIndicesColor
			wave M_TrackIndicesColor = trackingDisplayWavesFolder:M_TrackIndicesColor
			M_TrackIndicesColor = M_Colors[p / nValidTracks * nColors][q]
		endif
	endif
	
	Make /N=(5) /WAVE /FREE W_Result
	W_Result[0] = M_ActiveTracks
	W_Result[1] = M_TrackColors
	W_Result[2] = W_TrackIndices
	W_Result[3] = M_TrackIndicesLocation
	W_Result[4] = M_TrackIndicesColor
	return W_Result
End

Function CombineTracks_TracePopup()
	GetLastUserMenuInfo
	string baseWindowName = GetBaseWindowName(S_graphName)
	string traceName = S_traceName
	
	if (StringMatch(baseWindowName, "LocalizerViewer*") != 1)
		Abort "This menu should only be called on one of the viewers created by the localizer package"
	endif
	
	if ((StringMatch(traceName, "M_TrackIndicesLocation") != 1) && (StringMatch(traceName, "M_ActiveTracks") != 1))
		Abort "This menu should only be called on the trace showing particle tracks"
	endif
	
	// get the name of the track wave
	ControlInfo /W=$baseWindowName#TrackingControls PMSelectTrackWave
	string trackWaveName = S_Value
	wave /WAVE /Z W_TrackingWave = GetTrackingWaveReference(trackWaveName)
	if (WaveExists(W_TrackingWave) == 0)
		Abort "Unable the find the selected tracking wave"
	endif
	
	variable nTracks = DimSize(W_TrackingWave, 0)
	
	// ask the user for which tracks to combine
	variable track1Index, track2Index
	Prompt track1Index, "First track identifier:"
	Prompt track2Index, "Second track identifier:"
	DoPrompt "Select the tracks to combine", track1Index, track2Index
	if (V_flag != 0)
		return 0
	endif
	
	// some constraints need to be enforced
	if ((track1Index < 0) || (track2Index < 0) || (track1Index >= track2Index))
		Abort "Invalid track indices"
	endif
	if ((track1Index >= nTracks) || (track2Index >= nTracks))
		Abort "Invalid track indices, only " + num2str(nTracks) + " tracks are present"
	endif
	
	wave track1 = W_TrackingWave[track1Index]
	wave track2 = W_TrackingWave[track2Index]
	
	// the first position of the second track needs to come after this one
	if (track1[DimSize(track1, 0) - 1][0] >= track2[0][0])
		Abort "The second track needs to start after the first one is finished"
	endif
	
	// looks like we're all set
	// replace track1 with the combination of the emitters
	// and delete track2
	variable nInitialEmittersInTrack1 = DimSize(track1, 0)
	Redimension /N=(nInitialEmittersInTrack1 + DimSize(track2, 0), -1) track1
	track1[nInitialEmittersInTrack1, ][] = track2[p - nInitialEmittersInTrack1][q]
	
	DeletePoints track2Index, 1, W_TrackingWave
	
	// update the display
	GenerateGuiThreshold(baseWindowName)
End

Function SplitTracks_TracePopup()
	GetLastUserMenuInfo
	string baseWindowName = GetBaseWindowName(S_graphName)
	string traceName = S_traceName
	DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
	
	NVAR currentImage = windowDataFolder:V_currentImage
	
	if (StringMatch(baseWindowName, "LocalizerViewer*") != 1)
		Abort "This menu should only be called on one of the viewers created by the localizer package"
	endif
	
	if ((StringMatch(traceName, "M_TrackIndicesLocation") != 1) && (StringMatch(traceName, "M_ActiveTracks") != 1))
		Abort "This menu should only be called on the trace showing particle tracks"
	endif
	
	// get the name of the track wave
	ControlInfo /W=$baseWindowName#TrackingControls PMSelectTrackWave
	string trackWaveName = S_Value
	wave /WAVE /Z W_TrackingWave = GetTrackingWaveReference(trackWaveName)
	if (WaveExists(W_TrackingWave) == 0)
		Abort "Unable the find the selected tracking wave"
	endif
	
	variable nTracks = DimSize(W_TrackingWave, 0)
	
	// ask the user for which track to split
	variable track1Index
	Prompt track1Index, "Track identifier:"
	DoPrompt "Select track to split (new track starts at this frame)", track1Index
	if (V_flag != 0)
		return 0
	endif
	
	// some constraints need to be enforced
	if (track1Index < 0)
		Abort "Invalid track index"
	endif
	if (track1Index >= nTracks)
		Abort "Invalid track index, only " + num2str(nTracks) + " tracks are present"
	endif
	
	wave track1 = W_TrackingWave[track1Index]
	variable nPosInTrack = DimSize(track1, 0)
	
	// the track needs to be active in this image
	if ((track1[0][0] > currentImage) || (track1[nPosInTrack - 1][0] < currentImage))
		Abort "The selected track is not active in this frame"
	endif
	
	// looks like we're all set
	// determine where the track needs to be split
	variable nPosBefore, nPosAfter, i
	for (i = 0; i < nPosInTrack; i+=1)
		if (track1[i][0] >= currentImage)
			nPosBefore = i
			nPosAfter = nPosInTrack - nPosBefore
			break
		endif
	endfor
	
	if (nPosBefore < 2)
		Abort "Splitting the track here does not leave a valid first track"
	endif
	if (nPosAfter < 2)
		Abort "Splitting the track here does not leave a valid second track"
	endif
	
	// divide the track
	Make /O/N=(nPosAfter, DimSize(track1, 1)) /D /FREE track2
	track2 = track1[p + nPosBefore][q]
	
	Redimension /N=(nPosBefore, -1) track1
	
	// now look for the correct place to insert the second track
	// such that the sorting is maintained
	variable firstFrameInTrack2 = track2[0][0]
	for (i = 0; i < nTracks; i+=1)
		wave currentEmitter = W_TrackingWave[i]
		if (currentEmitter[0][0] > firstFrameInTrack2)
			break
		endif
	endfor
	
	InsertPoints i, 1, W_TrackingWave
	W_TrackingWave[i] = track2
	
	// update the display
	GenerateGuiThreshold(baseWindowName)
End

Function DeleteTrack_TracePopup()
	GetLastUserMenuInfo
	string baseWindowName = GetBaseWindowName(S_graphName)
	string traceName = S_traceName
	DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
	
	NVAR currentImage = windowDataFolder:V_currentImage
	
	if (StringMatch(baseWindowName, "LocalizerViewer*") != 1)
		Abort "This menu should only be called on one of the viewers created by the localizer package"
	endif
	
	if ((StringMatch(traceName, "M_TrackIndicesLocation") != 1) && (StringMatch(traceName, "M_ActiveTracks") != 1))
		Abort "This menu should only be called on the trace showing particle tracks"
	endif
	
	// get the name of the track wave
	ControlInfo /W=$baseWindowName#TrackingControls PMSelectTrackWave
	string trackWaveName = S_Value
	wave /WAVE /Z W_TrackingWave = GetTrackingWaveReference(trackWaveName)
	if (WaveExists(W_TrackingWave) == 0)
		Abort "Unable the find the selected tracking wave"
	endif
	
	variable nTracks = DimSize(W_TrackingWave, 0)
	
	// ask the user for which track to split
	variable track1Index
	Prompt track1Index, "Track identifier:"
	DoPrompt "Select track to remove", track1Index
	if (V_flag != 0)
		return 0
	endif
	
	// some constraints need to be enforced
	if (track1Index < 0)
		Abort "Invalid track index"
	endif
	if (track1Index >= nTracks)
		Abort "Invalid track index, only " + num2str(nTracks) + " tracks are present"
	endif
	
	// all set, just remove the track
	DeletePoints track1Index, 1, W_TrackingWave
	
	// update the display
	GenerateGuiThreshold(baseWindowName)
End
	
Function GenerateFittedPositionsWaves(PositionsWave, n)
	wave PositionsWave
	variable n
	
	Assert(WaveExists(PositionsWave))
	Assert(n >= 0)
	
	variable numberOfPositions, xCol, yCol, zCol
	
	wave /WAVE M_ExtractedPositions = ExtractPositionsInFrame(PositionsWave, n)
	wave positions = M_ExtractedPositions[0]
	
	
	numberOfPositions = DimSize(positions, 0)
	
	if (numberOfPositions == 0)	// no particles were found in this image
									// due to the (perhaps rather poor) design of the GUI
									// we need to create the waves anyway
									// we give them a size of 1 point, set to NaN
		Make /D/O/N=(numberOfPositions) ps_temp_X = NaN
		Make /D/O/N=(numberOfPositions) ps_temp_Y = NaN
	endif
	
	// get the indices of the columns containing the x and y coordinates
	getColumnsForEmitterPositions(PositionsWave, xCol, yCol, zCol)
	if ((xCol == -1) || (yCol == -1))
		Abort "The positions passed to GenerateFittedPositionsWaves() do not appear to contain any (x,y) information"
	endif
	
	// are these positions subsets starting from a particular origin?
	variable xStart, yStart
	if (NumType(NumberByKey("X START", note(PositionsWave))) != 2)
		xStart = NumberByKey("X START", note(PositionsWave))
	else
		xStart = 0
	endif
	
	if (NumType(NumberByKey("Y START", note(PositionsWave))) != 2)
		yStart = NumberByKey("Y START", note(PositionsWave))
	else
		yStart = 0
	endif
	
	Make /D/O/N=(numberOfPositions) ps_temp_X = positions[p][xCol] + xStart
	Make /D/O/N=(numberOfPositions) ps_temp_Y = positions[p][yCol] + yStart
End

Function MakeParticleWavesAndLabel()
	wave M_locatedParticles
	
	Assert(WaveExists(M_LocatedParticles))
	
	variable nParticles = DimSize(M_locatedParticles, 0)
	
	Make /D/O/N=(nParticles) W_particles_X
	Make /D/O/N=(nParticles) W_particles_Y
	
	W_particles_X = M_locatedParticles[p][1]
	W_particles_Y = M_locatedParticles[p][2]
End

Function DoPALMFitting(fp, windowName, positionsWaveName, abortOnError)
	STRUCT FitData &fp
	string windowName
	string &positionsWaveName
	variable abortOnError	// if nonzero then allow LocalizationAnalysis to abort with an error
							// if zero then do not cause aborts, but report a possible error as a nonzero return value
	
	string baseWindowName = GetBaseWindowName(windowName)
	DFREF wDF = GetWindowDataFolder(baseWindowName)
	
	NVAR pixelSize = wDF:V_CCDPixelSize
	
	// add a structure with a reference to the progress reporting function
	STRUCT LocalizerProgStruct progressStruct
	progressStruct.version = kLocalizerProgStructVersion
	FUNCREF ProgressFunctionPrototype progressStruct.func = ProgressReporterFunc
	
	// fit the data but also add the waveNote
	string waveNote
	variable err
	
	// make an output name for the position
	positionsWaveName = GetBasePosName(wDF, fp.CCDFilePath)
	
	if (abortOnError != 0)
		LocalizationAnalysis /M=(fp.localizationMethod)/D=(fp.thresholdMethod)/G={fp.preprocessing, fp.postprocessing}/F=(fp.particlefinder)/PVER={fp.PVerSymm,fp.PVerEllipse,fp.PVerOverlap}/RNG={fp.firstFrameToAnalyze,fp.lastFrameToAnalyze}/PFA=(fp.PFA)/T=(fp.directThresholdLevel)/R=(fp.minDistanceBetweenParticles)/W=(fp.PSFWidth)/PROG=progressStruct/DEST=wDF:$positionsWaveName fp.CCDFilePath
	else	// don't abort for errors, but return them as an error code (use the /Z flag for LocalizationAnalysis)
		LocalizationAnalysis /M=(fp.localizationMethod)/D=(fp.thresholdMethod)/G={fp.preprocessing, fp.postprocessing}/F=(fp.particlefinder)/PVER={fp.PVerSymm,fp.PVerEllipse,fp.PVerOverlap}/RNG={fp.firstFrameToAnalyze,fp.lastFrameToAnalyze}/PFA=(fp.PFA)/T=(fp.directThresholdLevel)/R=(fp.minDistanceBetweenParticles)/W=(fp.PSFWidth)/Z/PROG=progressStruct/DEST=wDF:$positionsWaveName fp.CCDFilePath
		err = V_flag
	endif
	
	// check if the output wave exists
	// it might not exist if the user immediately aborted the procedure
	// since it is by request, do not treat a user abort as an error
	wave /Z POS_out = wDF:$positionsWaveName
	
	if (WaveExists(POS_out) != 0)
		// append a note with information on the fit
		// most of the information is now appended by the XOP
		waveNote = ""
		
		if (pixelSize != 0)
			waveNote += "X PIXEL SIZE:" + num2str(pixelSize) + ";"
			waveNote += "Y PIXEL SIZE:" + num2str(pixelSize) + ";"
		endif
		waveNote += "CALCULATION DATE:" + date() + ";"
		
		Note /NOCR POS_out, waveNote
	endif
	
	if (abortOnError == 0)
		return err
	else
		return 0
	endif
End

Function /S GetBasePosName(inDataFolder, filePath)
	DFREF inDataFolder
	string filePath
	
	// get a temp name to use for the manual fit
	// base it on the file name, but add a suffix ("_pos")
	// and make sure it is a legal name
	string macFilePath = ParseFilePath(5, filePath, ":", 0, 1)
	string fileName = ParseFilePath(3, macFilePath, ":", 0, 1)
	string posName
	
	if (strlen(fileName) + 4 < kMaxWaveName)
		posName = fileName + "_pos"
		posName = CleanupName(posName, 0)
	else
		// make a subname
		// take the first couple of letters, and the last
		posName = fileName[0, 11] + "_" + fileName[strlen(fileName) - 11, strlen(fileName) - 1] + "_pos"
		posName =  CleanupName(posName, 0)
	endif
	
	// if there already is a positions wave with the same name, then we need to change the proposed name
	// unless that wave is in the same folder (i.e. we've updated the analysis settings and now
	// the earlier results need to be overwritten)
	if (IsUniquePositionsWaveName(posName) == 0)
		if (WaveExists(inDataFolder:$posName) && IsPositionsWave(inDataFolder:$posName))
			// posName is fine, we will overwrite a previous positions wave
		else
			posName = UniqueName(posName, 1, 0)
		endif
	endif
	
	return posName
End
	

Function /WAVE GetPositionsWaveReference(posName)
	string posName
	
	Assert(strlen(posName) > 0)
	
	// it's possible that the positions name is quoted
	// so take care of that by removing the quotes
	string noQuoteName = ReplaceString("'", posName, "")
	
	// look for a positions wave with the given name
	// only look in root:'Localized Positions' and in
	// the package folder, but do so recursively
	DFREF packageFolder = root:Packages:Localizer
	DFREF positionsFolder = root:'Localized Positions'
	
	string wavesInPackageFolder = RecursiveWaveList(packageFolder, noQuoteName, "DIMS:2")
	string wavesInPositionsFolder = RecursiveWaveList(positionsFolder, noQuoteName, "DIMS:2")
	
	// if the same positions wave appears twice then we'll keep that as an error for now
	Assert((ItemsInList(wavesInPackageFolder) == 0) || (ItemsInList(wavesInPositionsFolder) == 0))
	Assert(ItemsInList(wavesInPackageFolder) <= 1)
	Assert(ItemsInList(wavesInPositionsFolder) <= 1)
	
	if (ItemsInList(wavesInPackageFolder) == 1)
		return $StringFromList(0, wavesInPackageFolder)
	else
		return $StringFromList(0, wavesInPositionsFolder)
	endif
End

Function /WAVE GetTrackingWaveReference(trackName)
	string trackName
	
	Assert(strlen(trackName) > 0)
	
	// it's possible that the positions name is quoted
	// so take care of that by removing the quotes
	string noQuoteName = ReplaceString("'", trackName, "")
	
	// look for a tracking wave with the given name
	// only look in root:'Particle Tracks', but do so recursively
	DFREF trackFolder = root:'Particle Tracks'
	
	string wavesInTrackFolder = RecursiveWaveList(trackFolder, noQuoteName, "DIMS:1,BYTE:0,INTEGER:0,SP:0,DP:0,CMPLX:0,TEXT:0")
	
	// if the same positions wave appears twice then we'll keep that as an error for now
	Assert(ItemsInList(wavesInTrackFolder) <= 1)
	
	return $StringFromList(0, wavesInTrackFolder)
End

Function /WAVE GetPositionsWaveFromGraph(windowName)
	string windowName
	// assume that the specified window contains
	// an image or plot created with this package
	// try to get the positions wave
	// that was used to make the plot in this graph
	
	// we start by getting a special case out of the way:
	// if the user is interested in the interface panel
	// instead of one of the graphs created by the package
	// then that needs to be taken care of in a special way
	string baseWindowName = GetBaseWindowName(windowName)
	if (GrepString(baseWindowName, "^LocalizerViewer[0-9]*$") == 1)
		// this case is not very clear-cut
		// for now we assume that the positions wave of interest
		// is the one selected in the popupmenu
		ControlInfo /W=$baseWindowName#AnalysisControls PMSelectPositionsWave
		return GetPositionsWaveReference(S_Value)
	endif
	
	// get a list of traces in the graph
	string traceNames, imageNames
	traceNames = TraceNameList("", ";", 1)
	imageNames = ImageNameList("", ";")
	
	// how many traces and images do we have?
	variable nTraces = ItemsInList(traceNames), i
	variable nImages = ItemsInList(imageNames)
	
	// if there are no traces or images then return a null wave reference
	if ((nTraces == 0) && (nImages == 0))
		return $""
	endif
	
	// get the the full wave paths (names including the data folder)
	string fullWavePaths = ""
	
	for (i = 0; i < nTraces; i+=1)
		fullWavePaths += GetWavesDataFolder(TraceNameToWaveRef("", StringFromList(i, traceNames)), 2) + ";"
	endfor
	
	for (i = 0; i < nImages; i+=1)
		fullWavePaths += GetWavesDataFolder(ImageNameToWaveRef("", StringFromList(i, imageNames)), 2) + ";"
	endfor
	
	// find the wave paths that can provide us with a position
	// we assume that the generated image or traces are in 
	// root:'Localizer Images'
	string possiblePositionsWaves
	variable nPossiblePositionsWaves
	possiblePositionsWaves = ListMatch(fullWavePaths, "root:'Localizer Images':*")
	nPossiblePositionsWaves = ItemsInList(possiblePositionsWaves)
	
	// if we found no possibilities then return a null wave
	if (nPossiblePositionsWaves != 1)
		return $""
	endif
	
	// we found at least one possibility. The name of the positions wave
	// is now contained in possiblePositionsWaves as :root:'Localizer Images':<name>
	string positionsName = StringFromList(2, StringFromList(0, possiblePositionsWaves, ";"), ":")
	
	// if there is more than one match then require that these are the same
	string alternatePositionsName
	for (i = 1; i < nPossiblePositionsWaves; i+=1)	
		alternatePositionsName = StringFromList(2, StringFromList(i, possiblePositionsWaves, ";"), ":")
		if (StringMatch(positionsName, alternatePositionsName) != 1)
			return $""
		endif
	endfor
	
	// the wave name may be quoted, in which case we have to remove them
	positionsName = ReplaceString("'", positionsName, "")
	
	wave positionsWave = GetPositionsWaveReference(positionsName)
	return positionsWave
End

Function /WAVE GetPositionsWithinLimits(positions, xStart, xEnd, yStart, yEnd)
	wave positions
	variable xStart, xEnd, yStart, yEnd
	
	// returns a free wave
	
	variable nPos = DimSize(positions, 0)
	variable xCol, yCol, zCol
	variable nPositionsWithinLimits = 0, i, offset
	
	// get the indices of the columns containing the x and y coordinates
	getColumnsForEmitterPositions(positions, xCol, yCol, zCol)
	if ((xCol == -1) || (yCol == -1))
		Abort "The positions passed to ExtractSubsetOfPositions_menu() do not appear to contain any (x,y) information"
	endif
	
	
	for (i = 0; i < nPos; i+= 1)
		if ((positions[i][xCol] >= xStart) && (positions[i][yCol] >= yStart) && (positions[i][xCol] <= xEnd) && (positions[i][yCol] <= yEnd))
			nPositionsWithinLimits += 1
		endif
	endfor
	
	Make /D/FREE/N=(nPositionsWithinLimits, DimSize(positions, 1)) M_SelectedPositions
	
	offset = 0
	for (i = 0; i < nPos; i+= 1)
		if ((positions[i][xCol] >= xStart) && (positions[i][yCol] >= yStart) && (positions[i][xCol] <= xEnd) && (positions[i][yCol] <= yEnd))
			M_SelectedPositions[offset][] = positions[i][q]
			offset += 1
		endif
	endfor
	
	Note /K M_SelectedPositions, note(positions)
	
	return M_SelectedPositions
End

Function IsPositionsWave(w)
	wave w
	
	// Require that positions waves have a "LOCALIZATION METHOD" entry in the wave note
	string waveNote = Note(w)
	if (StringMatch(waveNote, "*LOCALIZATION METHOD*"))
		return 1
	else
		return 0
	endif
End

Function IsTrackingWave(w)
	wave w
	
	// Require that tracking waves have "TRACKING MAX SHIFT" entry in the wave note
	string waveNote = Note(w)
	if (StringMatch(waveNote, "*TRACKING MAX SHIFT*"))
		return 1
	else
		return 0
	endif
End

Function IsRegistrationMap(w)
	wave w
	
	// Require that tracking waves have "TYPE:RegistrationMap" entry in the wave note
	string waveNote = Note(w)
	if (StringMatch(waveNote, "*TYPE:RegistrationMap*"))
		return 1
	else
		return 0
	endif
End

Function /S GetPossiblePositionsWaves()
	string savDF = GetDataFolder(1)
	string listOfWaves = ""
	
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:Localizer
	NewDataFolder /O root:'Localized Positions'
	
	DFREF packageFolder = root:Packages:Localizer
	DFREF positionsFolder = root:'Localized Positions'
	
	// make a list of all 2D waves in the usual locations
	listOfWaves += RecursiveWaveList(packageFolder, "*", "DIMS:2")
	listOfWaves += RecursiveWaveList(positionsFolder, "*", "DIMS:2")
	
	// now select only those waves that really are positions
	variable nItems = ItemsInList(listOfWaves), nItemsInPath
	variable i
	string currentWaveName, validWaves = ""
	for (i = 0; i < nItems; i+=1)
		currentWaveName = StringFromList(i, listOfWaves)
		if (IsPositionsWave($currentWaveName))
			// this wave looks valid
			// return only the name, not the complete path
			nItemsInPath = ItemsInList(currentWaveName, ":")
			currentWaveName = StringFromList(nItemsInPath - 1, currentWaveName, ":")
			validWaves += currentWaveName + ";"
		endif
	endfor
	
	if (strlen(validWaves) == 0)
		validWaves = "* No Positions *;"
		return validWaves
	endif
	
	// sort the list alphabetically
	validWaves = SortList(validWaves, ";", 16)
	
	return validWaves
End

Function /S GetPossiblePositionsWavesForPM()
	// get all possible positions waves, but assume that this function
	// will only be called in the context of populating the popupmenu
	// in the analysis controls. Only return those positions that are global
	// (i.e. in the positions DF) or within the DF specific to the calling interface panel
	string windowName = GetNameOfTopInterfacePanel()
	string baseWindowName = GetBaseWindowName(windowName)
	string listOfWaves = ""
	
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:Localizer
	NewDataFolder /O root:'Localized Positions'
	
	DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
	DFREF positionsFolder = root:'Localized Positions'
	
	// make a list of all 2D waves in the usual locations
	listOfWaves += RecursiveWaveList(windowDataFolder, "*", "DIMS:2")
	listOfWaves += RecursiveWaveList(positionsFolder, "*", "DIMS:2")
	
	// now select only those waves that really are positions
	// determined by the fact that they have a "LOCALIZATION METHOD" entry in the wave note
	variable nItems = ItemsInList(listOfWaves), nItemsInPath
	variable i
	string currentWaveName, validWaves = ""
	for (i = 0; i < nItems; i+=1)
		currentWaveName = StringFromList(i, listOfWaves)
		if (IsPositionsWave($currentWaveName))
			// this wave looks valid
			// return only the name, not the complete path
			nItemsInPath = ItemsInList(currentWaveName, ":")
			currentWaveName = StringFromList(nItemsInPath - 1, currentWaveName, ":")
			validWaves += currentWaveName + ";"
		endif
	endfor
	
	if (strlen(validWaves) == 0)
		validWaves = "* No Positions *;"
		return validWaves
	endif
	
	// sort the list alphabetically
	validWaves = SortList(validWaves, ";", 16)
	
	return validWaves
End

Function /S GetSavedPositionsWaves()
	string listOfWaves = ""
	
	NewDataFolder /O root:'Localized Positions'
	
	DFREF positionsFolder = root:'Localized Positions'
	
	listOfWaves += RecursiveWaveList(positionsFolder, "*", "DIMS:2")
	
	// now select only those waves that really are positions
	variable nItems = ItemsInList(listOfWaves), nItemsInPath
	variable i
	string currentWaveName, validWaves = ""
	for (i = 0; i < nItems; i+=1)
		currentWaveName = StringFromList(i, listOfWaves)
		if (IsPositionsWave($currentWaveName))
			// this wave looks valid
			// return only the name, not the complete path
			nItemsInPath = ItemsInList(currentWaveName, ":")
			currentWaveName = StringFromList(nItemsInPath - 1, currentWaveName, ":")
			validWaves += currentWaveName + ";"
		endif
	endfor
	
	if (strlen(validWaves) == 0)
		validWaves = "* No Positions *;"
		return validWaves
	endif
	
	// sort the list alphabetically
	validWaves = SortList(validWaves, ";", 16)
	
	return validWaves
End

Function /S GetPossibleTrackingWaves()
	string listOfWaves = ""
	
	NewDataFolder /O root:'Particle Tracks'
	
	DFREF trackingFolder = root:'Particle Tracks'
	
	// make a list of all 2D waves in the usual locations
	listOfWaves += RecursiveWaveList(trackingFolder, "*", "DIMS:1,BYTE:0,CMPLX:0,DP:0,INTEGER:0,SP:0,TEXT:0")
	
	// now make use of the fact that the tracking waves
	// have a wavenote based on that of the positions
	variable nItems = ItemsInList(listOfWaves), nItemsInPath
	variable i
	string currentWaveName, validWaves = ""
	for (i = 0; i < nItems; i+=1)
		currentWaveName = StringFromList(i, listOfWaves)
		if (IsTrackingWave($currentWaveName))
			// this wave looks valid
			// return only the name, not the complete path
			nItemsInPath = ItemsInList(currentWaveName, ":")
			currentWaveName = StringFromList(nItemsInPath - 1, currentWaveName, ":")
			validWaves += currentWaveName + ";"
		endif
	endfor
	
	if (strlen(validWaves) == 0)
		validWaves = "* No Tracks *;"
		return validWaves
	endif
	
	// sort the list alphabetically
	validWaves = SortList(validWaves, ";", 16)
	
	return validWaves
End

Function /S GetPossibleRegistrationMaps()
	string listOfWaves = ""
	
	NewDataFolder /O root:'Registration Maps'
	
	DFREF registrationFolder = root:'Registration Maps'
	
	listOfWaves += RecursiveWaveList(registrationFolder, "*", "DIMS:2,DP:1,CMPLX:0")
	
	// now select only those waves that really are registration maps
	variable nItems = ItemsInList(listOfWaves), nItemsInPath
	variable i
	string currentWaveName, validWaves = ""
	for (i = 0; i < nItems; i+=1)
		currentWaveName = StringFromList(i, listOfWaves)
		if (IsRegistrationMap($currentWaveName))
			// this wave looks valid
			// return only the name, not the complete path
			nItemsInPath = ItemsInList(currentWaveName, ":")
			currentWaveName = StringFromList(nItemsInPath - 1, currentWaveName, ":")
			validWaves += currentWaveName + ";"
		endif
	endfor
	
	if (strlen(validWaves) == 0)
		validWaves = "* No Positions *;"
		return validWaves
	endif
	
	// sort the list alphabetically
	validWaves = SortList(validWaves, ";", 16)
	
	return validWaves
End

Function GetOutputStorageTypeAndPath(dataFilePath, outputFormat, outputPath)
	string dataFilePath	// used to provide some default output name
	variable &outputFormat
	string &outputPath
	
	outputPath = ""
	
	string macFilePath = ParseFilePath(5, dataFilePath, ":", 0, 1)
	string baseFileName = ParseFilePath(3, macFilePath, ":", 0 , 1)
	
	// ask how the user wants to save the data
	string strOutputFormat
	Prompt strOutputFormat, "Format: ", popup, "TIFF file;Deflate-Compressed TIFF File;PDE file;Igor wave"
	DoPrompt "Select the desired output format", strOutputFormat
	if (V_flag == 1)
		return 0
	endif
	
	string saveFileFilter = ""
	variable refNum
	
	// determine the output format
	StrSwitch (strOutputFormat)
		case "Igor wave":
			outputFormat = IMAGE_OUTPUT_TYPE_IGOR
			break
		case "TIFF file":
			outputFormat = IMAGE_OUTPUT_TYPE_TIFF
			saveFileFilter = "TIFF File (*.tif):.tif;"
			break
		case "Deflate-Compressed TIFF File":
			outputFormat = IMAGE_OUTPUT_TYPE_COMPR_TIFF
			saveFileFilter = "TIFF File (*.tif):.tif;"
			break
		case "PDE file":
			outputFormat = IMAGE_OUTPUT_TYPE_PDE
			saveFileFilter = "PDE File (*.pde):.pde;"
			break
		default:
			Abort "Bug: unknown argument to strswitch in GetOutputStorageTypeAndPath"
			break
	EndSwitch
	
	if (outputFormat == IMAGE_OUTPUT_TYPE_IGOR)
		outputPath = "root:'Localizer Movies':" + GetNewPositionsWaveName(CleanupName(baseFileName, 0), movieFolder, "")
		if (strlen(outputPath) == 0)
			return 0
		endif
	else
		Open /D /F=saveFileFilter refNum
		if (strlen(S_fileName) == 0)
			return 0
		endif
		outputPath = S_fileName
	endif
End

Function /WAVE ExtractPositionsInFrame(PositionsWave, frameNumber)
	wave PositionsWave
	variable frameNumber
	
	Assert(WaveExists(PositionsWave))
	Assert(frameNumber >= 0)
	
	// given a list of fitted positions, extract all positions that are present in frame frameNumber
	// do not return just the positions wave, but rather return a wave containing 2 wave references
	// the first reference is the positions wave, the second wave contains the indices of the found positions in the
	// original positions wave
	
	variable nPos = DimSize(PositionsWave, 0)
	variable start = 0, stop = nPos - 1, middle
	variable i, positionsHaveBeenConsolidated
	variable nExtractedPos
	
	Make /O/N=2 /WAVE /FREE M_ExtractedPositions
	
	variable nFramesPresentCol
	GetColumnForNFramesPresent(PositionsWave, nFramesPresentCol)
	
	// check if the positions have been consolidated before
	// if they haven't then use a binary search, if they have then perform a full loop
	ImageTransform /G=(nFramesPresentCol) sumCol, PositionsWave
	positionsHaveBeenConsolidated = V_value != DimSize(PositionsWave, 0)
	
	if (positionsHaveBeenConsolidated == 0)
		ImageTransform /G=0 getCol, PositionsWave
		wave W_ExtractedCol
		middle = BinarySearch(W_ExtractedCol, frameNumber)
		if ((middle < 0) || (PositionsWave[middle][0] != frameNumber))	// not found
			Make /D/FREE/O/N=(0, DimSize(PositionsWave, 1)) M_Positions
			Make /I/U/FREE/O/N=0 W_PositionIndices
		else
			for (i = middle; (i >= 0) && (PositionsWave[i][0] == frameNumber); i -= 1)
			endfor
			start = i + 1
			for (i = middle; (i < nPos) && (PositionsWave[i][0] == frameNumber); i += 1)
			endfor
			stop = i - 1
			
			nExtractedPos = stop - start + 1
			
			Make /D/FREE/O/N=(nExtractedPos, DimSize(PositionsWave, 1)) M_Positions
			Make /FREE/O/N=(nExtractedPos)/I/U W_PositionIndices
			
			M_Positions = PositionsWave[p + start][q]
			W_PositionIndices = p + start
		endif
	else		// positions have been consolidated
			// run over all frames up to the current frame to see if any of those emitters is still active
		Make /D/FREE/O/N=(0, DimSize(PositionsWave, 1)) M_Positions
		Make /I/U/FREE/O/N=0 W_PositionIndices
		variable offset = 0
		for (i = 0; (i < nPos) && (PositionsWave[i][0] <= frameNumber); i+=1)
			if ((PositionsWave[i][0] <= frameNumber) && ((PositionsWave[i][0] + PositionsWave[i][nFramesPresentCol]) > frameNumber))
				Redimension /N=(DimSize(M_Positions, 0) + 1, DimSize(M_Positions, 1)) M_Positions
				Redimension /N=(DimSize(W_PositionIndices, 0) + 1) W_PositionIndices
				M_Positions[offset][] = PositionsWave[i][q]
				W_PositionIndices[offset] = i
				offset += 1
			endif
		endfor
	endif
	
	M_ExtractedPositions[0] = M_Positions
	M_ExtractedPositions[1] = W_PositionIndices
	Note /NOCR /K M_ExtractedPositions[0], note(PositionsWave)
	
	KillWaves /Z W_ExtractedCol
	
	return M_ExtractedPositions
End

Function DoLClustering()
	string posName
	
	DFREF savDF = GetDataFolderDFR()
	NewDataFolder /O root:Packages
	NewDataFolder /O/S root:Packages:Localizer
	
	Variable /G root:Packages:Localizer:V_lClusteringRange
	Variable /G root:Packages:Localizer:V_lClusteringNBins
	Variable /G root:Packages:Localizer:V_lCluseringNPointsToSample
	Variable /G root:Packages:Localizer:V_lClusteringNMonteCarloSamples
	NVAR calculationRange = root:Packages:Localizer:V_lClusteringRange
	NVAR nBins = root:Packages:Localizer:V_lClusteringNBins
	NVAR nPointsToSample = root:Packages:Localizer:V_lCluseringNPointsToSample
	NVAR nMonteCarloSamples= root:Packages:Localizer:V_lClusteringNMonteCarloSamples
	
	if (calculationRange == 0)
		calculationRange = 30
	endif
	if (nBins == 0)
		nBins = 100
	endif
	if (nPointsToSample == 0)
		nPointsToSample = 15000
	endif
	if (nMonteCarloSamples == 0)
		nMonteCarloSamples = -1
	endif
	
	variable localCalculationRange = calculationRange
	variable localNBins = nBins
	variable localNPointsToSample = nPointsToSample
	variable localNMonteCarloSamples = nMonteCarloSamples
	
	Prompt posName, "Positions wave:", popup, GetPossiblePositionsWaves()
	Prompt localCalculationRange, "Calculation range (pixels):"
	Prompt localNBins, "Number of bins:"
	Prompt localNPointsToSample, "Number of points to randomly sample (-1 for all points):"
	Prompt localNMonteCarloSamples, "Monte Carlo trials for confidence analysis:"
	
	DoPrompt "Calculation parameters", posName, localCalculationRange, localNBins, localNPointsToSample, localNMonteCarloSamples
	if (V_flag == 1)	// cancel
		return 0
	endif
	
	calculationRange = localCalculationRange
	nBins = localNBins
	nPointsToSample = localNPointsToSample
	nMonteCarloSamples = localNMonteCarloSamples
	
	wave /Z pos = GetPositionsWaveReference(posName)
	if (WaveExists(pos) == 0)
		Abort "The selected wave doesn't seem to exist!"
	endif
	
	// try to get the CCD pixel size
	variable xSize, ySize, pixelSize
	GetCCDDimensionsFromPositions(pos, xSize, ySize, pixelSize)
	
	variable nPositions = DimSize(pos, 0)
	
	if ((nPointsToSample < 0) || (nPointsToSample > nPositions))
		localNPointsToSample = nPositions
	endif
	
	SetDataFolder root:Packages:Localizer
	if (localNPointsToSample < nPositions)
		// calculate only some points
		StatsSample /N=(localNPointsToSample) /MC pos
		wave positionsToAnalyze = M_Sampled
		Note /K positionsToAnalyze, note(pos)
	else
		// calculate over all points
		wave positionsToAnalyze = pos
	endif
	
	RipleyLFunctionClustering /RNGE={calculationRange, nBins} positionsToAnalyze
	
	SetDataFolder savDF
	
	wave W_lFunction = root:Packages:Localizer:W_lFunction
	Duplicate /O W_lFunction, root:Packages:Localizer:W_LFunctionSubX
	wave W_LFunctionSubX = root:Packages:Localizer:W_LFunctionSubX
	
	W_LFunctionSubX = W_lFunction - x
	
	// adjust the size to nm if possible
	if (pixelSize != 1)
		SetScale /I x, 0, calculationRange * pixelSize, W_LFunctionSubX
	endif
	
	DoWindow /F LFunction
	if (V_flag != 1)
		Display /K=1/N=LFunction W_LFunctionSubX
		Label /W=LFunction Left, "L Function - x"
	endif
	
	if (pixelSize == 0)
		Label /W=LFunction Bottom, "Distance (pixel)"
	else
		Label /W=LFunction Bottom, "Distance (nm)"
	endif
	
	if (nMonteCarloSamples > 1)
		// also simulate confidence bands
		wave /WAVE W_SimulationResult = SimulateRipleyConfidenceBounds(positionsToAnalyze, nMonteCarloSamples, calculationRange, nBins, xSize, ySize)
		wave lowerBound = W_SimulationResult[0]
		wave upperBound = W_SimulationResult[1]
		// convert to L(r) - r
		lowerBound = lowerBound - x
		upperBound = upperBound - x
		Duplicate /O lowerBound, root:Packages:Localizer:W_LFunctionLowerBound
		Duplicate /O upperBound, root:Packages:Localizer:W_LFunctionUpperBound
		SetScale /I x, 0, calculationRange * pixelSize, root:Packages:Localizer:W_LFunctionLowerBound, root:Packages:Localizer:W_LFunctionUpperBound
		AppendToGraph /W=LFunction root:Packages:Localizer:W_LFunctionLowerBound
		AppendToGraph /W=LFunction root:Packages:Localizer:W_LFunctionUpperBound
	else
		// remove the confidence bounds if they are being shown
		RemoveFromGraph /W=LFunction /Z W_LFunctionLowerBound
		RemoveFromGraph /W=LFunction /Z W_LFunctionUpperBound
	endif
End

Function /WAVE SimulateRipleyConfidenceBounds(positions, nSamples, calculationRange, nBins, xSize, ySize)
	wave positions	// only used as a template, not modified by this function
	variable  nSamples // number of simulations to peform
	variable calculationRange, nBins // passed to the /RNGE flag of the operation
	variable xSize, ySize // the dimensions of the region to simulate over
	
	// the percentiles will be returned as a wave containing two wave references
	// elem 0 is the low percentile, elem 1 the high
	
	DFREF savDF = GetDataFolderDFR()
	DFREF tempDF = NewFreeDataFolder()
	
	SetDataFolder tempDF
	
	variable i, xCol, yCol, zCol
	
	GetColumnsForEmitterPositions(positions, xCol, yCol, zCol)
	
	Duplicate /FREE positions, M_RandomSample
	Make /N=(nSamples, nBins) /D /FREE M_SimulationResults
	Make /O/N=(nBins) /FREE/D W_LowPercentile, W_HighPercentile
	Make /N=2 /FREE /WAVE W_Result
	
	for (i = 0; i < nSamples; i+=1)
		ProgressWindow("MonteCarlo", i, nSamples)
		M_RandomSample[][xCol] = enoise(xSize / 2)
		M_RandomSample[][yCol] = enoise(ySize / 2)
		
		RipleyLFunctionClustering /RNGE={calculationRange, nBins} M_RandomSample
		wave W_LFunction
		M_SimulationResults[i][] = W_LFunction[q]
	endfor
	
	SetScale /P x, DimOffset(W_LFunction, 0), DimDelta(W_LFunction, 0), W_LowPercentile, W_HighPercentile
	
	for (i = 0; i < nBins; i+=1)
		// get the percentiles of each bin
		MatrixOP /FREE W_ExtractedCol = col(M_SimulationResults, i)
		Sort W_ExtractedCol, W_ExtractedCol
		W_LowPercentile[i] = W_ExtractedCol(0.025 * (nSamples - 1) * DimDelta(W_ExtractedCol, 0) + DimOffset(W_ExtractedCol, 0))
		W_HighPercentile[i] = W_ExtractedCol(0.975 * (nSamples - 1) * DimDelta(W_ExtractedCol, 0) + DimOffset(W_ExtractedCol, 0))
	endfor
	SetDataFolder savDF
	
	W_Result[0] = W_LowPercentile
	W_Result[1] = W_HighPercentile
	
	return W_Result
End

Function GenerateScatterPlot_Menu()
	string posName
	string windowName
	string colorScaleString
	
	Prompt posName, "Positions wave:", popup, GetPossiblePositionsWaves()
	Prompt colorScaleString, "Color scale:", popup, "None;" + CTabList()
	
	DoPrompt "Select the positions wave", posName, colorScaleString
	if (V_flag == 1)	// cancel
		return 0
	endif
	
	if (StringMatch(colorScaleString, "None") == 1)
		colorScaleString = ""
	endif
	
	wave /Z pos = GetPositionsWaveReference(posName)
	if (WaveExists(pos) == 0)
		Abort "The selected wave doesn't seem to exist!"
	endif
	
	variable xSize, ySize, pixelSize
	GetCCDDimensionsFromPositions(pos, xSize, ySize, pixelSize)
	
	DFREF savDF = GetDataFolderDFR()
	NewDataFolder /O root:'Localizer Images'
	NewDataFolder /S/O root:'Localizer Images':$posName
	GenerateScatterPlot(pos, xSize, ySize, pixelSize, colorScaleString)
	wave Scatter_positions_X =  Scatter_positions_X 
	wave Scatter_positions_Y =  Scatter_positions_Y
	wave /Z Scatter_positions_color = Scatter_positions_color
	SetDataFolder savDF
	
	windowName = "Scatter plot from " + posName
	
	Display /K=1 /N=Scatter_Viewer Scatter_positions_Y vs Scatter_positions_X as windowName
	if (strlen(colorScaleString) != 0)
		ModifyGraph zColor(Scatter_positions_Y)={Scatter_positions_color,*,*,directRGB,0}
	endif
	ModifyGraph mode=3,marker=41,msize=2,tickUnit=1
	ModifyGraph /W=$S_name width={Aspect, xSize / ySize}, height=0
	
	if (pixelSize == 0)
		SetAxis /W=$S_name bottom, 0, xSize - 1
		SetAxis /W=$S_name left, 0, ySize - 1
		Label /W=$S_name left "Distance (pixel)"
		Label /W=$S_name bottom "Distance (pixel)"
	else
		SetAxis /W=$S_name bottom, 0, (xSize - 1) * pixelSize / 1000
		SetAxis /W=$S_name left, 0, (ySize - 1) * pixelSize / 1000
		Label /W=$S_name left "Distance (m)"
		Label /W=$S_name bottom "Distance (m)"
	endif
	
End

Function GenerateScatterPlot(positions, xSize, ySize, pixelSize, colorScaleString)
	wave positions
	variable xSize, ySize, pixelSize
	string colorScaleString	// the name of an Igor color table
	
	Assert(WaveExists(positions))
	Assert((xSize > 0) && (ySize > 0))
	
	variable n_positions = DimSize(positions, 0)
	variable xCol, yCol, zCol
	
	variable number_of_images = positions[n_positions - 1][0] - positions[0][0] + 1
	variable pixelSizeUM = pixelSize / 1000
	
	Make /D/O/N=(n_positions) Scatter_positions_X
	Make /D/O/N=(n_positions) Scatter_positions_Y
	wave Scatter_positions_X
	wave Scatter_positions_Y
	
	// get the indices of the columns containing the x and y coordinates
	getColumnsForEmitterPositions(positions, xCol, yCol, zCol)
	if ((xCol == -1) || (yCol == -1))
		Abort "The positions passed to GenerateScatterPlot() do not appear to contain any (x,y) information"
	endif
	
	Scatter_positions_X = positions[p][xCol]
	Scatter_positions_Y = positions[p][yCol]
	
	if (pixelSize != 0)
		Scatter_positions_X *= pixelSizeUM
		Scatter_positions_Y *= pixelSizeUM
		SetScale /P x, 0, 1, "um", Scatter_positions_X
		SetScale /P x, 0, 1, "um", Scatter_positions_Y
	else
		SetScale /P x, 0, 1, "pixel", Scatter_positions_X
		SetScale /P x, 0, 1, "pixel", Scatter_positions_Y
	endif
	
	if (strlen(colorScaleString) != 0)
		// choose the colors
		ColorTab2Wave $colorScaleString
		wave M_Colors
		Make /O/N=(n_positions, 3) Scatter_positions_color
		wave Scatter_positions_color = Scatter_positions_color
		
		variable nColors = DimSize(M_Colors, 0)
		
		Scatter_positions_color = M_Colors[floor(p / n_positions * nColors)][q]
	endif
	
	KillWaves /Z M_Colors
End

Function MakeAccumulatedImage_Menu()
	Variable /G root:Packages:Localizer:V_AccumulatedBinSize
	
	NVAR binSize = root:Packages:Localizer:V_AccumulatedBinSize
	if (binSize == 0)
		binSize = 0.2
	endif
	
	string posName
	string windowTitle
	variable colorScaleIndex, localBinSize
	string colorScaleString
	
	DFREF savDF = GetDataFolderDFR()
	
	localBinSize = binSize
	Prompt localBinSize, "Bin size (pixels):"
	Prompt posName, "Positions wave:", popup, GetPossiblePositionsWaves()
	Prompt colorScaleString, "Color scale:", popup, CTabList()
	DoPrompt "Select the pixel size and positions wave", posName, localBinSize, colorScaleString
	if (V_flag == 1)	// cancel
		return 0
	endif
	binSize = localBinSize
	
	wave /Z pos = GetPositionsWaveReference(posName)
	
	if (WaveExists(pos) == 0)
		Abort "The selected wave doesn't seem to exist!"
	endif
	
	NewDataFolder /O/S root:'Localizer Images'
	NewDataFolder /S/O root:'Localizer Images':$posName
	
	variable xSize, ySize, pixelSize
	GetCCDDimensionsFromPositions(pos, xSize, ySize, pixelSize)
	
	MakeAccumulatedImage(pos, xSize, ySize, binSize, pixelSize)
	
	wave M_AccumulatedImage
	windowTitle = "Accumulated image from " + posName
	
	// Typically the histogram for the resulting image is very inhomogeneous
	// a reasonable guess for an initial colorscale appears to be to set min to auto
	// and then take 10% of the maximum for the upper limit
	// this is similar to the bitmap image
	WaveStats /Q /M=1 M_AccumulatedImage
	
	Display /K=1 /N=AccumulatedImageViewer as windowTitle
	AppendImage /W=$S_name M_AccumulatedImage
	AdjustGraphForImageDisplay(S_name)
	ModifyImage /W=$S_name M_AccumulatedImage ctab= {*,V_max / 10,$colorScaleString,0}
	ModifyGraph /W=$S_name width={Aspect, xSize / ySize}, height=0,tickUnit=1
	
	if (pixelSize == 0)
		Label /W=$S_name left "Distance (pixels)"
		Label /W=$S_name bottom "Distance (pixels)"
	else
		Label /W=$S_name left "Distance (m)"
		Label /W=$S_name bottom "Distance (m)"
	endif
	
	SetDataFolder savDF
	
End

Function MakeAccumulatedImage(pos, xSize, ySize, binSize, pixelSize)
	wave pos
	variable xSize, ySize, binSize, pixelSize
	
	Assert(WaveExists(pos))
	Assert((xSize > 0) && (ySize > 0) && (binSize > 0))
	
	variable nPoints = DimSize(pos, 0)
	variable xWidth = ceil(xSize / binSize)
	variable yWidth = ceil(ySize / binSize)
	
	variable px, py, i, xCol, yCol, zCol
	
	// get the indices of the columns containing the x and y coordinates
	getColumnsForEmitterPositions(pos, xCol, yCol, zCol)
	if ((xCol == -1) || (yCol == -1))
		Abort "The positions passed to MakeAccumulatedImage() do not appear to contain any (x,y) information"
	endif
	
	Make /D/O/N=(xWidth, yWidth) M_AccumulatedImage
	if (pixelSize == 0)	// no pixel size specified
		SetScale /I x, 0, xSize - 1, "pixel", M_AccumulatedImage
		SetScale /I y, 0, ySize - 1, "pixel", M_AccumulatedImage
	else		// pixel size specified, rescale the nm pixel size into um
		SetScale /I x, 0, (xSize - 1) * pixelSize / 1000, "um", M_AccumulatedImage
		SetScale /I y, 0, (ySize - 1) * pixelSize / 1000, "um", M_AccumulatedImage
	endif
	
	M_AccumulatedImage = 0
	
	for (i = 0; i < nPoints; i+=1)
		px = floor(pos[i][xCol] / binSize)
		py = floor(pos[i][yCol] / binSize)
		
		M_AccumulatedImage[px][py] += 1
	endfor
	
End

Function MakePALMImage_Menu()
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:Localizer
	
	Variable /G root:Packages:Localizer:V_PALMBitmapConstantWidth
	Variable /G root:Packages:Localizer:V_PALMBitmapFitUncertainty
	Variable /G root:Packages:Localizer:V_PALMBitmapCamMagnification
	Variable /G root:Packages:Localizer:V_PALMBitmapCamOffset
	Variable /G root:Packages:Localizer:V_PALMCCDPixelSize
	Variable /G root:Packages:Localizer:V_PALMBitmapImageWidth
	Variable /G root:Packages:Localizer:V_gaussianWidth
	Variable /G root:Packages:Localizer:V_PALMBitmapSelectedMethod
	
	NVAR constantWidth = root:Packages:Localizer:V_PALMBitmapConstantWidth
	NVAR fitUncertainty = root:Packages:Localizer:V_PALMBitmapFitUncertainty
	NVAR camMagnification = root:Packages:Localizer:V_PALMBitmapCamMagnification
	NVAR camOffset = root:Packages:Localizer:V_PALMBitmapCamOffset
	NVAR imageWidth = root:Packages:Localizer:V_PALMBitmapImageWidth
	NVAR PSFWidth = root:Packages:Localizer:V_gaussianWidth
	NVAR SelectedMethod = root:Packages:Localizer:V_PALMBitmapSelectedMethod
	
	// provide initial values if this is the first time that we've used this function
	if (constantWidth == 0)
		constantWidth = 0.5
	endif
	if (fitUncertainty == 0)
		fitUncertainty = 3
	endif
	if (imageWidth == 0)
		imageWidth = 2096
	endif
	if (PSFWidth == 0)
		PSFWidth = 2
	endif
	if (camMagnification == 0)
		camMagnification = 1
	endif
	if (camOffset == 0)
		camOffset = 1000
	endif
	
	string listPositionsFunc = GetIndependentModuleName() + "#GetPossiblePositionsWaves()"
	
	DoWindow /F PALMBitmapWindow
	if (V_flag == 0)
		NewPanel /N=PALMBitmapWindow /K=1 /W=(137,135,397,490) as "Enter calculation parameters"
		GroupBox GBPALMBitmapColorBox,pos={4,65},size={250,25}
		PopupMenu PMPALMBitmapSelectPositionsWave,pos={14,8},size={220,20},bodyWidth=150,title="Positions wave:"
		PopupMenu PMPALMBitmapSelectPositionsWave,mode=1,value= #listPositionsFunc
		PopupMenu PMPALMBitmapSelectColorTable,pos={35,33},size={198,20},bodyWidth=145,title="Color table:"
		PopupMenu PMPALMBitmapSelectColorTable,mode=7,value= #"\"*COLORTABLEPOPNONAMES*\""
		
		GroupBox GBPALMBitmapEmitterWeighing,pos={4,102},size={250,57},title="Relative position weight:"
		CheckBox CBPALMBitmapSameEmitterWeight,pos={11,119},size={209,14},proc=CBPALMBitmapImage,title="Each localized position has the same weight"
		CheckBox CBPALMBitmapSameEmitterWeight,value= 1,mode=1
		CheckBox CBPALMBitmapPreserveIntegral,pos={11,139},size={188,14},proc=CBPALMBitmapImage,title="Preserve the fitted integrated intensity"
		CheckBox CBPALMBitmapPreserveIntegral,value= 0,mode=1
		
		GroupBox GBPALMBitmapDeviationBox,pos={3,171},size={250,150},title="Spot deviation:"
		CheckBox CBPALMBitmapConstantWidth,pos={10,191},size={168,14},proc=CBPALMBitmapImage,title="Constant width:\t\t\t\tCCD pixels"
		CheckBox CBPALMBitmapConstantWidth,value= 0,mode=1
		CheckBox CBPALMBitmapFitUncertainty,pos={10,214},size={166,14},proc=CBPALMBitmapImage,title="Scale fit uncertainty\t\t\t\ttimes"
		CheckBox CBPALMBitmapFitUncertainty,value= 0,mode=1
		CheckBox CBPALMBitmapGME,pos={10,239},size={130,14},proc=CBPALMBitmapImage,title="Gaussian mask estimation"
		CheckBox CBPALMBitmapGME,value= 0,mode=1
		SetVariable SVPALMBitmapImageWidth,pos={7,69},size={140,15},title="Image width (pixels)"
		SetVariable SVPALMBitmapImageWidth,limits={32,inf,0},value=imageWidth
		SetVariable SVPALMBitmapConstantWidth,pos={103,191},size={30,15},title=" "
		SetVariable SVPALMBitmapConstantWidth,limits={0,inf,0},value=constantWidth
		SetVariable SVPALMBitmapFitUncertainty,pos={121,214},size={30,15},title=" "
		SetVariable SVPALMBitmapFitUncertainty,limits={0,inf,0},value=fitUncertainty
		SetVariable SVPALMCameraMagnification,pos={29,258},size={190,15},title="Camera amplification factor:"
		SetVariable SVPALMCameraMagnification,limits={1e-4,inf,0},value=camMagnification
		SetVariable SVPALMCameraOffset,pos={28,278},size={190,15},title="Camera offset:"
		SetVariable SVPALMCameraOffset,limits={0,inf,0},value=camOffset
		SetVariable SVPALMBitmapPSFWidth,pos={28,298},size={190,15},title="PSF width (in CCD pixels):"
		SetVariable SVPALMBitmapPSFWidth,limits={0,inf,0},value=PSFWidth
		Button BTPALMBitmapImageCalculate,pos={179,329},size={75,20},proc=BTPALMBitmapImageCalculateProc,title="Calculate!"
		
		// if the window was opened previously then restore its size and position
		WC_WindowCoordinatesRestore("PALMBitmapWindow")
		SetWindow PALMBitmapWindow hook(WindowCoordsHook)=WC_WindowCoordinatesNamedHook
	endif
	
	// restore the selected method if we used it previously
	Switch (SelectedMethod)
		case 0:
			CheckBox CBPALMBitmapConstantWidth value= 1
			break
		case 1:
			CheckBox CBPALMBitmapFitUncertainty value= 1
			break
		case 2:
			CheckBox CBPALMBitmapGME value = 1
			break
	EndSwitch
	
End

Function BTPALMBitmapImageCalculateProc(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	switch( ba.eventCode )
		case 2: // mouse up
			// click code here
			NVAR constantWidth = root:Packages:Localizer:V_PALMBitmapConstantWidth
			NVAR uncertaintyScale = root:Packages:Localizer:V_PALMBitmapFitUncertainty
			NVAR camOffset = root:Packages:Localizer:V_PALMBitmapCamOffset
			NVAR camMagnification = root:Packages:Localizer:V_PALMBitmapCamMagnification
			NVAR imageWidth = root:Packages:Localizer:V_PALMBitmapImageWidth
			NVAR PSFWidth = root:Packages:Localizer:V_gaussianWidth
			string posName
			string colorTable
			variable method = -1
			variable weighingMethod = 0
			ControlInfo /W=PALMBitmapWindow PMPALMBitmapSelectPositionsWave
			posName = S_Value
			
			ControlInfo /W=PALMBitmapWindow PMPALMBitmapSelectColorTable
			colorTable = S_value
			
			ControlInfo /W=PALMBitmapWindow CBPALMBitmapSameEmitterWeight
			if (V_value == 1)
				weighingMethod = 0
			endif
			
			ControlInfo /W=PALMBitmapWindow CBPALMBitmapPreserveIntegral
			if (V_value == 1)
				weighingMethod = 1
			endif
			
			ControlInfo /W=PALMBitmapWindow CBPALMBitmapConstantWidth
			if (V_value == 1)
				method = PALMBITMAP_DEV_SAME
			endif
			ControlInfo /W=PALMBitmapWindow CBPALMBitmapFitUncertainty
			if (V_value == 1)
				method = PALMBITMAP_DEV_FITUNCERTAINTY
			endif
			ControlInfo /W=PALMBitmapWindow CBPALMBitmapGME
			if (V_value == 1)
				method = PALMBITMAP_DEV_GAUSSIANMASK
			endif
				
			
			
			wave /Z pos = GetPositionsWaveReference(posName)
			
			if (WaveExists(pos) == 0)
				Abort "The selected wave doesn't seem to exist!"
			endif
			
			DFREF savDF = GetDataFolderDFR()
			NewDataFolder /O root:'Localizer Images'
			NewDataFolder /O/S root:'Localizer Images':$posName
			
			// if there is already a PALM image in this data folder then the calculation will overwrite it
			// try to kill it before the actual calculation to maybe free up some memory
			KillWaves /Z $"M_LocalizationBitmap"
			
			variable xSize, ySize, pixelSize
			GetCCDDimensionsFromPositions(pos, xSize, ySize, pixelSize)
			
			variable imageScaleFactor = round(imageWidth / xSize)
			variable upperLimit = xSize
			
			// add a structure with a reference to the progress reporting function
			STRUCT LocalizerProgStruct progressStruct
			progressStruct.version = kLocalizerProgStructVersion
			FUNCREF ProgressFunctionPrototype progressStruct.func = ProgressReporterFunc
			
			switch (method)
				case PALMBITMAP_DEV_SAME:
					LocalizationBitmap /M=(method) /WGHT=(weighingMethod) /S=(constantWidth) /W={xSize, ySize, imageScaleFactor} /PROG=progressStruct pos
					break
				case PALMBITMAP_DEV_FITUNCERTAINTY:
					LocalizationBitmap /M=(method) /WGHT=(weighingMethod) /S=(uncertaintyScale) /L=(upperLimit) /W={xSize, ySize, imageScaleFactor} /PROG=progressStruct pos
					break
				case PALMBITMAP_DEV_GAUSSIANMASK:
					LocalizationBitmap /M=(method) /WGHT=(weighingMethod) /WDTH=(PSFWidth) /CAL={camOffset, camMagnification} /W={xSize, ySize, imageScaleFactor} /PROG=progressStruct pos
					break
				default:
					SetDataFolder savDF
					Abort "Unknown method"
					break
			endswitch
			
			wave M_LocalizationBitmap
			string windowTitle = "PALM bitmap image from " + posName
			
			// Typically the histogram for the resulting image is very inhomogeneous
			// a reasonable guess for an initial colorscale appears to be to set min to auto
			// and then take 10% of the maximum for the upper limit
			WaveStats /Q /M=1 M_LocalizationBitmap
			
			Display /K=1 /N=PALMImageViewer as windowTitle
			AppendImage /W=$S_name M_LocalizationBitmap
			AdjustGraphForImageDisplay(S_name)
			ModifyImage /W=$S_name M_LocalizationBitmap ctab= {*, V_max / 10,$colorTable,0}
			ModifyGraph /W=$S_name width={Aspect, xSize / ySize}, height=0
			
			SetDataFolder savDF
			
			if (pixelSize == 0)
				SetScale /I x, 0, xSize -1, "pixel", M_LocalizationBitmap
				SetScale /I y, 0, ySize -1, "pixel", M_LocalizationBitmap
				Label /W=$S_name left "Distance (pixels)"
				Label /W=$S_name bottom "Distance (pixels)"
			else
				SetScale /I x, 0, (xSize -1) * pixelSize / 1000, "um", M_LocalizationBitmap
				SetScale /I y, 0, (ySize -1) * pixelSize / 1000, "um", M_LocalizationBitmap
				Label /W=$S_name left "Distance (m)"
				Label /W=$S_name bottom "Distance (m)"
			endif
			
			AppendColorRangeSliders()
			
			DoWindow /K PALMBitmapWindow
					
			break
	endswitch

	return 0
End

Function CBPALMBitmapImage(cba) : CheckBoxControl
	STRUCT WMCheckboxAction &cba

	switch( cba.eventCode )
		case 2: // mouse up
			String controlName = cba.ctrlName
			Variable checked = cba.checked
			NVAR selectedMethod = root:Packages:Localizer:V_PALMBitmapSelectedMethod
			
			StrSwitch(controlName)
				// checkboxes that affect how different positions are weighed
				case "CBPALMBitmapSameEmitterWeight":
					CheckBox CBPALMBitmapPreserveIntegral value = 0
					break
				case "CBPALMBitmapPreserveIntegral":
					CheckBox CBPALMBitmapSameEmitterWeight value = 0
					break
				
				
				// checkboxes that affect the calculation of the spot standard deviation
				case "CBPALMBitmapConstantWidth":
					CheckBox CBPALMBitmapFitUncertainty value=0
					CheckBox CBPALMBitmapGME value=0
					selectedMethod = 0
					break
				case "CBPALMBitmapFitUncertainty":
					CheckBox CBPALMBitmapConstantWidth value=0
					CheckBox CBPALMBitmapGME value=0
					selectedMethod = 1
					break
				case "CBPALMBitmapGME":
					CheckBox CBPALMBitmapConstantWidth value=0
					CheckBox CBPALMBitmapFitUncertainty value=0
					selectedMethod = 2
					break
			EndSwitch
			
			break
	endswitch

	return 0
End

Function SaveCCDImagesAs()
	
	string windowName = GetNameOfTopInterfacePanel()
	if (strlen(windowName) == 0)
		Abort "No data file seems to have been opened, or the top panel does not belong tot this package"
	endif
	
	DFREF windowDataFolder = GetWindowDataFolder(windowName)
	SVAR CCDFilePath = windowDataFolder:S_filePath
	
	string outputFormatStr, outputFilePath, extension
	variable outputFormat
	
	Prompt outputFormatStr, "Output format:", popup, "TIFF File;Compressed TIFF File;PDE File;Presentation Movie;"
	DoPrompt "Choose the output format", outputFormatStr
	if (V_flag == 1)
		return 0
	endif
	
	StrSwitch (outputFormatStr)
		case "TIFF File":
			outputFormat = IMAGE_OUTPUT_TYPE_TIFF
			extension = ".tif"
			break
		case "Compressed TIFF File":
			outputFormat = IMAGE_OUTPUT_TYPE_COMPR_TIFF
			extension = ".tif"
			break
		case "PDE File":
			outputFormat = IMAGE_OUTPUT_TYPE_PDE
			extension = ".pde"
			break
		case "Presentation Movie":
			// this is handled by a different function, not by ProcessCCDImages
			MakeMovieFromOpenFile(windowName)
			return 0
			break
		default:
			Abort "Internal error in SaveCCDImagesAs(): no such format"
	EndSwitch
	
	variable refNum
	String fileFilter = "Image files (" + extension + "):" + extension
	string suggestedOutputPath = ParseFilePath(5, CCDFilePath, ":", 0, 1)
	suggestedOutputPath = ParseFilePath(3, CCDFilePath, ":", 0, 1) + extension
	Open /F=(fileFilter) /D refNum as suggestedOutputPath
	
	outputFilePath = S_fileName
	if (strlen(S_fileName) == 0)
		return 0
	endif
	
	// add a structure with a reference to the progress reporting function
	STRUCT LocalizerProgStruct progressStruct
	progressStruct.version = kLocalizerProgStructVersion
	FUNCREF ProgressFunctionPrototype progressStruct.func = ProgressReporterFunc
	
	ProcessCCDImages /O /M=(PROCESSCCDIMAGE_CONVERTFORMAT) /OUT=(outputFormat) /PROG=progressStruct CCDFilePath, outputFilePath
	
End

Function MakeParticleTrackingScatterPlot()
	string trackingWaveList = GetPossibleTrackingWaves()
	if (ItemsInList(trackingWaveList) == 0)
		Abort "No particle tracks seem to have been calculated"
	endif
	
	string trackingWaveName, colorTableName
	Prompt trackingWaveName, "Track wave:", popup, trackingWaveList
	Prompt colorTableName, "Color table:", popup, "None;" + CTabList()
	DoPrompt "Select the tracks to display", trackingWaveName, colorTableName
	
	wave /WAVE /Z trackWave = GetTrackingWaveReference(trackingWaveName)
	if (WaveExists(trackWave) == 0)
		Abort "Unable to find the requested track"
	endif
	
	variable nTracks = DimSize(trackWave, 0)
	
	variable xSize, ySize, pixelSize
	GetCCDDimensionsFromPositions(trackWave, xSize, ySize, pixelSize)
	variable xCol, yCol, zCol
	GetColumnsForEmitterPositions(trackWave, xCol, yCol, zCol)
	
	NewDataFolder /O root:'Tracking Images'
	NewDataFolder /O root:'Tracking Images':$trackingWaveName
	DFREF imageFolder = root:'Tracking Images':$trackingWaveName
	DFREF savDF = GetDataFolderDFR()
	
	Make /O/N=0 /D imageFolder:W_trackScatter_X, imageFolder:W_trackScatter_Y
	wave W_trackScatter_X = imageFolder:W_trackScatter_X
	wave W_trackScatter_Y = imageFolder:W_trackScatter_Y
	
	variable useColorTable = 0
	variable nColors
	if (StringMatch(colorTableName, "None") == 0)	
		useColorTable = 1
		
		SetDataFolder root:Packages:Localizer
		ColorTab2Wave $colorTableName
		SetDataFolder savDF
		
		wave M_Colors = root:Packages:Localizer:M_Colors
		nColors = DimSize(M_Colors, 0)
		Make /O/N=(0, 3) /W/U imageFolder:W_trackScatter_Colors
		wave W_trackScatter_Colors = imageFolder:W_trackScatter_Colors
	endif
	
	variable nPositionsInThisTrack
	variable offset, i, colorIndex
	for (i = 0; i < nTracks; i+=1)
		wave thisTrack = trackWave[i]
		nPositionsInThisTrack = DimSize(thisTrack, 0)
		offset = DimSize(W_trackScatter_Y, 0)
		Redimension /N=(DimSize(W_trackScatter_Y, 0) + nPositionsInThisTrack + 1) W_trackScatter_X, W_trackScatter_Y
		W_trackScatter_X[offset, offset + nPositionsInThisTrack - 1] = thisTrack[p - offset][xCol]
		W_trackScatter_Y[offset, offset + nPositionsInThisTrack - 1] = thisTrack[p - offset][yCol]
		
		// each track entry ends with a NaN
		W_trackScatter_X[DimSize(W_trackScatter_X, 0) - 1] = NaN
		W_trackScatter_Y[DimSize(W_trackScatter_Y, 0) - 1] = NaN
		
		// also make colors if required
		if (useColorTable != 0)
			colorIndex = floor(i / nTracks * nColors)
			offset = DimSize(W_trackScatter_Colors, 0)
			Redimension /N=(DimSize(W_trackScatter_X, 0), -1) W_trackScatter_Colors
			W_trackScatter_Colors[offset, offset + nPositionsInThisTrack][] = M_Colors[colorIndex][q]
		endif
	endfor
	
	Display /K=1 /N=TrackingScatter W_trackScatter_Y vs W_trackScatter_X
	if (useColorTable != 0)
		ModifyGraph zColor={W_trackScatter_Colors, *, *, DirectRGB}
	endif
	
	// take the pixel size into account
	if (pixelSize == 0)
		SetAxis /W=$S_name bottom, 0, xSize - 1
		SetAxis /W=$S_name left, 0, ySize - 1
		Label /W=$S_name left "Distance (pixel)"
		Label /W=$S_name bottom "Distance (pixel)"
	else
		W_trackScatter_X *= pixelSize / 1000
		W_trackScatter_Y *= pixelSize / 1000
		SetAxis /W=$S_name bottom, 0, (xSize - 1) * pixelSize / 1000
		SetAxis /W=$S_name left, 0, (ySize - 1) * pixelSize / 1000
		Label /W=$S_name left "Distance (m)"
		Label /W=$S_name bottom "Distance (m)"
	endif
End

Function MakeMovieFromOpenFile(windowName)
	String windowName
	
	Assert(strlen(windowName) > 0)
	
	string baseWindowName = GetBaseWindowName(windowName)
	DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
	
	SVAR dataFilePath = windowDataFolder:S_filePath
	NVAR nCCDFrames = windowDataFolder:V_nImages
	NVAR currentImage = windowDataFolder:V_currentImage
	
	variable firstFrameInMovie = 0, lastFrameInMovie = nCCDFrames - 1
	variable frameDelta = 1
	variable originalFrame = currentImage
	string movieFormat
	variable fps = 25, refnum, i
	string outputFilePath
	string error
	
	Prompt firstFrameInMovie, "First frame in movie:"
	Prompt lastFrameInMovie, "Last frame in movie:"
	Prompt frameDelta, "Display every nth frame:"
	#ifdef MACINTOSH
		Prompt movieFormat, "Movie format:", popup, "Quicktime;"
	#else
		Prompt movieFormat, "Movie format:", popup, "Quicktime;AVI;"
	#endif
	
	DoPrompt "Enter movie parameters", firstFrameInMovie, lastFrameInMovie, frameDelta, movieFormat
	if (V_Flag == 1)
		return 1
	endif
	
	string movieExtension
	StrSwitch(movieFormat)
		case "Quicktime":
			movieExtension = ".mov"
			break
		case "AVI":
			movieExtension = ".avi"
			break
		default:
			Abort "Unknown extension for movie format"
			break
	EndSwitch
	
	// remove the extension of the data file and replace it with .mov
	outputFilePath  = ParseFilePath(3, dataFilePath, ":", 0, 0)
	outputFilePath += movieExtension
	
	if (firstFrameInMovie < 0)
		Abort "Invalid first frame"
	elseif (lastFrameInMovie > nCCDFrames)
		Abort "The data contains less frames than requested"
	elseif (firstFrameInMovie >= lastFrameInMovie)
		Abort "Invalid frame range"
	endif
	
	Open /D /T=movieExtension refnum as outputFilePath
	if (strlen(S_fileName) == 0)
		return 1
	endif
	
	outputFilePath = S_fileName
	
	// remove a pre-exisiting movie viewer if it exists
	DoWindow /K MovieViewer
	
	// get a recreation macro for the graph with the movie
	// and parse it to get the macro for only the graph
	String recreationMacro = WinRecreation(baseWindowName, 0)
	recreationMacro = ReplaceString("\r\t", recreationMacro, ";")
	
	// don't get all of the macro, leave out the display command
	variable startOfMacro = strsearch(recreationMacro, ";Display", 0)
	startOfMacro = strsearch(recreationMacro, ";", startOfMacro + 1) + 1
	variable endOfMacro = strsearch(recreationMacro, "RenameWindow", startOfMacro) - 1
	
	string relevantMacro = recreationMacro[startOfMacro, endOfMacro]
	
	// parsing the recreation macro is surprisingly annoying. It turns out that when e.g.
	// particle tracks are displayed, a SetDataFolder call is inserted before the AppendImage
	// and all other appends are relative to this one. So try to kludge around that
	variable firstSetDataFolder = strsearch(recreationMacro, "SetDataFolder", 0)
	if ((firstSetDataFolder != -1) && (firstSetDataFolder < startOfMacro))
		string dataFolderSetTo = recreationMacro[strsearch(recreationMacro, "root:", firstSetDataFolder), strsearch(recreationMacro, ";Display", firstSetDataFolder) - 1]
		// even more annoying is the fact that the Display command will list a wave to append directly
		// so get that one out, too
		string displayCommand = recreationMacro[strsearch(recreationMacro, ";Display", firstSetDataFolder), strsearch(recreationMacro, ";Append", firstSetDataFolder) - 1]
		string waveToAppend
		SplitString /E=".[^ ]* (.*)$" displayCommand, waveToAppend
		
		// and generate the append command and prepend it to the macro
		string savDF = GetDataFolder(1)
		relevantMacro = "SetDataFolder " + dataFolderSetTo + "; AppendToGraph " + waveToAppend + "; " + relevantMacro
		relevantMacro = ReplaceString("SetDataFolder fldrSav0;", relevantMacro, "SetDataFolder " + savDF + ";")
	endif
		
	
	// we'll record the movie with the correct aspect size, so get hold of the image dimension
	wave image = windowDataFolder:M_CCDImage
	variable xSize = DimSize(image, 0)
	variable ySize = DimSize(image, 1)
	
	// display the graph
	DoWindow /K MovieViewer
	Display /N=MovieViewer /W=(67,128, 67 + xSize * ScreenResolution / 72, 128 + ySize * ScreenResolution / 72)
	
	// execute the recreation macro in a piecewise fashion
	// otherwise the string to execute can become too long
	for (i = 0; i < ItemsInList(relevantMacro); i += 1)
		string Dummy = StringFromList(i, relevantMacro)
		Execute StringFromList(i, relevantMacro)
	endfor
	
	// get rid of margins and axes
	ModifyGraph /W=MovieViewer margin=-1
	ModifyGraph /W=MovieViewer noLabel=2,axThick=0
	
	if (StringMatch(movieFormat, "AVI") == 1)
		NewMovie /A /Z /O /L /F=(fps) /I as outputFilePath
	else
		NewMovie /Z /O /L /F=(fps) /I as outputFilePath
	endif
	if (V_flag == -1)
		// user canceled
		DoWindow /K MovieViewer
		return 0
	endif
	if (V_flag != 0)
		error = GetErrMessage(V_flag, 3)
		DoWindow /K MovieViewer
		Abort error
	endif
	
	for (i = firstFrameInMovie; i < lastFrameInMovie; i += frameDelta)
		ChangeCurrentImage(i, baseWindowName)
		currentImage = i
		DoUpdate
		AddMovieFrame
	endfor
	
	CloseMovie
	DoWindow /K MovieViewer
	
	// restore the image we were originally looking at
	ChangeCurrentImage(originalFrame, baseWindowName)
End

Function SetupBatchLocalizationPanel()
	
	// check if there is a viewer panel open
	// this is required because we take the analysis settings
	// from the topmost panel
	string topWinName = GetNameOfTopInterfacePanel()
	if (strlen(topWinName) == 0)
		Abort "Please open a data file and set the desired processing options in the panel"
	endif
	string baseWindowName = GetBaseWindowName(topWinName)
	
	Make /O/N=(0, 2) /T root:Packages:Localizer:M_BatchProcessListWave
	Make /O/B/U/N=(0, 2, 2) root:Packages:Localizer:M_BatchProcessSelectionWave
	Make /O/W/U/N=(2, 3) root:Packages:Localizer:M_BatchProcessColorWave
	Make /O/N=2 /T root:Packages:Localizer:M_BatchProcessColumnHeadings
	wave /T M_BatchProcessColumnHeadings = root:Packages:Localizer:M_BatchProcessColumnHeadings
	wave M_BatchProcessColorWave = root:Packages:Localizer:M_BatchProcessColorWave
	wave M_BatchProcessSelectionWave = root:Packages:Localizer:M_BatchProcessSelectionWave
	
	M_BatchProcessColumnHeadings[0] = "CCD Files"
	M_BatchProcessColumnHeadings[1] = "Output position names"
	
	M_BatchProcessColorWave = 0
	M_BatchProcessColorWave[1][0] = 65535
	SetDimLabel 2,1, backColors, M_BatchProcessSelectionWave
	M_BatchProcessSelectionWave[][][1] = 0	// provide the default Igor colors
	
	DoWindow /F BatchProcess
	if (V_flag == 0)
		NewPanel /K=1 /N=BatchProcess /W=(238,189,792,491) as "Select the files to localize (double click to remove)"
		SetDrawLayer UserBack
		SetDrawEnv fsize= 11,fstyle= 2
		DrawText 9,20,"Note: all files will be processed with the settings in the top viewer panel"
		ListBox LBBatchListLocalization,pos={7,27},size={541,246}, proc=LBDeleteDoubleClickedRowProc, listWave=root:Packages:Localizer:M_BatchProcessListWave
		ListBox LBBatchListLocalization,frame=2,userColumnResize=1,titleWave=root:Packages:Localizer:M_BatchProcessColumnHeadings
		ListBox LBBatchListLocalization,mode=1,selWave=root:Packages:Localizer:M_BatchProcessSelectionWave
		ListBox LBBatchListLocalization,colorWave = root:Packages:Localizer:M_BatchProcessColorWave
		ListBox LBBatchListLocalization,userdata(ResizeControlsInfo)= A"!!,@C!!#=;!!#Cl5QF/pz!!#](Aon\"Qzzzzzzzzzzzzzz!!#o2B4uAezz"
		ListBox LBBatchListLocalization,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#u:Du]k<zzzzzzzzzzz"
		ListBox LBBatchListLocalization,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
		Button BTBatchStartAnalysis,pos={434,277},size={100,20},title="Start!",proc=BTStartBatchLocalization
		Button BTBatchStartAnalysis,userdata(ResizeControlsInfo)= A"!!,I?!!#BDJ,hpW!!#<Xz!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
		Button BTBatchStartAnalysis,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
		Button BTBatchStartAnalysis,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
		Button BTBatchAddLocalizationFiles,pos={320,276},size={100,20},title="Add files",proc=BTBatchAddDataFiles
		Button BTBatchAddLocalizationFiles,userdata(ResizeControlsInfo)= A"!!,H[!!#BD!!#@,!!#<Xz!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
		Button BTBatchAddLocalizationFiles,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
		Button BTBatchAddLocalizationFiles,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
		CheckBox CBBatchSaveExperiment,pos={6,279},size={175,14},title="Autosave experiment after each run"
		CheckBox CBBatchSaveExperiment,userdata(ResizeControlsInfo)= A"!!,@#!!#BEJ,hqi!!#;mz!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
		CheckBox CBBatchSaveExperiment,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
		CheckBox CBBatchSaveExperiment,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
		CheckBox CBBatchSaveExperiment,value= 1
		
		// the controls in this panel will need to know the name of the panel
		// with the settings. Encode this in the userdata
		SetWindow kwTopWin, userdata = baseWindowName
		
		// make the controls resize dynamically when the window is resized thanks to the <Resize Controls> package
		SetWindow BatchProcess,hook(ResizeControls)=ResizeControls#ResizeControlsHook
		SetWindow kwTopWin,userdata(ResizeControlsInfo)= A"!!*'\"z!!#CoJ,hs'zzzzzzzzzzzzzzzzzzzzz"
		SetWindow kwTopWin,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzzzzzzzzzzzzzzz"
		SetWindow kwTopWin,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzzzzzzzzz!!!"
		
		// if the window was opened previously then restore its size and position
		WC_WindowCoordinatesRestore("BatchProcess")
		SetWindow BatchProcess hook(WindowCoordsHook)=WC_WindowCoordinatesNamedHook
	endif
	
End

Function BTBatchAddDataFiles(ba) : ButtonControl
	STRUCT WMButtonAction &ba
	
	string baseFileName, extension, outputFilePath, macintoshFilePath
	variable nFiles, nNewFiles, i, offset
	
	switch( ba.eventCode )
		case 2: // mouse up
				// click code here
			wave /T M_BatchProcessListWave = root:Packages:Localizer:M_BatchProcessListWave
			wave M_BatchProcessSelectionWave = root:Packages:Localizer:M_BatchProcessSelectionWave
			wave M_BatchProcessColorWave = root:Packages:Localizer:M_BatchProcessColorWave
			
			Open /D/R /MULT=1/T="????" refNum
			if (strlen(S_fileName) == 0)
				// user canceled
				return 0
			endif
			
			offset = DimSize(M_BatchProcessListWave, 0)
			nNewFiles = ItemsInList(S_fileName, "\r") 
			nFiles = DimSize(M_BatchProcessListWave, 0) + nNewFiles
			Redimension /N=(nFiles, 2) M_BatchProcessListWave
			Redimension /N=(nFiles, 2, 2) M_BatchProcessSelectionWave
			for (i = 0; i < nNewFiles; i+=1)
				M_BatchProcessListWave[offset][0] = StringFromList(i, S_fileName, "\r")
				// the output data will be saved as igor waves
				baseFileName = ParseFilePath(3, StringFromList(i, S_fileName, "\r"), ":", 0, 0)
				//provide a valid Igor output name
				baseFileName = CleanupName(baseFileName, 0)
				M_BatchProcessListWave[offset][1] = baseFileName
				offset += 1
			endfor
			
			M_BatchProcessSelectionWave[][1][0] = 2	// be sure that the output names are editable
			
			break
	endswitch	
	
	return 0
End

Function BTStartBatchLocalization(ba) : ButtonControl
	STRUCT WMButtonAction &ba
	
	variable nFiles, nNewFiles, i, j, startTime
	variable autoSave
	variable invalidOutputNames = 0
	variable identicalNames = 0
	Struct FitData fitParams
	variable err
	
	switch( ba.eventCode )
		case 2: // mouse up
				// click code here
			// get the name of the panel with the settings
			string interfacePanelName = GetUserData("BatchProcess", "", "")
			string baseWindowName = GetBaseWindowName(interfacePanelName)
			DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
			
			wave /T M_BatchProcessListWave = root:Packages:Localizer:M_BatchProcessListWave
			wave M_BatchProcessSelectionWave = root:Packages:Localizer:M_BatchProcessSelectionWave
			wave M_BatchProcessColorWave = root:Packages:Localizer:M_BatchProcessColorWave
			
			Make /T/O/N=(0, 2) root:Packages:Localizer:M_BatchProcessErrors	// keep track of data files that had errors
			wave /T M_BatchProcessErrors = root:Packages:Localizer:M_BatchProcessErrors
			
			
			nFiles = DimSize(M_BatchProcessListWave, 0)
			
			// make sure that none of the cells are highlighted
			M_BatchProcessSelectionWave[][][1] = 0	// provide the default Igor colors
				
			// are any of the output file names invalid?
			for (i = 0; i < nFiles; i+=1)
				if ((StringMatch(M_BatchProcessListWave[i][1], CleanupName(M_BatchProcessListWave[i][1], 0)) != 1) || (strlen(M_BatchProcessListWave[i][1]) == 0))
					M_BatchProcessSelectionWave[i][1][1] = 1	// give the cell a red color
					invalidOutputNames = 1
				endif
			endfor
			
			if (invalidOutputNames == 1)
				Abort "Some of the output names are invalid, please correct these"
			endif
			
			// do any of the positions have identical names?
			for (i = 0; i < nFiles - 1; i+=1)
				for (j = i + 1; j < nFiles; j+=1)
					if (StringMatch(M_BatchProcessListWave[i][1], M_BatchProcessListWave[j][1]) == 1)
						M_BatchProcessSelectionWave[i][1][1] = 1// give the cell a red color
						M_BatchProcessSelectionWave[j][1][1] = 1
						identicalNames = 1
					endif
				endfor
			endfor
			
			if (identicalNames == 1)
				Abort "Some of the output names are identical, please correct these"
			endif
			
			// if we're here then we're all set to go
			ControlInfo /W=BatchProcess CBBatchSaveExperiment
			autoSave = V_Value
			
			// if the user selected to autosave the experiment, save it now
			// this avoids the situation where the experiment was not saved before and the 'SaveExperiment'
			// operation opens up a save dialog, but only after the first run has finished
			// and the user might have gone home
			if (autoSave != 0)
				SaveExperiment
			endif
			
			startTime = StopMSTimer(-2)
			string outputWaveName
			
			for (i = 0; i < nFiles; i+=1)
				ProgressWindow("file", i + 1, nFiles)
				// get all the fit options and parameters the user defined in the GUI
				GetGUIFitSettingsAndOptions(fitParams, baseWindowName)
				
				// set the data filepath to the requested file
				fitParams.CCDFilePath = M_BatchProcessListWave[i][0]
				
				// run the actual fit
				err = DoPALMFitting(fitParams, baseWindowName, outputWaveName, 0)
				if (err != 0)
					if (err == kUserAbort)
						// the user wants to abort the analysis
						Print "Batch fitting stopped due to user abort"
						KillWaves /Z M_BatchProcessListWave, M_BatchProcessSelectionWave, M_BatchProcessColorWave, M_BatchProcessErrors
						return 0
					endif
					
					// if we got here then it means that there was an error fitting the current data file
					// don't abort now, but remember the filePath that caused the error and report it to the user
					Redimension /N=(DimSize(M_BatchProcessErrors, 0) + 1, DimSize(M_BatchProcessErrors, 1)) M_BatchProcessErrors
					M_BatchProcessErrors[DimSize(M_BatchProcessErrors, 0) - 1][0] = M_BatchProcessListWave[i][0]
					M_BatchProcessErrors[DimSize(M_BatchProcessErrors, 0) - 1][1] = M_BatchProcessListWave[i][1]
					
					// move on to the next file
					continue
				endif
				
				// the output positions wave will have been created with a name that depends on the
				// filepath 
				wave POS_Out = windowDataFolder:$outputWaveName
				// copy the wave to the requested output positions name
				Duplicate /O POS_Out, root:'Localized Positions':$M_BatchProcessListWave[i][1]
				// and delete the original
				KillWaves /Z POS_Out
				
				// if the users selected to autosave the experiment, do so now
				if (autoSave != 0)
					SaveExperiment
				endif
			endfor
			
			// if we're here then the fitting is finished
			Printf "Total duration is %s\r", PrettyPrintTimeElapsed((StopMSTimer(-2) - startTime) / 1e6)
			
			// if there were any errors then report them now
			if (DimSize(M_BatchProcessErrors, 0) != 0)
				Print "WARNING: the following data file(s) reported errors and should not be trusted:"
				for (i = 0; i < DimSize(M_BatchProcessErrors, 0); i+=1)
					Printf "%d. %s,  %s %s\r", i + 1, M_BatchProcessErrors[i][0], "positions were to be written in", M_BatchProcessErrors[i][1]
				endfor
			endif
			
			
			// clean up
			KillWaves /Z M_BatchProcessListWave, M_BatchProcessSelectionWave, M_BatchProcessColorWave, M_BatchProcessErrors
			
			DoWindow /K BatchProcess
			
			break
	endswitch
	
	return 0
End

Function MergePositions()
	
	string savDF = GetDataFolder(1)
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:Localizer
	
	DFREF packageFolder = root:Packages:Localizer
	
	string listPositionsFunc = GetIndependentModuleName() + "#GetPossiblePositionsWaves()"
	
	
	Make /T/O/N=(0) packageFolder:candidatesListBoxWave
	Make /O/N=(0) packageFolder:candidatesListBoxSelWave
	
	DoWindow /F positionsWavesViewer
	if (V_flag == 0)
		NewPanel /K=1/W=(101,286,335,497) /N=positionsWavesViewer as "Select positions waves (double click to remove)"
		ListBox ListBoxCandidateWaves,pos={1,2},size={231,183},proc=LBDeleteDoubleClickedRowProc
		ListBox ListBoxCandidateWaves,listWave=packageFolder:candidatesListBoxWave,mode=4
		ListBox ListBoxCandidateWaves,selRow= 0,selWave=packageFolder:candidatesListBoxSelWave
		ListBox ListBoxCandidateWaves,userdata(ResizeControlsInfo)= A"!!,<7!!#7a!!#B!!!#AFz!!#](Aon\"Qzzzzzzzzzzzzzz!!#o2B4uAezz"
		ListBox ListBoxCandidateWaves,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#u:Du]k<zzzzzzzzzzz"
		ListBox ListBoxCandidateWaves,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
		PopupMenu PMAddPositionsWaves,pos={2,190},size={109,20},title="Add positions",proc=PMAddPositionsWavesProc
		PopupMenu PMAddPositionsWaves,mode=0,value= #listPositionsFunc
		PopupMenu PMAddPositionsWaves,userdata(ResizeControlsInfo)= A"!!,=b!!#AM!!#@<!!#<Xz!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
		PopupMenu PMAddPositionsWaves,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
		PopupMenu PMAddPositionsWaves,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
		Button BTCreateImageFromMultiplePos,pos={164,189},size={50,20},proc=BTMergePositionsProc,title="OK"
		Button BTCreateImageFromMultiplePos,userdata(ResizeControlsInfo)= A"!!,G4!!#AL!!#>V!!#<Xz!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
		Button BTCreateImageFromMultiplePos,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
		Button BTCreateImageFromMultiplePos,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
		SetWindow kwTopWin,hook(closeControl)=SetupImageViewer_IgorWaves_hook
		SetWindow kwTopWin,hook(ResizeControls)=ResizeControls#ResizeControlsHook
		SetWindow kwTopWin,userdata(ResizeControlsInfo)= A"!!*'\"z!!#B$!!#Abzzzzzzzzzzzzzzzzzzzzz"
		SetWindow kwTopWin,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzzzzzzzzzzzzzzz"
		SetWindow kwTopWin,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzzzzzzzzz!!!"
		
		// if the window was opened previously then restore its size and position
		WC_WindowCoordinatesRestore("positionsWavesViewer")
		SetWindow positionsWavesViewer hook(WindowCoordsHook)=WC_WindowCoordinatesNamedHook
	endif
End

Function PMAddPositionsWavesProc(pa) : PopupMenuControl
	STRUCT WMPopupAction &pa

	switch( pa.eventCode )
		case 2: // mouse up
			DFREF packageFolder = root:Packages:Localizer
			Variable popNum = pa.popNum
			String popStr = pa.popStr
			
			wave/T candidatesListBoxWave = packageFolder:candidatesListBoxWave
			wave candidatesListBoxSelWave = packageFolder:candidatesListBoxSelWave
			
			if (StringMatch(popStr, "* No Positions *") == 1)
				return 0
			endif
			
			wave /Z newPos = GetPositionsWaveReference(popStr)
			
			if (WaveExists(newPos) == 0)
				Abort "Internal error: the positions wave does not exist"
			endif
			
			Redimension /N=(DimSize(candidatesListBoxWave, 0) + 1) candidatesListBoxWave
			Redimension /N=(DimSize(candidatesListBoxSelWave, 0) + 1) candidatesListBoxSelWave
			
			candidatesListBoxWave[DimSize(candidatesListBoxWave, 0) - 1] = popStr
			candidatesListBoxSelWave[DimSize(candidatesListBoxSelWave, 0) - 1] = 0
			
			ControlUpdate /W=positionsWavesViewer ListBoxCandidateWaves
			break
	endswitch

	return 0
End


Function BTMergePositionsProc(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	switch( ba.eventCode )
		case 2: // mouse up
			// click code here
			
			DFREF positionsFolder = root:'Localized Positions'
			DFREF packageFolder = root:Packages:Localizer
			
			wave/T waveNames = root:Packages:Localizer:candidatesListBoxWave
			string outputName = ""
			variable nWaves = DimSize(waveNames, 0)
			variable i, j
			variable offset = 0
			variable nPoints, totalNumberOfPoints = 0
			variable lastFrameIndex = 0
			string waveNote
			string originalFilePaths
			string originalCalculationDates
			variable totalCalculationTime
			
			if (nWaves < 2)
				DoWindow /K positionsWavesViewer
				return 0
			endif
			
			outputName = GetNewPositionsWaveName("", positionsFolder, "Enter a name for the merged positions wave")
			if (strlen(outputName) == 0)	// the user canceled the dialog
				return -1
			endif
			
			for (i = 0; i < nWaves; i+=1)
				wave pos = GetPositionsWaveReference(waveNames[i])
				totalNumberOfPoints += DimSize(pos, 0)
			endfor
			
			// did the positions originate from the same localization algorithm?
			variable localizationMethod
			localizationMethod = NumberByKey("LOCALIZATION METHOD", note(pos))
			
			for (i = 1; i < nWaves; i+=1)
				wave pos = GetPositionsWaveReference(waveNames[i])
				if (localizationMethod != NumberByKey("LOCALIZATION METHOD", note(pos)))
					Abort "These positions were acquired using different localization algorithms"
				endif
			endfor 
			
			Make /D/O/N=(totalNumberOfPoints, DimSize(pos, 1)) positionsFolder:$outputName
			
			wave concatenatedPositions = positionsFolder:$outputName
			
			for (i = 0; i < nWaves; i+=1)
				if (i > 0)
					lastFrameIndex += pos[DimSize(pos, 0) - 1][0]
				endif
				wave pos = GetPositionsWaveReference(waveNames[i])
				nPoints = DimSize(pos, 0)
				for (j = 0; j <  nPoints; j+=1)
					concatenatedPositions[offset][] = pos[j][q]
					concatenatedPositions[offset][0] += lastFrameIndex
					offset += 1
				endfor
			endfor
			
			// set up an appropriate wave note that contains the calculation information
			waveNote = note(pos)
			originalFilePaths = StringByKey("ORIGINAL FILE PATH", waveNote)
			originalCalculationDates = StringByKey("CALCULATION DATE", waveNote)
			totalCalculationTime = NumberByKey("CALCULATION DURATION", waveNote)
			for (i = 1; i < nWaves; i+=1)
				originalFilePaths += " and "
				originalFilePaths += StringByKey("ORIGINAL FILE PATH", note(GetPositionsWaveReference(waveNames[i])))
				
				originalCalculationDates += " and "
				originalCalculationDates += StringByKey("CALCULATION DATE", note(GetPositionsWaveReference(waveNames[i])))
				
				totalCalculationTime += NumberByKey("CALCULATION DURATION", note(GetPositionsWaveReference(waveNames[i])))
			endfor
			
			waveNote = ReplaceStringByKey("ORIGINAL FILE PATH", waveNote, originalFilePaths)
			waveNote = ReplaceStringByKey("CALCULATION DATE", waveNote, originalCalculationDates)
			waveNote = ReplaceNumberByKey("CALCULATION DURATION", waveNote, totalCalculationTime)
			
			Note /K concatenatedPositions, waveNote
			
			DoWindow /K positionsWavesViewer
			break
	endswitch

	return 0
End

Function ConsolidateSameEmitters_menu()
	// present a graphical interface to ConsolidateSameEmitters
	string posName
	
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:Localizer
	NewDataFolder /O root:'Localized Positions'
	
	Variable /G root:Packages:Localizer:V_consolidateMaxDistance
	Variable /G root:Packages:Localizer:V_consolidateMaxBlinking
	Variable /G root:Packages:Localizer:V_consolidateMinTrajLength
	Variable /G root:Packages:Localizer:V_consolidateMaxTrajLength
	
	NVAR maxPositionDifference = root:Packages:Localizer:V_consolidateMaxDistance
	NVAR maxBlinking = root:Packages:Localizer:V_consolidateMaxBlinking
	NVAR consolidateMinTrajLength = root:Packages:Localizer:V_consolidateMinTrajLength
	NVAR consolidateMaxTrajLength = root:Packages:Localizer:V_consolidateMaxTrajLength
	
	maxPositionDifference = (maxPositionDifference == 0) ? 1.0 : maxPositionDifference
	consolidateMinTrajLength = (consolidateMinTrajLength == 0)?  -1 : consolidateMinTrajLength
	consolidateMaxTrajLength = (consolidateMaxTrajLength == 0) ? -1 : consolidateMaxTrajLength
	
	variable maxPositionDifferenceLocal = maxPositionDifference
	variable maxBlinkingLocal = maxBlinking
	variable MinTrajLengthLocal = consolidateMinTrajLength
	variable MaxTrajLengthLocal = consolidateMaxTrajLength
	
	Prompt posName, "Positions wave:", popup, GetPossiblePositionsWaves()
	Prompt maxPositionDifferenceLocal, "Maximum difference in position (in pixels)"
	Prompt maxBlinkingLocal, "Maximum allowed gap due to blinking (in frames)"
	Prompt MinTrajLengthLocal, "Min number of observations for localization error analysis:"
	Prompt MaxTrajLengthLocal, "Max number of observations for localization error analysis:"
	
	DoPrompt "Enter calculation parameters", posName, maxPositionDifferenceLocal, maxBlinkingLocal, MinTrajLengthLocal, MaxTrajLengthLocal
	if (V_flag == 1)	// cancel
		return 0
	endif
	
	maxBlinking = maxBlinkingLocal
	maxPositionDifference = maxPositionDifferenceLocal
	consolidateMinTrajLength = MinTrajLengthLocal
	consolidateMaxTrajLength = MaxTrajLengthLocal
	
	if (MinTrajLengthLocal == -1)
		MinTrajLengthLocal = 2
	endif
	if (MaxTrajLengthLocal == -1)
		MaxTrajLengthLocal = inf
	endif
	
	if ((MinTrajLengthLocal <= 1) || (MaxTrajLengthLocal <= 1) || (MinTrajLengthLocal > MaxTrajLengthLocal))
		Abort "Invalid input for the min and/or max trajectory length for error analysis"
	endif
	
	wave /Z pos = GetPositionsWaveReference(posName)
	if (WaveExists(pos) == 0)
		Abort "The selected wave doesn't seem to exist!"
	endif
	
	// get a wave name for the output positions
	string suggestedOutputName = CleanupName(NameOfWave(pos) + "_cons", 0)
	string outputName = GetNewPositionsWaveName(suggestedOutputName, root:'Localized Positions', "Enter a name for the output wave")
	if (strlen(outputName) == 0)
		return 0
	endif
	
	// do the actual calculation
	ConsolidateSameEmitters(pos, "root:'Localized Positions':" + outputName, maxPositionDifference, maxBlinking, minLocalizationsForError = consolidateMinTrajLength, maxLocalizationsForError = MaxTrajLengthLocal)
	
End

Function ConsolidateSameEmitters(positions, posOutputName, maxPositionDifference, maxBlinking, [ minLocalizationsForError, maxLocalizationsForError])
	wave positions
	string posOutputName
	variable maxPositionDifference, maxBlinking
	variable minLocalizationsForError, maxLocalizationsForError
	
	Assert(WaveExists(positions))
	Assert(maxPositionDifference >= 0)
	Assert(maxBlinking >= 0)
	
	// Go over all the positions in the positions wave and look for emitters that show up in
	// different frames but might actually be the same spot. The two emitters have to be close
	// enough to be counted as the same molecule (maxPositionDifference). To allow for blinking
	// a molecule can disappear from the images for a number of frames up to maxBlinking
	// and still be accepted as the same emitter
	
	// the algorithm works as follows: every new position is temporarily added to a 'waiting list'
	// that contains the index into positions where it is listed, its x and y position, and a counter
	// that is initially set to maxBlinking
	
	// when moving on to a new frame every position on the waiting list is checked to see if a point
	// within a distance of maxPositionDifference is present in the new frame. If so then the number of
	// frames associated with the position on the waiting list is incremented by one, the counter is reset
	// to maxBlinking, and the matching point is deleted from the positions wave.
	// If not then the counter is decremented. If the counter becomes negative then the point is deleted .
	
	// the format of the waiting list is index of first appearance / x position / y position / counter
	
	// in its original form the algorithm actively deleted points that were the same as a previous emitter
	// during the processing. But this introduces the overhead of copying the wave elements each time.
	// therefore in its current form the algorithm does not delete points directly, but works on a duplicate
	// of the positions where the entries to be deleted are marked by a negative frame number
	
	variable minLocForError, maxLocForError
	if (ParamIsDefault(minLocalizationsForError))
		minLocForError = 2
	else
		minLocForError = minLocalizationsForError
	endif
	if (ParamIsDefault(maxLocalizationsForError))
		maxLocForError = inf
	else
		maxLocForError = maxLocalizationsForError
	endif
	if (minLocForError > maxLocForError)
		Abort "Invalid limits for localization error"
	endif
	
	variable nFramesPresentColumn, xCol, yCol, zCol, intensityColumn, xUncertaintyCol, yUncertaintyCol, zUncertaintyCol
	GetColumnForNFramesPresent(positions, nFramesPresentColumn)	// handle different types of positions
	GetColumnsForEmitterPositions(positions, xCol, yCol, zCol)
	GetColumnForIntegratedIntensity(positions, intensityColumn)
	GetColumnsForLocalizationError(positions, xUncertaintyCol, yUncertaintyCol, zUncertaintyCol)
	
	variable xSize, ySize, pixelSize
	GetCCDDimensionsFromPositions(positions, xSize, ySize, pixelSize)
	
	// check if these positions have not been consolidated before
	// if so then the results are unpredictable
	ImageTransform /G=(nFramesPresentColumn) sumCol, positions
	if (V_value != DimSize(positions, 0))
		DoAlert 1, "These positions look as if they have already been combined, or were created using an old version of this program. In the latter case, would you like me to correct this and continue?"
		if (V_flag == 1)	// yes
			positions[][nFramesPresentColumn] = 1
		else
			return 0
		endif
	endif
	
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:Localizer

	// group the emitters
	wave /WAVE W_GroupedEmitters = GroupEmitters(positions, CombinedPosEstimator_Mean, maxPositionDifference, maxBlinking)
	variable nGroups = DimSize(W_GroupedEmitters, 0)
	variable i, localizationStdDev, trackLength
	
	// allocate an output positions wave with the right number of entries
	Make /O/N=(nGroups, DimSize(positions, 1)) /D $posOutputName
	wave consolidatedPositions = $posOutputName
	
	// make room for some other statistics
	Make /N=(nGroups) /D /FREE W_LocalizationErrors
	Make /N=(nGroups) /D /FREE W_TrackLength
	
	for (i = 0; i < nGroups; i+=1)
		ProgressWindow("Combining", i, nGroups)
		wave currentEmitters = W_GroupedEmitters[i]
		wave combinedEmitters = CombineEmittersIntoPosition(currentEmitters, localizationStdDev, trackLength)
		consolidatedPositions[i][] = combinedEmitters[0][q]
		W_LocalizationErrors[i] = localizationStdDev
		W_TrackLength[i] = trackLength
	endfor
	
	Make /N=0 /D /O root:Packages:Localizer:W_LocalizationErrorHist
	Make /N=0 /D /O root:Packages:Localizer:W_TrackLengthHist
	Histogram /B=4 W_LocalizationErrors, root:Packages:Localizer:W_LocalizationErrorHist
	Histogram /B=4 W_TrackLength, root:Packages:Localizer:W_TrackLengthHist
	
	// set the wave scaling of the statistics to match the pixel size if it was specified
	GetCCDDimensionsFromPositions(positions, xSize, ySize, pixelSize)
	if (pixelSize != 0)
		SetScale /P x, 0, DimDelta(root:Packages:Localizer:W_LocalizationErrorHist, 0) * pixelSize, root:Packages:Localizer:W_LocalizationErrorHist
		SetScale /P x, 0, DimDelta(root:Packages:Localizer:W_TrackLengthHist, 0) * pixelSize, root:Packages:Localizer:W_TrackLengthHist
	endif
	
	// add info on the calculation params to the wave note
	string waveNote = Note(positions) + "CONSOLIDATE MAX SHIFT:" + num2str(maxPositionDifference) + ";" + "CONSOLIDATE MAX BLINKING:" + num2str(maxBlinking) + ";"
	Note /K consolidatedPositions, waveNote

	// perform a quality check
	ImageTransform /G=(nFramesPresentColumn) sumCol, consolidatedPositions
	if (DimSize(positions, 0) != V_Value)
		Abort "The calculation appears to have gone wrong, the results should not be trusted"
	endif
	
	// display the histograms
	DoWindow /F LocalizationErrorViewer
	if (V_flag != 1)
		Display /K=1 /N=LocalizationErrorViewer root:Packages:Localizer:W_LocalizationErrorHist
		if (pixelSize == 0)
			Label Bottom, "Estimated localization error (pixels)"
		else
			Label Bottom, "Estimated localization error (nm)"
		endif
		Label Left, "Occurrence"
	endif
	
	DoWindow /F TrackLengthViewer
	if (V_flag != 1)
		Display /K=1 /N=TrackLengthViewer root:Packages:Localizer:W_TrackLengthHist
		Label Bottom, "Track length (frames)"
		Label Left, "Occurrence"
		AutoPositionWindow /R=LocalizationErrorViewer TrackLengthViewer
	endif
End

Function ParticleTracking(positions, posOutputName, maxPositionJump, maxBlinking, [ minTrackLength, maxTrackLength])
	wave positions
	string posOutputName
	variable maxPositionJump, maxBlinking
	variable minTrackLength, maxTrackLength
	
	Assert(WaveExists(positions))
	Assert(maxPositionJump >= 0)
	Assert(maxBlinking >= 0)
	
	// perform single particle tracking, with possible limitation on the length of an
	// acceptable track
	// This output of this function is a wave containing wave references
	// each wave reference contains a single track
	// the tracks are sorted into ascending frame of first occurrence
	
	// TODO: 
	// 1. the analysis ignores the possibility that more than one positions may be localized
	// within maxPositionJump. In this case, the emitter that the algorithm will choose is currently
	// undefined
	// 2. the analysis ignores the possibility that the particle may diffuse while it is not emitting
	
	variable localMinTrackLength, localMaxTrackLength
	if (ParamIsDefault(minTrackLength))
		localMinTrackLength = 2
	else
		localMinTrackLength = minTrackLength
	endif
	if (ParamIsDefault(maxTrackLength))
		localMaxTrackLength = inf
	else
		localMaxTrackLength = maxTrackLength
	endif
	if (localMinTrackLength > localMaxTrackLength)
		Abort "Invalid limits for localization error"
	endif
	
	variable nFramesPresentColumn, xCol, yCol, zCol, intensityColumn, xUncertaintyCol, yUncertaintyCol, zUncertaintyCol
	GetColumnForNFramesPresent(positions, nFramesPresentColumn)	// handle different types of positions
	GetColumnsForEmitterPositions(positions, xCol, yCol, zCol)
	GetColumnForIntegratedIntensity(positions, intensityColumn)
	GetColumnsForLocalizationError(positions, xUncertaintyCol, yUncertaintyCol, zUncertaintyCol)
	
	// check if these positions have not been consolidated before
	// if so then there's no meaning in performing this
	ImageTransform /G=(nFramesPresentColumn) sumCol, positions
	if (V_value != DimSize(positions, 0))
		Abort "These positions have been consolidated. There is no point in continuing"
	endif
	
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:Localizer

	// group the emitters
	wave /WAVE W_GroupedEmitters = GroupEmitters(positions, CombinedPosEstimator_LastPos, maxPositionJump, maxBlinking)
	
	// filter the emitter tracks to make sure that they are within the acceptable bounds
	variable i
	variable nGroups = DimSize(W_GroupedEmitters, 0)
	variable nGroupsRemaining = nGroups
	// first see how many emitters will remain
	for (i = 0; i < nGroups; i+=1)
		wave currentEmitter = W_GroupedEmitters[i]
		if ((DimSize(currentEmitter, 0) < localMinTrackLength) || (DimSize(currentEmitter, 0) > localMaxTrackLength))
			nGroupsRemaining -= 1
		endif
	endfor
	
	Make /FREE /WAVE /N=(nGroupsRemaining) W_FilteredEmitters
	variable offset
	for (i = 0; i < nGroups; i+=1)
		wave currentEmitter = W_GroupedEmitters[i]
		if ((DimSize(currentEmitter, 0) < localMinTrackLength) || (DimSize(currentEmitter, 0) > localMaxTrackLength))
			continue
		endif
		W_FilteredEmitters[offset] = W_GroupedEmitters[i]
		offset += 1
	endfor
	
	// no more modification is needed, just make a copy of the positions and add a wavenote
	string waveNote = Note(positions) + "PARTICLETRACKING MAX SHIFT:" + num2str(maxPositionJump) + ";" + "PARTICLETRACKING MAX BLINKING:" + num2str(maxBlinking) + ";"
	Note /K W_FilteredEmitters, waveNote
	Duplicate /WAVE /O W_FilteredEmitters, $posOutputName
End

Function /WAVE GroupEmitters(positions, PositionEstimator, maxPositionDifference, maxBlinking)
	wave positions
	FUNCREF CombinedPosEstimator_Prototype PositionEstimator
	variable maxPositionDifference, maxBlinking
	
	// given a list of positions, group these into subsets that combine positions occurring in subsequent
	// frames and that are not shifted in position by more than maxPositionDifference
	// also allow the emitters to blink for maxBlinking number of frames before being dropped from consideration
	
	variable nFramesPresentColumn, xCol, yCol, zCol, intensityColumn, xUncertaintyCol, yUncertaintyCol, zUncertaintyCol
	GetColumnForNFramesPresent(positions, nFramesPresentColumn)	// handle different types of positions
	GetColumnsForEmitterPositions(positions, xCol, yCol, zCol)
	GetColumnForIntegratedIntensity(positions, intensityColumn)
	GetColumnsForLocalizationError(positions, xUncertaintyCol, yUncertaintyCol, zUncertaintyCol)
	
	// maintain a copy of the string containing the format specification for these positions
	// each grouped set of emitters needs this string
	string positionsFormatStr = "LOCALIZATION METHOD:" + StringByKey("LOCALIZATION METHOD", Note(positions)) + ";"
	
	// Make a copy of the positions
	Duplicate /FREE /O positions, M_TempConsolidatedPositions
	
	// set up some waves that will provide temporary storage
	Make /N=0 /WAVE /FREE W_WaitingList
	// M_WaitingList is a wave containing wave references. Each of these references points to
	// a 2D wave, that contains those localizations that are thought to arise from the same particle.
	// each entry is a full localization, so contains the same number of columns as the positions wave
	Make /N=0 /I /FREE W_RemainingFrames
	// contains the number of frames this emiter can be off while still being counted as active
	
	Make /N=0 /FREE /D W_ClosestEmitterDistances
	Make /N=0 /I /FREE /D W_ClosestEmitterIndex
	// these waves will be used internally by the algorithm
	// to allow the globally closest emitter - previous location
	// to be found
	// W_ClosestEmitterDistances will contain the distance from the corresponding entry in the waiting
	// list to the closest emitter in the current frame. If there is no such emitter then it will be set to NaN
	// W_ClosestEmitterIndex contains the index of the closest emitter in this frame to the corresponding
	// emitter in the waiting list. If there is no such emitter then it is set to -1
	
	Make /N=0 /WAVE /FREE W_GroupedEmitters
	// the result of this function
	// each entry is a reference to a wave containing one or more positions
	// that are considered to belong together
	// every position shows up somewhere, and only once, in this wave
	// though it may be present in a one-row wave or together with a bunch of others
	
	variable n, err, i
	variable startingFrame = positions[0][0]	// warning: magic numbers
	variable endingFrame = positions[DimSize(positions,0) - 1][0]
	variable nFramesInCalculation = endingFrame - startingFrame + 1
	
	variable lastX, lastY, nPointsInThisTrajectory, localizationStdDev, trackLength, minIndex
	variable nearestX, nearestY, index, distance, offset, nPointsToBeDeleted = 0
	variable currentXPos, currentYPos, currentNFrames, currentXUncertainty, currentYUncertainty
	variable previouslyMatchedEmitter
	string trackWaveName
	
	for (n = startingFrame; n <= endingFrame; n+=1)
		// the calculation can be slow occasionally, so add a progress bar
		// but only if there is a sufficiently large number of frames
		if (nFramesInCalculation > 10)
			ProgressWindow("Grouping", n - startingFrame, endingFrame - startingFrame + 1)
		endif
		
		// get all emitters localized in this frame
		wave /WAVE M_Extracted = ExtractPositionsInFrame(positions, n)
		wave M_emittersInCurrentFrame = M_Extracted[0]
		wave W_emittersInCurrentFrameIndices = M_Extracted[1]
		
		// for every emitter in the waiting list we need to check if there is a matching emitter in the 
		// current frame. One of the side issues with that is that more than one of the emitters in the
		// waiting list can match the same one in the current frame. So iterate over this procedure
		// in such a way that the closest emitter is taken at each point
		Redimension /N=(DimSize(W_WaitingList, 0)) W_ClosestEmitterDistances, W_ClosestEmitterIndex
		W_ClosestEmitterDistances = inf
		W_ClosestEmitterIndex = -1
		previouslyMatchedEmitter = NaN
		do
			
			for (i = 0; i < DimSize(W_WaitingList, 0); i+=1)
				if (NumType(W_ClosestEmitterDistances[i]) == 2)
					// every emitter that has been matched already, or that has no corresponding match,
					// will have its entry in W_ClosestEmitterDistances set to NaN
					// so these points need to be skipped
					continue
				endif
				
				if ((W_ClosestEmitterIndex[i] != previouslyMatchedEmitter) && (NumType(previouslyMatchedEmitter) != 2))
					// in all iterations of the do-while loop following the first there is no need to recalculate any of the matching
					// emitters, except when this emitter matched to the same one as the one matched in the previous iteration
					
					// however, if the matching index of the current emitter is larger than the previously matched one, then we need to
					// decrement it by one. This compensates for the deletion of the emitter from the extracted positions list upon a
					// successful match
					if (W_ClosestEmitterIndex[i] > previouslyMatchedEmitter)
						W_ClosestEmitterIndex[i] -= 1
					endif
					
					continue
				endif
				
				// get the point in the current frame that is the closest to the last observation of this emitter
				// in the waiting list
				wave currentEmitter = W_WaitingList[i]
				
				// the measure of the last position of the emitter depends on what we want to do with the results
				// so it is parameterized using PositionEstimator. This function takes care of filling in
				// lastX and lastY
				PositionEstimator(currentEmitter, xCol, yCol, lastX, lastY)
				
				err = returnPointNearestFitPositions(M_emittersInCurrentFrame, lastX, lastY, nearestX, nearestY, index)
				if (err != 0)
					// some error looking for the nearest point
					// there are probably no more emitters in M_emittersInCurrentFrame
					W_ClosestEmitterDistances[i] = NaN
					W_ClosestEmitterIndex[i] = -1
					continue
				endif
				
				distance = sqrt((nearestX - lastX)^2 + (nearestY - lastY)^2)
				
				if (distance > maxPositionDifference)
					// the point on the waiting list is not present in the current frame
					W_ClosestEmitterDistances[i] = NaN
					W_ClosestEmitterIndex[i] = -1
				
				else // the point on the waiting list is also present in this frame
					// add the information to the intermediate waves
					W_ClosestEmitterDistances[i] = distance
					W_ClosestEmitterIndex[i] = index
				endif
			endfor
			
			// W_ClosestEmitterDistances now contains all distances between the waiting list and the closest matching positions
			// select the minimum distance, i.e. the closest possible match, and evaluate that
			// leave the other matching emitters to subsequent iterations of the do-while loop
			// unfortunately WaveStats throws an error if there are no points in W_ClosestEmitterDistances
			// (which occurs when the waiting list is empty)
			if (DimSize(W_WaitingList, 0) == 0)
				minIndex = -1
			else
				WaveStats /Q /M=1 W_ClosestEmitterDistances
				minIndex = V_minLoc
			endif
			if (minIndex == -1)
				// if W_ClosestEmitterDistances only contains NaN then WaveStats sets V_minLoc to -1
				// time to move on to the next frame
				break
			endif
			
			// if we're here then we have a new matching point
			// add it to the output
			wave currentEmitter = W_WaitingList[minIndex]
			nPointsInThisTrajectory = DimSize(currentEmitter, 0)
			index = W_ClosestEmitterIndex[minIndex]
			Assert(index != -1)
					
			Redimension /N=(nPointsInThisTrajectory + 1, -1) currentEmitter
			currentEmitter[nPointsInThisTrajectory][] = M_emittersInCurrentFrame[index][q]
			W_ClosestEmitterDistances[minIndex] = NaN
			previouslyMatchedEmitter = index
					
			// delete the matching point in the frame from the extracted positions list
			DeletePoints index, 1, M_emittersInCurrentFrame
			DeletePoints index, 1, W_emittersInCurrentFrameIndices
		while (1)
		
		// W_ClosestEmitterDistances must now contain only NaN's
		if (DimSize(W_ClosestEmitterDistances, 0) > 0)
			WaveStats /M=1 /Q W_ClosestEmitterDistances
			Assert(V_numNaNs == DimSize(W_ClosestEmitterDistances, 0))
		endif
		
		// update the blinking counters
		// for this the algorithm needs to know whether the emitters were successfully matched or not
		// this cannot be checked using W_ClosestEmitterDistances since it's NaN everywhere
		// instead make use of W_ClosestEmitterIndex: if it is -1 then there was no successful match
		for (i = 0; i < DimSize(W_WaitingList, 0); i+=1)
			if (W_ClosestEmitterIndex[i] == -1)
				W_RemainingFrames[i] -= 1
			else
				W_RemainingFrames[i] = maxBlinking
			endif
		endfor
		
		// any emitter in the waiting list that has a negative value in W_RemainingFrames is finished
		// the corresponding entry in currentEmitter represents a finished group of positions
		// add it as an entry to W_GroupedEmitters
		for (i = 0; i < DimSize(W_WaitingList, 0); i+=1)
			if (W_RemainingFrames[i] < 0)
				wave currentEmitter = W_WaitingList[i]
				Redimension /N=(DimSize(W_GroupedEmitters, 0) + 1) W_GroupedEmitters
				W_GroupedEmitters[DimSize(W_GroupedEmitters, 0) - 1] = currentEmitter
				DeletePoints i, 1, W_WaitingList
				DeletePoints i, 1, W_RemainingFrames
				i -= 1
			endif
		endfor
		
		// any emitters that remain in the current frame are new, add them to the waiting list
		offset = DimSize(W_WaitingList, 0)
		Redimension /N=(offset + DimSize(M_emittersInCurrentFrame, 0)) W_WaitingList
		Redimension /N=(offset + DimSize(M_emittersInCurrentFrame, 0)) W_RemainingFrames
		
		for (i = 0; i < Dimsize(M_emittersInCurrentFrame, 0); i+=1)
			Make /O/D/FREE /N=(1, DimSize(positions, 1)) M_NewEmitter
			// downstream routines (such as when combining these emitters for consolidation
			// must know what positions format this wave is in
			Note M_NewEmitter, positionsFormatStr
			M_NewEmitter = positions[W_emittersInCurrentFrameIndices[i]][q]
			W_WaitingList[i + offset] = M_NewEmitter
			W_RemainingFrames[offset + i] = maxBlinking
		endfor
		
	endfor
	
	// if we're here than we've looped over all the frames in the movie
	// it's possible that some emitters were 'on' throughout the movie
	// in that case they are still in W_WaitingList and need to be added
	// without regard to their blinking allowance
	for (i = 0; i < DimSize(W_WaitingList, 0); i+=1)
		wave currentEmitter = W_WaitingList[i]
		Redimension /N=(DimSize(W_GroupedEmitters, 0) + 1) W_GroupedEmitters
		W_GroupedEmitters[DimSize(W_GroupedEmitters, 0) - 1] = currentEmitter
		DeletePoints i, 1, W_WaitingList
		DeletePoints i, 1, W_RemainingFrames
		i -= 1
	endfor
	
	
	// no more emitters may be in the waiting list now
	Assert(DimSize(W_WaitingList, 0) == 0)
	
	// sort the grouped emitters such that the first frame they occur in
	// is in order
	// this is somewhat involved
	variable nEntries = DimSize(W_GroupedEmitters, 0)
	Make /O/N=(nEntries) /I/U /FREE W_FirstFrames
	Make /O/N=(nEntries) /I/U /FREE W_Indices
	W_Indices = p
	for (i = 0; i < nEntries; i+=1)
		wave currentEmitter = W_GroupedEmitters[i]
		W_FirstFrames[i] = currentEmitter[0][0]
	endfor
	Sort W_FirstFrames, W_Indices
	Duplicate /O/FREE /WAVE W_GroupedEmitters, W_TempGroupedEmitters
	W_GroupedEmitters = W_TempGroupedEmitters[W_Indices[p]]
	
	// perform a quality check on the algorithm
	// the total number of positions must not have been changed
	variable nGroupedPositions = 0;
	for (i = 0; i < nEntries; i+=1)
		wave currentEmitter = W_GroupedEmitters[i]
		nGroupedPositions += DimSize(currentEmitter, 0)
	endfor
	
	if (DimSize(positions, 0) != nGroupedPositions)
		Abort "The calculation appears to have gone wrong, the results should not be trusted"
	endif
	
	return W_GroupedEmitters
End

Function CombinedPosEstimator_Prototype(combinedPositions, xCol, yCol, xPos, yPos)
	wave combinedPositions
	variable xCol, yCol
	variable &xPos
	variable &yPos
	// to decide whether emitters belong together, we need some kind of metric
	// for the actual position of an emitter already fitted over several frames
	// this metric depends on what we want to do: for simple consolidation it should
	// probably be the average of all previous positions,
	// while for particle tracking it should really be the last observed positions
	
	// the current solution is to parametrize these different behaviors by providing different functions that
	// can then be passed to GroupEmitters
	// this is the prototype, which should not be called
	Abort "CombinedPosEstimator_Prototype should not be called"
End

Function CombinedPosEstimator_Mean(combinedPositions, xCol, yCol, xPos, yPos)
	wave combinedPositions
	variable xCol, yCol
	variable &xPos
	variable &yPos
	// estimate the actual position of the emitter as the mean of all observations
	variable nPositions = DimSize(combinedPositions, 0)
	
	ImageTransform /G=(xCol) sumCol, combinedPositions
	xPos = V_Value / nPositions
	ImageTransform /G=(yCol) sumCol, combinedPositions
	yPos = V_Value / nPositions
End

Function CombinedPosEstimator_LastPos(combinedPositions, xCol, yCol, xPos, yPos)
	wave combinedPositions
	variable xCol, yCol
	variable &xPos
	variable &yPos
	// estimate the actual position of the emitter as the mean of all observations
	variable nPositions = DimSize(combinedPositions, 0)
	
	xPos = combinedPositions[nPositions - 1][xCol]
	yPos = combinedPositions[nPositions - 1][yCol]
End

Function CreateRegistrationMap_Menu()
	
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:Localizer
	DFREF packageFolder = root:Packages:Localizer
	
	variable /G packageFolder:V_RegistrationMaxJump
	NVAR registrationMaxJump = packageFolder:V_RegistrationMaxJump
	
	if (registrationMaxJump == 0)
		registrationMaxJump = 5
	endif
	
	DoWindow RegistrationMapPanel
	if (V_flag != 0)
		DoWindow /K RegistrationMapPanel
	endif
	
	string possiblePositionsWaves = GetSavedPositionsWaves()
	variable nPositionsWaves = ItemsInList(possiblePositionsWaves)
	variable i
	
	Make /O/N=(nPositionsWaves)/T packageFolder:W_PossiblePosForRegistration1
	Make /O/N=(nPositionsWaves)/T packageFolder:W_PossiblePosForRegistration2
	wave /T W_PossiblePosForRegistration1 = packageFolder:W_PossiblePosForRegistration1
	wave /T W_PossiblePosForRegistration2 = packageFolder:W_PossiblePosForRegistration2
	for (i = 0; i < nPositionsWaves; i+=1)
		W_PossiblePosForRegistration1[i] = StringFromList(i, possiblePositionsWaves)
		W_PossiblePosForRegistration2[i] = StringFromList(i, possiblePositionsWaves)
	endfor
	Make /O/N=(0,2)/T packageFolder:M_RegistrationCombinations
	wave /T M_RegistrationCombinations = packageFolder:M_RegistrationCombinations
	Make /O/T=2 packageFolder:W_RegistrationTitleWave = {"Channel 1", "Channel 2"}
	wave /T W_RegistrationTitleWave = packageFolder:W_RegistrationTitleWave
	
	DoWindow /F RegistrationMapPanel
	if (V_flag != 1)
		NewPanel /K=1/N=RegistrationMapPanel /W=(225,102,1088,370) as "Create registration map"
		SetDrawLayer UserBack
		DrawText 8,24,"Channel 1 (e.g. blue, cyan, green)"
		DrawText 218,24,"Channel 2 (e.g. yellow, orange, red)"
		DrawText 459,24,"Calibration pairs"
		ListBox LBChannel1PositionsWaves,pos={6,26},size={199,176},mode= 2,selRow= 0,listWave=W_PossiblePosForRegistration1
		ListBox LBChannel2PositionsWaves,pos={216,28},size={193,175},mode= 2,selRow= 0,listWave=W_PossiblePosForRegistration1
		ListBox LBCalibrationPairs,pos={458,28},size={396,176},mode= 1,selRow= 0,listWave=M_RegistrationCombinations,titleWave=W_RegistrationTitleWave
		Button BTCreateRegistrationMap,pos={776,236},size={75,20},title="Do it",proc=BTCreateRegistrationMap
		Button BTAddPair,pos={414,93},size={40,20},title=">>",proc=BTAddRegistrationPair
		Button BTRemovePair,pos={414,126},size={40,20},title="<<",proc=BTRemoveRegistrationPair
		TitleBox TBErrorString,pos={464,126},size={50,20}
		TitleBox TBErrorString,pos={9,219},size={146,20},title="",fColor=(65535,0,0)
		SetVariable SVMaxRegistrationJump,pos={559,214},size={296,15},bodyWidth=100,title="Max difference in position between channels",value=registrationMaxJump
		
		string error = CheckRegistrationPanelForError()
		TitleBox TBErrorString win=RegistrationMapPanel,title=error
	endif
End

Function BTAddRegistrationPair(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	switch( ba.eventCode )
		case 2: // mouse up
			// click code here
			DFREF packageFolder = root:Packages:Localizer
			
			wave /T W_PossiblePosForRegistration1 = packageFolder:W_PossiblePosForRegistration1
			wave /T W_PossiblePosForRegistration2 = packageFolder:W_PossiblePosForRegistration2
			wave /T M_RegistrationCombinations = packageFolder:M_RegistrationCombinations
			
			variable selectedRowInChannel1
			variable selectedRowInChannel2
			
			ControlInfo /W=RegistrationMapPanel LBChannel1PositionsWaves
			selectedRowInChannel1 = V_Value
			ControlInfo /W=RegistrationMapPanel LBChannel2PositionsWaves
			selectedRowInChannel2 = V_Value
			
			if ((selectedRowInChannel1 == -1) || (selectedRowInChannel2 == -1) || (DimSize(W_PossiblePosForRegistration1, 0) == 0) || (DimSize(W_PossiblePosForRegistration2, 0) == 0))
				return 1
			endif
			
			string posName1 = W_PossiblePosForRegistration1[selectedRowInChannel1]
			string posName2 = W_PossiblePosForRegistration2[selectedRowInChannel2]
			Redimension /N=(DimSize(M_RegistrationCombinations, 0) + 1, -1) M_RegistrationCombinations
			M_RegistrationCombinations[DimSize(M_RegistrationCombinations, 0) - 1][0] = posName1
			M_RegistrationCombinations[DimSize(M_RegistrationCombinations, 0) - 1][1] = posName2
			
			DeletePoints selectedRowInChannel1, 1, W_PossiblePosForRegistration1
			DeletePoints selectedRowInChannel2, 1, W_PossiblePosForRegistration2
			
			// check for error
			string error = CheckRegistrationPanelForError()
			TitleBox TBErrorString win=RegistrationMapPanel,title=error
			if (strlen(error) > 0)
				Button BTCreateRegistrationMap win=RegistrationMapPanel, disable=2
			else
				Button BTCreateRegistrationMap win=RegistrationMapPanel, disable=0
			endif
			return 1
			break
		case -1: // control being killed
			break
	endswitch

	return 0
End

Function BTRemoveRegistrationPair(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	switch( ba.eventCode )
		case 2: // mouse up
			// click code here
			DFREF packageFolder = root:Packages:Localizer
			
			wave /T W_PossiblePosForRegistration1 = packageFolder:W_PossiblePosForRegistration1
			wave /T W_PossiblePosForRegistration2 = packageFolder:W_PossiblePosForRegistration2
			wave /T M_RegistrationCombinations = packageFolder:M_RegistrationCombinations
			
			variable selectedCombinationRow
			ControlInfo /W=RegistrationMapPanel LBCalibrationPairs
			selectedCombinationRow = V_Value
			
			if ((selectedCombinationRow == -1) || (DimSize(M_RegistrationCombinations, 0) == 0))
				return 1
			endif
			
			string combinationPosName1 = M_RegistrationCombinations[selectedCombinationRow][0]
			string combinationPosName2 = M_RegistrationCombinations[selectedCombinationRow][1]
			
			Redimension /N=(DimSize(W_PossiblePosForRegistration1, 0) + 1, -1) W_PossiblePosForRegistration1, W_PossiblePosForRegistration2
			W_PossiblePosForRegistration1[DimSize(W_PossiblePosForRegistration1, 0) - 1] = combinationPosName1
			W_PossiblePosForRegistration2[DimSize(W_PossiblePosForRegistration2, 0) - 1] = combinationPosName2
			
			DeletePoints /M=0 selectedCombinationRow, 1, M_RegistrationCombinations
			Redimension /N=(-1, 2) M_RegistrationCombinations	// avoids DeletePoints setting the wave to 1D
			
			// check for error
			string error = CheckRegistrationPanelForError()
			TitleBox TBErrorString win=RegistrationMapPanel,title=error
			if (strlen(error) > 0)
				Button BTCreateRegistrationMap win=RegistrationMapPanel, disable=2
			else
				Button BTCreateRegistrationMap win=RegistrationMapPanel, disable=0
			endif
			
			return 1
			break
		case -1: // control being killed
			break
	endswitch

	return 0
End

Function /S CheckRegistrationPanelForError()
	DFREF packageFolder = root:Packages:Localizer
	wave /T W_PossiblePosForRegistration1 = packageFolder:W_PossiblePosForRegistration1
	wave /T W_PossiblePosForRegistration2 = packageFolder:W_PossiblePosForRegistration2
	wave /T M_RegistrationCombinations = packageFolder:M_RegistrationCombinations
	
	variable nCombinationsRequested = DimSize(M_RegistrationCombinations, 0)
	if (nCombinationsRequested == 0)
		return "add one or more sets of calibration positions using the '>>' button"
	endif
	
	variable i, j
	variable localizationMethod = -1
	for (i = 0; i < nCombinationsRequested; i+=1)
		for (j = 0; j < 2; j += 1)
			wave thisPos = GetPositionsWaveReference(M_RegistrationCombinations[i][j])
			if (WaveExists(thisPos) == 0)
				return M_RegistrationCombinations[i][j] + " does to appear to be a valid positions wave"
			endif
			
			// all waves should have been localized in the same way
			if (NumType(NumberByKey("LOCALIZATION METHOD", note(thisPos))) == 2)
				// no localization method
				// flag as error
				return M_RegistrationCombinations[i][j] + " does to appear to be a valid positions wave"
			endif
			
			if (localizationMethod == -1)	// first run
				localizationMethod = NumberByKey("LOCALIZATION METHOD", note(thisPos))
			else
				if (localizationMethod != NumberByKey("LOCALIZATION METHOD", note(thisPos)))
					return "the positions appear to have been analyzed using different localization algorithms (e.g. Gaussian Fitting)"
				endif
			endif
		endfor
	endfor
	
	return ""
End

Function BTCreateRegistrationMap(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	switch( ba.eventCode )
		case 2: // mouse up
			// click code here
			DFREF packageFolder = root:Packages:Localizer
			wave /T M_RegistrationCombinations = packageFolder:M_RegistrationCombinations
			
			NVAR registrationMaxJump = packageFolder:V_RegistrationMaxJump
			
			variable nCombinations = DimSize(M_RegistrationCombinations, 0)
			if (nCombinations == 0)
				return 1
			endif
			
			// get an output wave name
			NewDataFolder /O root:'Registration Maps'
			DFREF outputDF = root:'Registration Maps'
			string outputName = GetNewPositionsWaveName("", outputDF, "Select an output name")
			if (strlen(outputName) == 0)
				return 1	// user cancel
			endif
			
			 // store the positions waves in wave waves
			 Make /FREE/N=(nCombinations)/WAVE W_Channel1Positions, W_Channel2Positions
			 variable i
			 for (i = 0; i < nCombinations; i+=1)
			 	W_Channel1Positions[i] = GetPositionsWaveReference(M_RegistrationCombinations[i][0])
			 	W_Channel2Positions[i] = GetPositionsWaveReference(M_RegistrationCombinations[i][1])
			 endfor
			 
			 wave M_RegistrationMap =  CreateRegistrationMap(W_Channel1Positions, W_Channel2Positions, registrationMaxJump)
			 Duplicate /O M_RegistrationMap, outputDF:$outputName
			 wave output = outputDF:$outputName
			 
			 // add a wave note
			 string listOfPos1 = "", listOfPos2 = ""
			 for (i = 0; i < nCombinations; i+=1)
			 	listOfPos1 += M_RegistrationCombinations[i][0] + ","
			 	listOfPos2 += M_RegistrationCombinations[i][1] + ","
			 endfor
			 // trim added comma
			 listOfPos1 = listOfPos1[0, strlen(listOfPos1) - 2]
			 listOfPos2 = listOfPos2[0, strlen(listOfPos2) - 2]
			 string waveNote = "TYPE:RegistrationMap;MAX SHIFT:" + num2str(registrationMaxJump) + ";"
			 waveNote += "POSITIONS1:" + listOfPos1 + ";"
			 waveNote += "POSITIONS2:" + listOfPos2 + ";"
			 
			 Note /K output, waveNote
			 
			 DoWindow /K RegistrationMapPanel
			 
			 // provide a graph for the user to check whether the registration makes sense
			 // but only use the first set of provided positions (if there are several) to
			 // keep things simple
			 wave pos1 = W_Channel1Positions[0]
			 wave pos2 = W_Channel2Positions[0]
			 variable nPositionsInChannel1 = DimSize(pos1, 0)
			 variable nPositionsInChannel2 = DimSize(pos2, 0)
			 
			 // allocate waves that will hold all of the positions and the shifted coordinates
			 Make /O/N=(nPositionsInChannel1)/D packageFolder:W_Channel1X, packageFolder:W_Channel1Y
			 Make /O/N=(nPositionsInChannel2)/D packageFolder:W_Channel2X, packageFolder:W_Channel2Y
			 Make /O/N=(nPositionsInChannel1)/D packageFolder:W_CorrectedChannel1X, packageFolder:W_CorrectedChannel1Y
			 wave W_Channel1X = packageFolder:W_Channel1X; wave W_Channel1Y = packageFolder:W_Channel1Y
			 wave W_Channel2X = packageFolder:W_Channel2X; wave W_Channel2Y = packageFolder:W_Channel2Y
			 wave W_CorrectedChannel1X = packageFolder:W_CorrectedChannel1X; wave W_CorrectedChannel1Y = packageFolder:W_CorrectedChannel1Y
			 
			 // and fetch them
			 variable xCol, yCol, zCol
			 GetColumnsForEmitterPositions(pos1, xCol, yCol, zCol)
			 W_Channel1X = pos1[p][xCol]
			 W_Channel1Y = pos1[p][yCol]
			 GetColumnsForEmitterPositions(pos2, xCol, yCol, zCol)
			 W_Channel2X = pos2[p][xCol]
			 W_Channel2Y = pos2[p][yCol]
			 
			wave Corrected = ApplyRegistrationMap(pos1, output)
			GetColumnsForEmitterPositions(Corrected, xCol, yCol, zCol)
			W_CorrectedChannel1X = Corrected[p][xCol]
			W_CorrectedChannel1Y = Corrected[p][yCol]
			
			// and plot the result
			DoWindow /F CreateRegistrationResultViewer
			if (V_flag != 1)
				Display /K=1/N=CreateRegistrationResultViewer W_Channel1Y vs W_Channel1X as "Registration result"
				AppendToGraph W_Channel2Y vs W_Channel2X
				AppendToGraph W_CorrectedChannel1Y vs W_CorrectedChannel1X
				ModifyGraph mode=3,rgb(W_Channel2Y)=(0,0,65535),marker(W_CorrectedChannel1Y)=8,rgb(W_CorrectedChannel1Y)=(1,26214,0)
				Legend/C/N=text0/J "\\s(W_Channel1Y) Channel 1\r\\s(W_Channel2Y) Channel 2\r\\s(W_CorrectedChannel1Y) Ch1/Corrected"
				Label left, "Distance (pixels)"
				Label bottom, "Distance (pixels)"
			endif
			
			return 1
			break
		case -1: // control being killed
			break
	endswitch

	return 0
End

Function /WAVE CreateRegistrationMap(W_Channel1Positions, W_Channel2Positions, maxShiftDistance)
	wave /WAVE W_Channel1Positions, W_Channel2Positions
	variable maxShiftDistance
	
	// given sets of positions measured in different detector channels, create a registration map
	// that will contain the shifts needed to overlap the emitters in pos1 onto those in pos2
	
	// The emitters need to be close enough together to qualify as the same particle,
	// where 'close enough' is determined by maxShiftDistance
	
	// The high level overview is at follows: the x and y positions in every combination
	// are combined using GroupEmitters() to yield x-y pairs that appear to belong together.
	
	// see http://www.cs.wright.edu/~agoshtas/IVC88.pdf
	
	if (DimSize(W_Channel1Positions, 0) != DimSize(W_Channel2Positions, 0))
		Abort "Need to have same number of positions waves for both channels in CreateRegistrationMap"
	endif
	variable nPositionsWaves = DimSize(W_Channel1Positions, 0)
	
	Make /N=(0, 4)/D/FREE M_CombinedEmitterPositions
	
	variable n, localizationMethod
	for (n = 0; n < nPositionsWaves; n+=1)
		wave pos1 = W_Channel1Positions[n]
		wave pos2 = W_Channel2Positions[n]
		
		// verify that the positions aren't empty
		if ((DimSize(pos1, 0) == 0) || (DimSize(pos2, 0) == 0))
			Abort "One or more of the positions waves does not contain any positions"
		endif
		
		// check if all emitters were localized in the same way
		if (n == 0)
			localizationMethod = NumberByKey("LOCALIZATION METHOD", note(pos1))
			if (NumType(localizationMethod) == 2)	// NaN
				Abort "The input wave named " + NameOfWave(pos1) + " does not appear to be a regular set of positions"
			endif
		else
			if (localizationMethod != NumberByKey("LOCALIZATION METHOD", note(pos1)))
				Abort "The positions appear to have been localized using different algorithms. Please use the same algorithm for all the images"
			endif
		endif
		
		if (localizationMethod != NumberByKey("LOCALIZATION METHOD", note(pos2)))
			Abort "The positions appear to have been localized using different algorithms. Please use the same algorithm for all the images"
		endif
		
		variable xCol, yCol, zCol
		GetColumnsForEmitterPositions(pos1, xCol, yCol, zCol)
		
		// For now this routine looks only at the first frame in each of the waves
		// and ignores the rest
		variable firstFrameIndex1 = pos1[0][0]
		variable firstFrameIndex2 = pos2[0][0]
		wave /WAVE extractedPositionsInFrame1= ExtractPositionsInFrame(pos1, firstFrameIndex1)
		wave /WAVE extractedPositionsInFrame2 = ExtractPositionsInFrame(pos2, firstFrameIndex2)
		
		wave extractedPositions1 = extractedPositionsInFrame1[0]
		wave extractedPositions2 = extractedPositionsInFrame2[0]
		
		// try to estimate whether there is a systematic shift between the particles
		// but only if it is significant
		variable systematicShiftX = 0, systematicShiftY = 0
		GuessSystematicShiftInPositions(extractedPositions1, extractedPositions2, maxShiftDistance, systematicShiftX, systematicShiftY)
		systematicShiftX = (systematicShiftX > 0.5) ? systematicShiftX : 0
		systematicShiftY = (systematicShiftY > 0.5) ? systematicShiftY : 0
		
		// subract the systematic shift from the positions in channel2. It will be added back in
		// after the emitters have been combined
		extractedPositions2[][xCol] -= systematicShiftX
		extractedPositions2[][yCol] -= systematicShiftY
		
		// the grouping functions that already exist are designed for consolidation or particle tracking
		// so make a new positions wave, in which pos2 is appended to pos1
		variable nEmittersIn1 = DimSize(extractedPositions1, 0)
		variable nEmittersIn2 = DimSize(extractedPositions2, 0)
		Duplicate /O /FREE pos1, M_AppendedEmitters	// duplicate takes care of wave note
		M_AppendedEmitters[][0] = 0
		Redimension /N=(nEmittersIn1 + nEmittersIn2, -1) M_AppendedEmitters
		M_AppendedEmitters[nEmittersIn1, ] = extractedPositions2[p - nEmittersIn1][q]
		M_AppendedEmitters[nEmittersIn1, ][0] = 1
		
		wave /WAVE W_GroupedEmitters = GroupEmitters(M_AppendedEmitters, CombinedPosEstimator_LastPos, maxShiftDistance, 0)
		
		// In W_GroupedEmitters, we're only interested in those subwaves that contain 2 rows (i.e. for which a match was successfully found). We selectively
		// extract those and convert them to a format suitable for passing to CreateRegistration().
		variable i, nMatchedEmitters = 0
		for (i = 0; i < DimSize(W_GroupedEmitters, 0); i+=1)
			wave thisGroup = W_GroupedEmitters[i]
			if (DimSize(thisGroup, 0) == 2)
				nMatchedEmitters += 1
			endif
		endfor
		
		variable offset = DimSize(M_CombinedEmitterPositions, 0)
		Redimension /N=(DimSize(M_CombinedEmitterPositions, 0) + nMatchedEmitters, -1) M_CombinedEmitterPositions
		
		for (i = 0; i < DimSize(W_GroupedEmitters, 0); i+=1)
			wave thisGroup = W_GroupedEmitters[i]
			if (DimSize(thisGroup, 0) == 2)
				M_CombinedEmitterPositions[offset][0] = thisGroup[0][xCol]
				M_CombinedEmitterPositions[offset][1] = thisGroup[0][yCol]
				M_CombinedEmitterPositions[offset][2] = thisGroup[1][xCol]
				M_CombinedEmitterPositions[offset][3] = thisGroup[1][yCol]
				
				// add the systematic offset for channel 2 back in
				M_CombinedEmitterPositions[offset][2] += systematicShiftX
				M_CombinedEmitterPositions[offset][3] += systematicShiftY
				
				offset += 1
			endif
		endfor
	endfor
	
	if (DimSize(M_CombinedEmitterPositions, 0) < 6)
		Abort "Insufficient number of reference points in the positions"
	endif
	
	wave M_RegistrationMap = CreateRegistration(M_CombinedEmitterPositions)
	
	return M_RegistrationMap
End

Function /WAVE CreateRegistration(M_EmitterPositions)
	wave M_EmitterPositions	// wave containing a row for every calibration emitter, using the format x (ch1) / y (ch1) / x (ch2) / y (ch2)
	
	assert(DimSize(M_EmitterPositions, 0) >= 6)
	assert(DimSize(M_EmitterPositions, 1) == 4)
	
	variable nEmitters = DimSize(M_EmitterPositions, 0)
	
	Make /FREE/N=(nEmitters)/D W_xPositions = M_EmitterPositions[p][0]
	Make /FREE/N=(nEmitters)/D W_yPositions = M_EmitterPositions[p][1]
	Make /FREE/N=(nEmitters)/I/U W_SortedIndices
	
	Make /FREE/N=(6)/D W_xValuesForFit, W_yValuesForFit, W_shiftInX, W_shiftInY
	
	Make /FREE/N=(nEmitters, 14)/D M_RegistrationMap
	
	variable i, xLoc, yLoc
	for (i = 0; i < nEmitters; i+=1)
		xLoc = M_EmitterPositions[i][0]
		yLoc = M_EmitterPositions[i][1]
		
		M_RegistrationMap[i][0] = xLoc
		M_RegistrationMap[i][1] = yLoc
		
		// find the 6 nearest emitters (which will include the point under consideration)
		Make /FREE/D/N=(nEmitters) W_Distances2
		MatrixOP /FREE W_Distances = sqrt(magSqr(W_xPositions - xLoc) + magSqr(W_yPositions - yLoc))
		MakeIndex W_Distances, W_SortedIndices
		W_xValuesForFit = M_EmitterPositions[W_SortedIndices[p]][0]
		W_yValuesForFit = M_EmitterPositions[W_SortedIndices[p]][1]
		W_shiftInX = M_EmitterPositions[W_SortedIndices[p]][2] - W_xValuesForFit[p]
		W_shiftInY = M_EmitterPositions[W_SortedIndices[p]][3] - W_yValuesForFit[p]
		
		// perform a 2nd order 2D poly fit for the shift in x
		CurveFit /N=1/W=2/Q poly2D 2, W_shiftInX /X=W_xValuesForFit /Y=W_yValuesForFit
		// todo: handle failed fits
		wave W_Coef
		M_RegistrationMap[i][2,7] = W_Coef[q - 2]
		
		// perform a 2nd order 2D poly fit for the shift in y
		CurveFit /N=1/W=2/Q poly2D 2, W_shiftInY /X=W_xValuesForFit /Y=W_yValuesForFit
		// todo: handle failed fits
		wave W_Coef
		M_RegistrationMap[i][8,13] = W_Coef[q - 8]
		
	endfor
	
	return M_RegistrationMap
End

Function GuessSystematicShiftInPositions(pos1, pos2, maxJump, dx, dy)
	wave pos1, pos2
	variable maxJump, &dx, &dy
	
	variable xCol, yCol, zCol
	GetColumnsForEmitterPositions(pos1, xCol, yCol, zCol)
	
	// For now this routine looks only at the first frame in each of the waves
	// and ignores the rest
	variable firstFrameIndex1 = pos1[0][0]
	variable firstFrameIndex2 = pos2[0][0]
	wave /WAVE extractedPositionsInFrame1= ExtractPositionsInFrame(pos1, firstFrameIndex1)
	wave /WAVE extractedPositionsInFrame2 = ExtractPositionsInFrame(pos2, firstFrameIndex2)
	
	wave extractedPositions1 = extractedPositionsInFrame1[0]
	wave extractedPositions2 = extractedPositionsInFrame2[0]
	
	variable nEmittersIn1 = DimSize(extractedPositions1, 0)
	variable nEmittersIn2 = DimSize(extractedPositions2, 0)
	Duplicate /O /FREE pos1, M_AppendedEmitters	// duplicate takes care of wave note
	Redimension /N=(nEmittersIn1 + nEmittersIn2, -1) M_AppendedEmitters
	M_AppendedEmitters[0, nEmittersIn1 - 1] = extractedPositions1[p][q]	// extractedPositions1 is never modified
	
	variable shiftX = 0, shiftY = 0
	variable estimatedShiftInThisRunX = 0, estimatedShiftInThisRunY = 0
	variable nIterations = 0, nMatchedEmitters, i
	do
		if (nIterations > 10)
			break
		endif
		
		extractedPositions2[][xCol] -= estimatedShiftInThisRunX
		extractedPositions2[][yCol] -= estimatedShiftInThisRunY
		
		// the grouping functions that already exist are designed for consolidation or particle tracking
		// so make a new positions wave, in which pos2 is appended to pos1
		M_AppendedEmitters[nEmittersIn1, ] = extractedPositions2[p - nEmittersIn1][q]
		M_AppendedEmitters[][0] = 0
		M_AppendedEmitters[nEmittersIn1, ][0] = 1
		
		wave /WAVE W_GroupedEmitters = GroupEmitters(M_AppendedEmitters, CombinedPosEstimator_LastPos, maxJump, 0)
		nMatchedEmitters = 0
		for (i = 0; i < DimSize(W_GroupedEmitters, 0); i+=1)
			wave thisWave = W_GroupedEmitters[i]
			if (DimSize(thisWave, 0) == 2)
				nMatchedEmitters += 1
			endif
		endfor
		
		Make /FREE/N=(nMatchedEmitters)/D W_xShift, W_yShift
		variable offset = 0
		for (i = 0; i < DimSize(W_GroupedEmitters, 0); i+=1)
			wave thisWave = W_GroupedEmitters[i]
			if (DimSize(thisWave, 0) == 2)
				W_xShift[offset] = thisWave[1][xCol] - thisWave[0][xCol]
				W_yShift[offset] = thisWave[1][yCol] - thisWave[0][yCol]
				offset += 1
			endif
		endfor
		
		estimatedShiftInThisRunX = mean(W_xShift)
		estimatedShiftInThisRunY = mean(W_yShift)
		
		shiftX += estimatedShiftInThisRunX
		shiftY += estimatedShiftInThisRunY
		
		nIterations += 1
	while ((estimatedShiftInThisRunX > 0.25) || (estimatedShiftInThisRunY > 0.25))
	
	dx = shiftX
	dy = shiftY
	
End

Function ApplyRegistrationMap_Menu()
	DFREF mapFolder = root:'Registration Maps'
	DFREF positionsFolder = root:'Localized Positions'
	
	// start by getting a list of positions waves and registration maps
	string positionsWaves = GetPossiblePositionsWaves()
	string registrationMaps = GetPossibleRegistrationMaps()
	
	variable nPositionsWaves = ItemsInList(positionsWaves)
	variable nRegistrationMaps = ItemsInList(registrationMaps)
	
	if (nPositionsWaves == 0)
		Abort "No positions waves seem to be present"
	endif
	if (nRegistrationMaps == 0)
		Abort "No registration maps seem to be present"
	endif
	
	string posName, registrationMapName
	Prompt posName, "Positions to correct (from the first channel!):", popup, positionsWaves
	Prompt registrationMapName, "Registration map to use", popup, registrationMaps
	DoPrompt "Select calculation parameters", posName, registrationMapName
	if (V_flag == 1)
		return 0
	endif
	
	wave posWave = GetPositionsWaveReference(posName)
	wave mapWave = mapFolder:$registrationMapName
	
	string newOutputName = GetNewPositionsWaveName(posName + "_mapped", positionsFolder, "Enter the name of the output wave")
	if (strlen(newOutputName) == 0)
		return 0
	endif
	
	variable nPositions = DimSize(posWave, 0)
	if (nPositions < 5000)
		wave M_Mapped = ApplyRegistrationMap(posWave, mapWave)
	else
		wave M_Mapped = ApplyRegistrationMap_Threaded(posWave, mapWave)
	endif
	Duplicate /O M_Mapped, positionsFolder:$newOutputName
End

Function /WAVE ApplyRegistrationMap(positions, M_RegistrationMap)
	wave positions, M_RegistrationMap
	
	assert(DimSize(M_RegistrationMap, 0) >= 6)
	
	variable nPositions = DimSize(positions, 0)
	variable nReferencePoints = DimSize(M_RegistrationMap, 0)
	
	variable xCol, yCol, zCol
	GetColumnsForEmitterPositions(positions, xCol, yCol, zCol)
	
	Make /FREE/N=(nReferencePoints)/D W_ReferenceX = M_RegistrationMap[p][0]
	Make /FREE/N=(nReferencePoints)/D W_ReferenceY = M_RegistrationMap[p][1]
	Make /FREE/N=(nReferencePoints)/D W_SortedIndices
	
	Make /FREE/N=6 /D W_PolyParamsX, W_PolyParamsY
	
	Duplicate /FREE positions, M_CorrectedPositions
	
	variable i, j, dx, dy, xLoc, yLoc, weight, distanceToFurthestReference, summedWeights
	for (i = 0; i < nPositions; i+=1)
		if ((mod(i, 999) == 0) && (nPositions > 2000))
			ProgressWindow("Mapping", i + 1, nPositions)
		endif
		xLoc = positions[i][xCol]
		yLoc = positions[i][yCol]
		
		// get the distances to the nearest reference points
		MatrixOP /FREE W_Distances = sqrt(magSqr(W_ReferenceX - xLoc) + magSqr(W_ReferenceY - yLoc))
		MakeIndex W_Distances, W_SortedIndices
		
		distanceToFurthestReference = W_Distances[W_SortedIndices[5]]
		summedWeights = 0
		dx = 0
		dy = 0
		// only the 6 closest reference points need to be included
		// start with x
		for (j = 0; j < 6; j+=1)
			weight = 1 - 3 * (W_Distances[W_SortedIndices[j]] / distanceToFurthestReference)^2 + 2 * (W_Distances[W_SortedIndices[j]] / distanceToFurthestReference)^3
			W_PolyParamsX = M_RegistrationMap[W_SortedIndices[j]][p + 2]
			W_PolyParamsY = M_RegistrationMap[W_SortedIndices[j]][p + 8]
			
			dx += weight * Poly2D(W_PolyParamsX, xLoc, yLoc)
			dy += weight * Poly2D(W_PolyParamsY, xLoc, yLoc)
			summedWeights += weight
		endfor
		
		dx /= summedWeights
		dy /= summedWeights
		
		M_CorrectedPositions[i][xCol] += dx
		M_CorrectedPositions[i][yCol] += dy
	endfor
	
	return M_CorrectedPositions
End

Function /WAVE ApplyRegistrationMap_Threaded(positions, M_RegistrationMap)
	wave positions, M_RegistrationMap
	
	variable nPositions = DimSize(positions, 0)
	
	// allocate the output
	Duplicate /FREE positions, M_CorrectedPositions
	// and the wave used to report progress
	Make /FREE/D/N=1 W_Progress = 0
	
	// set up the threads
	// assume each thread needs at least a 1000 positions to be worthwhile
	variable nThreads = min(floor(nPositions / 1000), ThreadProcessorCount)
	if (nThreads == 0)
		nThreads = 1
	endif
	
	variable tgID = ThreadGroupCreate(nThreads)
	
	variable i, nPositionsPerThread = floor(nPositions / nThreads)
	variable startOfThisThread, endOfThisThread = 0
	Make /FREE/N=(nThreads, 2)/D M_ThreadStartStop
	M_ThreadStartStop[][0] = p * nPositionsPerThread
	M_ThreadStartStop[][1] = ((p + 1) * nPositionsPerThread) - 1
	M_ThreadStartStop[nThreads - 1][1] = nPositions - 1	// last thread is special since it may not have a full set of data to work on
	
	for (i = 0; i < nThreads; i+=1)
		ThreadStart tgID, i, ApplyRegistrationMap_Worker(positions, M_RegistrationMap, M_CorrectedPositions, M_ThreadStartStop[i][0], M_ThreadStartStop[i][1], W_Progress)
	endfor
	
	// now wait for the threads to finish and occasionally report progress
	variable threadStatus, nFinished
	for (;;)
		threadStatus = ThreadGroupWait(tgID, 100)
		if (threadStatus == 0)
			// all threads finished
			break
		endif
		
		// progress notification
		if (nPositions > 5000)
			nFinished = W_Progress[0]
			ProgressWindow("Mapping", nFinished, nPositions)
		endif
	endfor
	
	threadStatus = ThreadGroupRelease(tgID)
	
	return M_CorrectedPositions
End

ThreadSafe Function ApplyRegistrationMap_Worker(positions, M_RegistrationMap, M_CorrectedPositions, firstIndexToWorkOn, lastIndexToWorkOn, W_Progress)
	wave positions, M_RegistrationMap
	wave M_CorrectedPositions	// corrected result goes here
	variable firstIndexToWorkOn, lastIndexToWorkOn
	wave W_Progress	// a wave containing a single point that will be incremented by all worker functions to show progress.
						// intrinsically unsafe, but since it's only for progress reporting there's no problem if the incrementing
						// goes wrong from time to time.
	
	variable nPositions = lastIndexToWorkOn - firstIndexToWorkOn + 1
	variable nReferencePoints = DimSize(M_RegistrationMap, 0)
	
	variable xCol, yCol, zCol
	GetColumnsForEmitterPositions(positions, xCol, yCol, zCol)
	
	Make /FREE/N=(nReferencePoints)/D W_ReferenceX = M_RegistrationMap[p][0]
	Make /FREE/N=(nReferencePoints)/D W_ReferenceY = M_RegistrationMap[p][1]
	Make /FREE/N=(nReferencePoints)/D W_SortedIndices
	
	Make /FREE/N=6 /D W_PolyParamsX, W_PolyParamsY
	
	variable i, j, dx, dy, xLoc, yLoc, weight, distanceToFurthestReference, summedWeights
	variable nPreviouslyFinished = 0, nFinished = 0
	for (i = firstIndexToWorkOn; i <= lastIndexToWorkOn; i+=1)
		if (mod(i - firstIndexToWorkOn, 50) == 0)
			W_Progress[0] += nFinished - nPreviouslyFinished
			nPreviouslyFinished = nFinished
		endif
		xLoc = positions[i][xCol]
		yLoc = positions[i][yCol]
		
		// get the distances to the nearest reference points
		MatrixOP /FREE W_Distances = sqrt(magSqr(W_ReferenceX - xLoc) + magSqr(W_ReferenceY - yLoc))
		MakeIndex W_Distances, W_SortedIndices
		
		distanceToFurthestReference = W_Distances[W_SortedIndices[5]]
		summedWeights = 0
		dx = 0
		dy = 0
		// only the 6 closest reference points need to be included
		// start with x
		for (j = 0; j < 6; j+=1)
			weight = 1 - 3 * (W_Distances[W_SortedIndices[j]] / distanceToFurthestReference)^2 + 2 * (W_Distances[W_SortedIndices[j]] / distanceToFurthestReference)^3
			W_PolyParamsX = M_RegistrationMap[W_SortedIndices[j]][p + 2]
			W_PolyParamsY = M_RegistrationMap[W_SortedIndices[j]][p + 8]
			
			dx += weight * Poly2D(W_PolyParamsX, xLoc, yLoc)
			dy += weight * Poly2D(W_PolyParamsY, xLoc, yLoc)
			summedWeights += weight
		endfor
		
		dx /= summedWeights
		dy /= summedWeights
		
		M_CorrectedPositions[i][xCol] += dx
		M_CorrectedPositions[i][yCol] += dy
		
		nFinished += 1
	endfor
End

Function /WAVE CombineEmittersIntoPosition(positions, localizationStdDev, trackLength)
	wave positions
	variable &localizationStdDev
	variable &trackLength
	// this function is called during the emitter consolidation
	// it takes a set of fitted positions from different frames,
	// thought to arise from the same emitter
	// these localizations are combined into a single one and returned
	
	// additionally the following is performed:
	// the localization coordinates are obtained as an intensity-weighted average
	// the integrated intensity is the sum of all the intensities
	// the localization uncertainty is set to the average divided by the square root of the number of positions
	
	// the consolidation also provides some estimate of the localization error by looking at the shifts in position that occur
	// in what is assumed to be the same emitter. This is returned by reference in localizationStdDev.
	// while this is not perfect by any means, it does provide a seemingly unbiased estimate
	// it also returns the number of positions that were combined via trackLength
	variable nPositions = DimSize(positions, 0)
	trackLength = nPositions
	variable i
	
	if (nPositions == 0)
		return positions
	endif
	
	variable nFramesPresentColumn, xCol, yCol, zCol, intensityColumn, xUncertaintyCol, yUncertaintyCol, zUncertaintyCol
	GetColumnForNFramesPresent(positions, nFramesPresentColumn)	// handle different types of positions
	GetColumnsForEmitterPositions(positions, xCol, yCol, zCol)
	GetColumnForIntegratedIntensity(positions, intensityColumn)
	GetColumnsForLocalizationError(positions, xUncertaintyCol, yUncertaintyCol, zUncertaintyCol)
	
	// make the output wave with the combined positions
	Make /N=(1, DimSize(positions, 1)) /FREE /D M_Combined
	M_Combined[0][] = positions[0][q]
	
	// combine the different localizationa into one
	// give each localization the same weight, unless there is an estimate
	// for the intensity of each emitter
	variable sumIntensity = 0, xLoc = 0, yLoc = 0, zLoc = 0, xDev = 0, yDev = 0, zDev = 0
	variable thisIntensity
	for (i = 0; i < nPositions; i+=1)
		if (intensityColumn != -1)
			thisIntensity = positions[i][intensityColumn]
		else
			thisIntensity = 1
		endif
		xLoc += thisIntensity * positions[i][xCol]
		yLoc += thisIntensity * positions[i][yCol]
		if (zCol != -1)
			zLoc += positions[i][intensityColumn] * positions[i][zCol]
		endif
		if (xUncertaintyCol != -1)
			xDev += positions[i][xUncertaintyCol]
		endif
		if (yUncertaintyCol != -1)
			yDev += positions[i][yUncertaintyCol]
		endif
		if (zUncertaintyCol != -1)
			zDev += positions[i][zUncertaintyCol]
		endif
		sumIntensity += thisIntensity
	endfor
	xLoc /= sumIntensity
	yLoc /= sumIntensity
	zLoc /= sumIntensity
	xDev /= (nPositions * sqrt(nPositions))
	yDev /= (nPositions * sqrt(nPositions))
	zDev /= (nPositions * sqrt(nPositions))
	
	if (intensityColumn != -1)
		M_Combined[0][intensityColumn] = sumIntensity
	endif
	M_Combined[0][xCol] = xLoc
	M_Combined[0][yCol] = yLoc
	if (zCol != -1)
		M_Combined[0][zCol] = zLoc
	endif
	if (xUncertaintyCol != -1)
		M_Combined[0][xUncertaintyCol] = xDev
	endif
	if (yUncertaintyCol != -1)
		M_Combined[0][yUncertaintyCol] = yDev
	endif
	if (zUncertaintyCol != -1)
		M_Combined[0][zUncertaintyCol] = zDev
	endif
	M_Combined[0][nFramesPresentColumn] = nPositions
	
	// now try to get the localization estimate
	// we don't do intensity weighing
	MatrixOP /FREE W_summedVariance = varCols(col(positions, xCol)) + varCols(col(positions, yCol))
	localizationStdDev = sqrt(W_summedVariance[0][0])
	
	return M_Combined
End
	
	

Function BTCorrectDriftProc(ba) : ButtonControl
	STRUCT WMButtonAction &ba
	
	switch( ba.eventCode )
		case 2: // mouse up
			// click code here
			string baseWindowName = GetBaseWindowName(ba.win)
			
			ControlInfo /W=$baseWindowName#OtherControls PMDriftCorrectionMode
			string correctionModeName = S_Value
			
			string positionsWaves = GetPossiblePositionsWaves()
			if (ItemsInList(positionsWaves) == 0)
				Abort "Please make sure that there are some fitted positions in the experiment first"
			endif
			
			ControlInfo /W=$baseWindowName#OtherControls PMDriftCorrSelectPositionsWave
			string positionsWaveName = S_Value	
			if (StringMatch(positionsWaveName, "* No Positions *") == 1)
				Abort "Please select positions to correct first"
			endif
			
			wave positions = GetPositionsWaveReference(positionsWaveName)
			if (WaveExists(positions) != 1)
				Abort "The selected positions do not appear to be valid"
			endif
			
			// check if the positions have been consolidated, if so then refuse
			variable nFramesPresentCol
			GetColumnForNFramesPresent(positions, nFramesPresentCol)
			ImageTransform /G=(nFramesPresentCol) sumCol, positions
			if (V_value != DimSize(positions, 0))
				Abort "The requested positions have been consolidated, there is no point in performing the drift correction"
			endif
			
			StrSwitch (correctionModeName)
				case "Automatic":
					DoAutomaticDriftCorrection(positions, baseWindowName)
					break
				case "Manual":
					DoManualDriftCorrection(baseWindowName)
					break
				default:
					Abort "Unknown correction mode in BTCorrectDriftProc()"
					break
			EndSwitch
			
			break
	EndSwitch
End

Function /WAVE DoManualDriftCorrection(baseWindowName)
	string baseWindowName
	
	DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
	NVAR driftCorrectionMaxJump = windowDataFolder:V_driftCorrectionMaxJump
	variable msTimer, sec, err
	string positionsWaves, panelName
	
	// set up the waves that will contain the clicked positions and the listbox entries (both updated by the window hook function)
	Make /D/O/N=(0, 2) windowDataFolder:M_ClickedCoordinates
	Make /T/O/N=(0, 2) windowDataFolder:M_ClickedTextCoordinates
	Make /T/O/N=(2) windowDataFolder:M_DriftListBoxTitles = {"X", "Y"}
	
	wave M_ClickedCoordinates = windowDataFolder:M_ClickedCoordinates
	wave /T M_ClickedTextCoordinates = windowDataFolder:M_ClickedTextCoordinates
	wave /T M_DriftListBoxTitles = windowDataFolder:M_DriftListBoxTitles
	
	// draw the main panel
	DoWindow /K DriftCorrection	// in case the user still had a panel floating around
	
	NewPanel /K=1 /N=DriftCorrection /W=(949,50,1197,306)
	panelName = S_name
	SetDrawLayer UserBack
	SetDrawEnv fsize= 11
	DrawText 17,38,"1. Make sure the first 'good' frame\rin the movie is displayed"
	SetDrawEnv fsize= 11
	DrawText 18,65,"2. Click the fiducial markers in the image."
	SetDrawEnv fsize= 11
	DrawText 22,222,"3. Click continue"
	ListBox LBDriftCorrectionMarkerPosition,pos={68,78},size={107,118}, listWave = M_ClickedTextCoordinates, titleWave=M_DriftListBoxTitles
	Button BTDriftCorrectionContinue,pos={74,227},size={100,20},proc=BTDriftMarkerContinue,title="Continue"
	SetWindow $panelName, hook(KillHook) = DriftCorrection_WindowHook
	
	// set the name of the corresponding interface panel in the userdata
	// since it's difficult to retrieve otherwise
	SetWindow $panelName, UserData = baseWindowName
	
	AutoPositionWindow /R=$baseWindowName /M=0 $panelName
	// attach the window hook function to the CCD Viewer
	SetWindow $baseWindowName, hook(DriftMarkerHook) = CaptureClickedCoordinates_Hook
End

Function DriftCorrection_WindowHook(s)
	STRUCT WMWinHookStruct &s
	
	switch (s.eventcode)
		case 2:	// window kill, remove the window hook function from CCDViewer
			string interfacePanelName = GetUserData(s.winName, "", "")
			DoWindow $interfacePanelName
			if (V_flag != 0)
				SetWindow $interfacePanelName hook(DriftMarkerHook) = $""
				RemoveFromGraph /Z /W=$interfacePanelName#CCDViewer M_ClickedCoordinates
			endif
			break
		default:
			return 0
	endswitch
	
	return 0
End

Function CaptureClickedCoordinates_Hook(s) 
	STRUCT WMWinHookStruct &s
	// this function attaches to a window of choice and appends the X- and Y-coordinates where the user clicked the mouse to a specific wave
	// both a numeric and a text version of the wave is made (for use with listbox controls)
	
	switch (s.eventcode)
		case 3:	// mousedown
			variable xVal, yVal
			string baseWindowName = GetBaseWindowName(s.winName)
			DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
			
			wave /Z M_ClickedCoordinates = windowDataFolder:M_ClickedCoordinates
			wave /T/Z M_ClickedTextCoordinates = windowDataFolder:M_ClickedTextCoordinates
			
			// get the axis range of the viewer graph so clicks outside the graph window can be rejected
			variable minX, minY, maxX, maxY
			GetAxis /Q/W=$baseWindowName#CCDViewer Bottom
			minX = V_min
			maxX = V_max
			GetAxis /Q/W=$baseWindowName#CCDViewer Left
			minY = V_min
			maxY = V_max
			
			if (WaveExists(M_ClickedCoordinates) == 0)
				Make /D/N=(0, 2) windowDataFolder:M_ClickedCoordinates
				wave M_ClickedCoordinates = windowDataFolder:M_ClickedCoordinates
			endif
			if (WaveExists(M_ClickedTextCoordinates) == 0)
				Make /T /N=(0, 2) windowDataFolder:M_ClickedTextCoordinates
				wave /T M_ClickedTextCoordinates = windowDataFolder:M_ClickedTextCoordinates
			endif
			
			if (DimSize(M_ClickedCoordinates, 0) == 0)
				// no clicked coordinates yet
				// append the wave to the graph so that the user can see
				// where they clicked
				AppendToGraph /W=$baseWindowName#CCDViewer M_ClickedCoordinates[][1] vs M_ClickedCoordinates[][0]
				ModifyGraph /W=$baseWindowName#CCDViewer mode(M_ClickedCoordinates)=3
			endif
			
			xVal = AxisValFromPixel(baseWindowName + "#CCDViewer", "Bottom", s.mouseLoc.h)
			yVal = AxisValFromPixel(baseWindowName + "#CCDViewer", "Left", s.mouseLoc.v)
			
			if ((xVal < minX) || (xVal > maxX) || (yVal < minY) || (yVal > maxY))
				return 1
			endif
			
			Redimension /N=(DimSize(M_ClickedCoordinates, 0) + 1, 2) M_ClickedCoordinates
			M_ClickedCoordinates[DimSize(M_ClickedCoordinates, 0) - 1][0] = xVal
			M_ClickedCoordinates[DimSize(M_ClickedCoordinates, 0) - 1][1] = yVal
			
			Redimension /N=(DimSize(M_ClickedTextCoordinates, 0) + 1, 2) M_ClickedTextCoordinates
			M_ClickedTextCoordinates[DimSize(M_ClickedTextCoordinates, 0) - 1][0] = num2str(xVal)
			M_ClickedTextCoordinates[DimSize(M_ClickedTextCoordinates, 0) - 1][1] = num2str(yVal)
			
			return 1
			break
		default:
			return 0
			break
	endswitch
End

Function BTDriftMarkerContinue(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	switch( ba.eventCode )
		case 2: // mouse up
			// click code here
			string interfacePanelName = GetUserData(ba.win, "", "")
			DFREF windowDataFolder = GetWindowDataFolder(interfacePanelName)
			NVAR currentFrame = windowDataFolder:V_currentImage
			NVAR maxJump = windowDataFolder:V_driftCorrectionMaxJump
			
			wave /T M_ClickedTextCoordinates = windowDataFolder:M_ClickedTextCoordinates
			wave M_ClickedCoordinates = windowDataFolder:M_ClickedCoordinates
			
			variable nMarkers = DimSize(M_ClickedTextCoordinates, 0), nPositions
			string positionsWaveName
			variable i, nearestX, nearestY, index
			
			ControlInfo /W=$interfacePanelName#OtherControls PMDriftCorrSelectPositionsWave
			positionsWaveName = S_Value
			
			DoWindow /K DriftCorrection
			
			if (StringMatch(positionsWaveName, "* No Positions *") == 1)
				KillWaves /Z M_ClickedCoordinatesText
				KillWaves /Z windowDataFolder:M_ClickedCoordinates
				return 0
			endif
			
			wave positions = GetPositionsWaveReference(positionsWaveName)
			nPositions = DimSize(positions, 0)
			
			if (nMarkers == 0)	// no markers indicated
				KillWaves /Z M_ClickedCoordinatesText
				KillWaves /Z M_ClickedCoordinates
				return 0
			endif
			
			// check if the positions have been consolidated, if so then refuse
			variable nFramesPresentCol
			GetColumnForNFramesPresent(positions, nFramesPresentCol)
			ImageTransform /G=(nFramesPresentCol) sumCol, positions
			if (V_value != DimSize(positions, 0))
				Abort "The requested positions have been consolidated, there is no point in performing the drift correction"
			endif
			
			// extract the positions that are fitted in the current frame
			wave /WAVE M_ExtractedPositions = ExtractPositionsInFrame(positions, currentFrame)
			wave extractedPositions = M_ExtractedPositions[0]
			wave extractedPositionsIndices = M_ExtractedPositions[1]
			
			if (DimSize(extractedPositions, 0) == 0)	// no positions were fitted in this frame
				KillWaves /Z M_ClickedCoordinatesText, extractedPositions
				KillWaves /Z M_ClickedCoordinates
				Abort "No positions were fitted in the current frame, not even the markers!"
			endif
			
			Make /D/O/N=(nMarkers, 2) windowDataFolder:RecoveredMarkerPositions	// the coordinates that we have recovered for the markers
			wave RecoveredMarkerPositions = windowDataFolder:RecoveredMarkerPositions
			RecoveredMarkerPositions = str2num(M_ClickedTextCoordinates[p][q])
			
			for (i = 0; i < nMarkers; i+=1)
				returnPointNearestFitPositions(extractedPositions, RecoveredMarkerPositions[i][0], RecoveredMarkerPositions[i][1], nearestX, nearestY, index)
				RecoveredMarkerPositions[i][0] = nearestX
				RecoveredMarkerPositions[i][1] = nearestY
			endfor
			
			// the user might have clicked the same marker more than once
			// so delete identical positions
			variable offset = 0, j, k
			for (i = 0; i < DimSize(RecoveredMarkerPositions, 0); i+=1)
				for (j = i + 1; j < DimSize(RecoveredMarkerPositions, 0); j+=1)
					if ((RecoveredMarkerPositions[i][0] == RecoveredMarkerPositions[j][0]) && (RecoveredMarkerPositions[i][1] == RecoveredMarkerPositions[j][1]))
						// remove the markers pointed to by j
						Duplicate /FREE RecoveredMarkerPositions, OldRecoveredMarkerPositions
						Redimension /N=(DimSize(OldRecoveredMarkerPositions, 0) - 1, -1) RecoveredMarkerPositions
						for (k = 0; k < DimSize(OldRecoveredMarkerPositions, 0); k+=1)
							if (offset == j)
								continue
							endif
							RecoveredMarkerPositions[offset][] = OldRecoveredMarkerPositions[k][q]
							offset += 1
						endfor
						j -= 1
					endif
				endfor
			endfor
					
			nMarkers = DimSize(RecoveredMarkerPositions, 0)
			
			Make /D/O/N=(nMarkers) windowDataFolder:W_Markers_X
			Make /D/O/N=(nMarkers) windowDataFolder:W_Markers_Y
			wave W_Markers_X = windowDataFolder:W_Markers_X
			wave W_Markers_Y = windowDataFolder:W_Markers_Y
			
			W_Markers_X = RecoveredMarkerPositions[p][0]
			W_Markers_Y = RecoveredMarkerPositions[p][1]
			
			DoWindow /F $interfacePanelName
			AppendToGraph /W=$interfacePanelName#CCDViewer W_Markers_Y vs W_Markers_X
			ModifyGraph /W=$interfacePanelName#CCDViewer mode(W_Markers_Y)=3
			ModifyGraph /W=$interfacePanelName#CCDViewer rgb(W_Markers_Y)=(0,65535,0)
			
			DoUpdate
			
			DoAlert 1, "The green markers on the image highlight the presumed marker positions. Accept?"
			if (V_flag != 1)	// user didn't choose 'yes'
				RemoveFromGraph /W=$interfacePanelName#CCDViewer W_Markers_Y
				KillWaves /Z M_ClickedCoordinatesText, extractedPositions, RecoveredMarkerPositions, W_Markers_X, W_Markers_Y
				KillWaves /Z windowDataFolder:M_ClickedCoordinates
				return 0
			endif
			
			RemoveFromGraph /W=$interfacePanelName#CCDViewer W_Markers_Y
			KillWaves /Z W_Markers_X, W_Markers_Y
			
			// ask for an output name
			DFREF positionsFolder = root:'Localized Positions'
			string outputName = CleanupName(NameOfWave(positions) + "_driftCorr", 0)
			outputName = GetNewPositionsWaveName(outputName, positionsFolder, "Enter a name for the corrected positions wave")
			if (strlen(outputName) == 0)	// the user canceled the dialog
				DoWindow /K positionsWavesViewer
				return 0
			endif
			
			// do the real work
			wave M_DriftParameters = CalculateDrift_ExplicitMarkers(positions, RecoveredMarkerPositions, maxJump)
			ApplyAndDisplayDriftCorrection(positions, M_DriftParameters, outputName)
	endswitch
	return 0
End

Function DoAutomaticDriftCorrection(positions, baseWindowName)
	wave positions
	string baseWindowName
	
	DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
	NVAR maxJump = windowDataFolder:V_driftCorrectionMaxJump
	
	variable maxJumpLocal = maxJump
	variable trackLengthLocal = 25
	variable nBlinkingLocal = 0
	
	Prompt maxJumpLocal, "Max positions jump:"
	Prompt trackLengthLocal, "Min number of frames emitter must be visible:"
	Prompt nBlinkingLocal, "Emitter may disappear for ? frames:"
	DoPrompt "Enter calculation parameters", maxJumpLocal, trackLengthLocal, nBlinkingLocal
	if (V_flag == 1)
		return 0
	endif
	
	maxJump = maxJumpLocal
	
	// ask for an output name
	DFREF positionsFolder = root:'Localized Positions'
	string outputName = CleanupName(NameOfWave(positions) + "_driftCorr", 0)
	outputName = GetNewPositionsWaveName(outputName, positionsFolder, "Enter a name for the corrected positions wave")
	if (strlen(outputName) == 0)	// the user canceled the dialog
		return 0
	endif
	
	wave M_DriftParameters = DriftCorrection_automatic(positions, maxJumpLocal, trackLengthLocal, nBlinkingLocal)
	ApplyAndDisplayDriftCorrection(positions, M_DriftParameters, outputName)
End

Function /WAVE CalculateDrift_ExplicitMarkers(positions, markers, maxStepSize)
	wave positions, markers
	variable maxStepSize
	
	variable nPositions = DimSize(positions, 0)
	
	Assert(WaveExists(positions))
	Assert(WaveExists(markers))
	Assert(maxStepSize >= 0)
	
	variable nMarkers = DimSize(markers, 0)
	variable nTotalFrames = positions[nPositions - 1][0] + 1
	variable startFrame = positions[0][0]
	variable i
	variable currentFrame
	variable driftX, driftY, nearestX, nearestY, index
	variable nMarkersRecovered
	
	Make /N=(nTotalFrames)/FREE/D W_RelativeDriftX = 0, W_RelativeDriftY = 0, W_nMarkersUsed = 0
	
	Make /D/O/N=(nMarkers, 2) /FREE M_previousPositions	// the positions of the markers in the previous frame, not taking drift into account
	M_previousPositions = markers[p][q]	// initialize the marker positions to the given values
	
	for (currentFrame = 0; currentFrame < nTotalFrames; currentFrame += 1)
		ProgressWindow("Estimating", currentFrame + 1, nTotalFrames)
		// extract the positions found in the current frame
		wave /WAVE M_Extracted = ExtractPositionsInFrame(positions, currentFrame)
		wave extractedPositions = M_Extracted[0]
		
		if (DimSize(extractedPositions, 0) == 0)	// no positions in the current frame
			continue
		endif
		
		driftX = 0
		driftY = 0
		nMarkersRecovered = 0
		for (i = 0; i < nMarkers; i += 1)
				returnPointNearestFitPositions(extractedPositions, M_previousPositions[i][0], M_previousPositions[i][1], nearestX, nearestY, index)
				if (sqrt((nearestX - M_previousPositions[i][0])^2 + (nearestY - M_previousPositions[i][1])^2) > maxStepSize)
					// it looks like this particular marker was not recovered in the fit in this frame
					// we shouldn't use it
					continue
				else	// the marker was recovered
					nMarkersRecovered += 1
					driftX += nearestX - M_previousPositions[i][0]
					driftY += nearestY - M_previousPositions[i][1]
					M_previousPositions[i][0] = nearestX
					M_previousPositions[i][1] = nearestY
				endif
		endfor
		
		if (nMarkersRecovered == 0)
			continue
		endif
		
		// take the average of the individual drifts as the absolute drift
		// and store the relative drift compared to the previous frame
		W_RelativeDriftX[currentFrame] = driftX / nMarkersRecovered
		W_RelativeDriftY[currentFrame] = driftY / nMarkersRecovered
		W_nMarkersUsed[currentFrame] += nMarkersRecovered
	endfor
	
	// combine the waves and return them as a single 2D wave
	Make /N=(nTotalFrames, 3)/FREE/D M_DriftParameters
	M_DriftParameters[][0] = W_nMarkersUsed[p]
	M_DriftParameters[][1] = W_RelativeDriftX[p]
	M_DriftParameters[][2] = W_RelativeDriftY[p]
	
	return M_DriftParameters
End

Function /WAVE DriftCorrection_automatic(positions, maxStepSize, minTrackLength, maxBlinking)
	wave positions
	variable maxStepSize, minTrackLength, maxBlinking
	
	assert(minTrackLength >= 1)
	assert(maxStepSize >= 0)
	assert(maxBlinking >= 0)
	
	// get the indices of the columns containing the x and y coordinates
	variable xCol, yCol, zCol
	getColumnsForEmitterPositions(positions, xCol, yCol, zCol)
	
	variable nPositions = DimSize(positions, 0)
	variable nTotalFrames = positions[nPositions - 1][0] + 1
	
	Make /N=(nTotalFrames)/FREE/D W_RelativeDriftX = 0, W_RelativeDriftY = 0, W_nMarkersUsed = 0
	
	wave /WAVE W_GroupedEmitters = GroupEmitters(positions, CombinedPosEstimator_LastPos, maxStepSize, maxBlinking)
	
	variable nGroups = DimSize(W_GroupedEmitters, 0), nPositionsInThisGroup
	variable i, driftX, driftY, j
	for (i = 0; i < nGroups; i+=1)
		ProgressWindow("Estimating", i + 1, nGroups)
		wave thisGroup = W_GroupedEmitters[i]
		nPositionsInThisGroup = DimSize(thisGroup, 0)
		
		if (nPositionsInThisGroup < minTrackLength)
			continue
		endif
		
		for (j = 1; j < nPositionsInThisGroup; j+=1)
			W_RelativeDriftX[thisGroup[j][0]] += thisGroup[j][xCol] - thisGroup[j - 1][xCol]
			W_RelativeDriftY[thisGroup[j][0]] += thisGroup[j][yCol] - thisGroup[j - 1][yCol]
			W_nMarkersUsed[thisGroup[j][0]] += 1
		endfor
	endfor
	
	W_RelativeDriftX /= W_nMarkersUsed[p]
	W_RelativeDriftY /= W_nMarkersUsed[p]
	
	// remove NaN's
	MatrixOP /FREE W_RelativeDriftX = ReplaceNaNs(W_RelativeDriftX, 0)
	MatrixOP /FREE W_RelativeDriftY = ReplaceNaNs(W_RelativeDriftY, 0)
	
	// combine the waves and return them as a single 2D wave
	Make /N=(nTotalFrames, 3)/FREE/D M_DriftParameters
	M_DriftParameters[][0] = W_nMarkersUsed[p]
	M_DriftParameters[][1] = W_RelativeDriftX[p]
	M_DriftParameters[][2] = W_RelativeDriftY[p]
	
	return M_DriftParameters
End

Function ApplyAndDisplayDriftCorrection(positions, M_DriftParameters, outputPositionsName)
	wave positions, M_DriftParameters
	string outputPositionsName
	
	assert(DimSize(M_DriftParameters, 1) == 3)
	
	DFREF packageFolder = root:Packages:Localizer
	variable nPositions = DimSize(positions, 0)
	variable nFrames = DimSize(M_DriftParameters, 0)
	
	variable xCol, yCol, zCol
	getColumnsForEmitterPositions(positions, xCol, yCol, zCol)
	
	// start by calculating the absolute drift
	Make /O/N=(nFrames, 3) packageFolder:M_AbsoluteDrift
	wave M_AbsoluteDrift = packageFolder:M_AbsoluteDrift
	
	M_AbsoluteDrift[0][] = 0
	M_AbsoluteDrift[1,][0] = M_DriftParameters[p][0]
	M_AbsoluteDrift[1,][1,2] = M_AbsoluteDrift[p - 1][q] + M_DriftParameters[p][q]

	DFREF positionsFolder = root:'Localized Positions'
	Duplicate /O positions, positionsFolder:$outputPositionsName
	wave correctedPos = positionsFolder:$outputPositionsName
	
	variable i
	for (i = 0; i < nPositions; i+=1)
		correctedPos[i][xCol] += M_AbsoluteDrift[correctedPos[i][0]][1]
		correctedPos[i][yCol] += M_AbsoluteDrift[correctedPos[i][0]][2]
	endfor
	
	DoWindow /F AbsoluteDriftViewer
	if (V_flag == 0)
		Display /K=1 /N=AbsoluteDriftViewer M_AbsoluteDrift[][1] as "Estimated drift"
		AppendToGraph M_AbsoluteDrift[][2]
		AppendToGraph /R M_AbsoluteDrift[][0]
		ModifyGraph rgb(M_AbsoluteDrift)=(0,0,65535)
		ModifyGraph rgb(M_AbsoluteDrift#2)=(2,39321,1)
		Legend/C/N=text0/J "\\s(M_AbsoluteDrift) Drift in X\r\\s(M_AbsoluteDrift#1) Drift in Y\r\\s(M_AbsoluteDrift#2) # Emitters used"
		Label left "Drift (pixel)"
		Label right "Number of emitters used in the correction"
		Label bottom "Frame Number"
	endif
End

Function RecoverEmittersByBleachingSteps(maxJumpDistance, windowName)
	variable maxJumpDistance
	string windowName
	
	string baseWindowName = GetBaseWindowName(windowName)
	DFREF windowDataFolder = GetWindowDataFolder(baseWindowName)
	
	NewDataFolder /O root:'Localized Positions'
	DFREF savDF = GetDataFolderDFR()
	
	Struct FitData fp
	// get all the fit options and parameters the user defined in the GUI
	GetGUIFitSettingsAndOptions(fp, baseWindowName)
	
	// Currently only symmetric 2D Gaussian fitting is supported
	if ((fp.localizationMethod != LOCALIZATION_GAUSS_FITTING) && (fp.localizationMethod != LOCALIZATION_GAUSS_FITTING_FIX))
		Abort "Currently only 2D symmetric Gaussian fitting is supported for the bleaching analysis"
	endif
	
	variable previousLoadedImage = fp.currentImage
	variable i, j, nPositionsFound, nPositionsOnWaitingList, nearestX, nearestY, closestIndex
	variable xCol, yCol, zCol, intensityCol, widthCol, dummyCol
	variable firstRun = 1
	variable offset, err
	
	for (i = fp.numberOfImages - 1; i >= 0; i-=1)
		// provide a progress window
		ProgressWindow("Bleaching", fp.numberOfImages - 1 - i, fp.numberOfImages - 1)
		
		ChangeCurrentImage(i, baseWindowName)
		wave Viewer_Temp = windowDataFolder:M_CCDImage
		
		 if (firstRun == 1)
			// set up a wave that will contain the subtracted frame
			Duplicate /O Viewer_Temp, windowDataFolder:M_Subtracted
			wave M_Subtracted = windowDataFolder:M_Subtracted
			Redimension /D M_Subtracted
			
			// set up a wave that will contain the contributions of all emitters that have been identified
			Duplicate /O Viewer_Temp, windowDataFolder:M_SummedEmitters
			Wave M_SummedEmitters = windowDataFolder:M_SummedEmitters
			Redimension /D M_SummedEmitters
			FastOP M_SummedEmitters = 0
		endif
		
		SetDataFolder windowDataFolder
		
		// subtract all emitters that have been localized in previous frames
		MatrixOP /O M_Subtracted = Viewer_Temp - M_SummedEmitters
		
		LocalizationAnalysis /M=(fp.localizationMethod) /D=(fp.thresholdMethod) /Y=(CAMERA_TYPE_IGOR_WAVE) /G={fp.preprocessing, fp.postprocessing} /F=(fp.particlefinder) /PFA=(fp.PFA) /T=(fp.directThresholdLevel) /R=(fp.minDistanceBetweenParticles) /W=(fp.PSFWidth) /PVER={fp.PVerSymm, fp.PVerEllipse, fp.PVerOverlap} /Z /Q /DEST=windowDataFolder:POS_out "M_Subtracted"
		SetDataFolder savDF
		
		if (V_flag != 0)	// an error occurred during the fitting
			string errorMessage = "An error occurred while fitting frame " + num2str(i)
			Abort errorMessage
		endif
		
		wave POS_out = windowDataFolder:POS_out
		 nPositionsFound = DimSize(POS_out, 0)
		 
		 if (firstRun == 1)
			// make the output wave for the positions
			Duplicate /O POS_out, root:'Localized Positions':BleachingAnalysis
			wave M_BleachingAnalysis = root:'Localized Positions':BleachingAnalysis
			Redimension /N=(0, DimSize(POS_out, 1)) M_BleachingAnalysis
			
			Duplicate /O POS_out, windowDataFolder:M_WaitingList			// a wave that will contain a list of points encountered in the previous frame
			wave M_WaitingList = windowDataFolder:M_WaitingList			// the reason for its existence is that the frame in which we first encounter an emitter may not show its correct intensity since the emitter bleaches somewhere in that frame
			Redimension /N=(0, DimSize(POS_out, 1)) M_WaitingList			// so when a new point is found, add it to the waiting list first, and then only include it in the analysis if the next frame shows it again
			
			// get some indices into the columns of the positions wave
			GetColumnsForEmitterPositions(POS_out, xCol, yCol, zCol)
			GetColumnForIntegratedIntensity(POS_out, intensityCol)
			GetColumnForFittedWidth(POS_out, widthCol, dummyCol)
			
			// if no width or amplitude is provided then abort
			if ((intensityCol < 0) || (widthCol < 0))
				Abort "The bleaching analysis requires a fitting method that provides amplitude and width information (e.g. Gaussian fitting)"
			endif
			
			// set up some viewer windows
			DoWindow /F BleachingAnalysis_Original
			if (V_flag == 0)
				Display /K=1 /N=BleachingAnalysis_Original ; AppendImage Viewer_Temp
				AppendToGraph /W=BleachingAnalysis_Original M_BleachingAnalysis[][yCol] vs M_BleachingAnalysis[][xCol]
				ModifyGraph /W=BleachingAnalysis_Original mode=3
			endif
			DoWindow /F BleachingAnalysis_Subtracted
			if (V_flag == 0)
				Display /K=1 /N=BleachingAnalysis_Subtracted ; AppendImage M_Subtracted
				AutoPositionWindow /M=0 /R=BleachingAnalysis_Original BleachingAnalysis_Subtracted
			endif
			
			ModifyGraph /W=BleachingAnalysis_Original width={Aspect, (DimSize(Viewer_Temp, 0) / DimSize(Viewer_Temp, 1))}, height=0
			ModifyGraph /W=BleachingAnalysis_Subtracted width={Aspect, (DimSize(Viewer_Temp, 0) / DimSize(Viewer_Temp, 1))}, height=0
			
			firstRun = 0
		endif
		 
		 if (i != 0)	// for all frames except the first, work with a waiting list
			 // did we localize any emitters that are on the waiting list?
			 for (j = 0; j < DimSize(M_WaitingList, 0); j+=1)
			 	// did we localize this position again?
			 	err = returnPointNearestFitPositions(POS_Out, M_WaitingList[j][xCol], M_WaitingList[j][yCol], nearestX, nearestY, closestIndex)
			 	
			 	if ((err != 0) || sqrt((nearestX - M_WaitingList[j][xCol])^2 + (nearestY - M_WaitingList[j][yCol])^2) > maxJumpDistance * fp.PSFWidth)	// is it close enough to count as the same emitter?
			 		// delete this point from the waiting list, we can't reproduce it
			 		DeletePoints j, 1, M_WaitingList
			 		j -= 1
			 		continue
			 		
			 	else		// the point corresponds to the same emitter as the one on the waiting list
			 			// add it to the list of recovered emitters, and remove it from the fitted positions list (otherwise it will end up on the waiting list again)
			 		Redimension /N=(DimSize(M_BleachingAnalysis, 0) + 1, DimSize(M_BleachingAnalysis, 1)) M_BleachingAnalysis
			 		M_BleachingAnalysis[DimSize(M_BleachingAnalysis, 0) - 1][] = POS_out[closestIndex][q]
			 		M_BleachingAnalysis[DimSize(M_BleachingAnalysis, 0) - 1][0] = i
			 		
			 		// add its contribution
			 		AddEmitter(M_SummedEmitters, POS_out[closestIndex][intensityCol], POS_out[closestIndex][widthCol], POS_out[closestIndex][xCol], POS_out[closestIndex][yCol])
			 		
			 		DeletePoints closestIndex, 1, POS_out
			 		DeletePoints j, 1, M_WaitingList
			 		j -= 1
			 	endif
			 endfor
		else	// for the very first frame, don't use the waiting list but just add all fitted points to the output waves
			// requested by Rob 30102009
			
			offset = DimSize(M_BleachingAnalysis, 0)
			Redimension /N=(DimSize(M_BleachingAnalysis, 0) + DimSize(POS_Out, 0), DimSize(M_BleachingAnalysis, 1)) M_BleachingAnalysis
	 		
	 		for (j = 0; j < DimSize(POS_Out, 0); j+=1)
	 			M_BleachingAnalysis[j + offset][] = POS_Out[j][q]
	 			M_BleachingAnalysis[j + offset][0] = i
	 			AddEmitter(M_SummedEmitters, POS_out[j][intensityCol], POS_out[j][widthCol], POS_out[j][xCol], POS_out[j][yCol])
	 		endfor
			
		endif
		 
		 // all points that are in the fitting list now count as new potential candidates
		 // add them to the waiting list
		 if (DimSize(POS_out, 0) > 0)
			Redimension /N=(DimSize(POS_out, 0), DimSize(POS_out, 1)) M_WaitingList
			M_WaitingList = POS_out
		endif
		 
		 DoUpdate
	endfor
	
	// now reverse the order of the fitted positions so the positions are located according to incrementing frame number
	Duplicate /FREE/O M_BleachingAnalysis, ShiftedBleaching
	ShiftedBleaching = M_BleachingAnalysis[DimSize(M_BleachingAnalysis, 0) - 1 - p][q]
	M_BleachingAnalysis = ShiftedBleaching
	
	KillWaves /Z POS_out
	
	ChangeCurrentImage(previousLoadedImage, baseWindowName)
	
End

Function AddEmitter(image, integral, width, xPos, yPos)
	wave image
	variable integral, width, xPos, yPos
	
	SubtractEmitter(image, -1 * integral, width, xPos, yPos)
End

Function SubtractEmitter(image, integral, width, xPos, yPos)
	wave image
	variable integral, width, xPos, yPos
	
	Assert(WaveExists(image))
	Assert((width > 0) && (xPos >= 0) && (yPos >= 0))
	
	// given image, subtract the contribution of an emitter centered at the given position with the given characteristics
	// we don't make an effort to check if the emitter is really there or not, we just assume that it is
	
	variable startP, endP
	variable startQ, endQ
	variable i, j
	variable xSize = DimSize(image, 0)
	variable ySize = DimSize(image, 1)
	variable value, distance, width_exp, amplitude
	
	startP = floor(xPos - 4 * width)
	endP = ceil(xPos + 4 * width)
	startQ = floor(yPos - 4 * width)
	endQ = ceil(yPos + 4 * width)
	
	startP = (startP >= 0) ? startP : 0
	startQ = (startQ >= 0) ? startQ : 0
	endP = (endP < xSize) ? endP : xSize - 1
	endQ = (endQ < ySize) ? endQ : ySize - 1
	
	amplitude = integral / (2 * pi * width * width)
	width_exp = 2 * width^2
	
	for (i = startP; i <= endP; i+=1)
		for (j = startQ; j <= endQ; j+=1)
			distance = (i - xPos)^2 + (j - yPos)^2
			value = amplitude * exp(-distance / width_exp)
			image[i][j] -= value
		endfor
	endfor
	
End

Function GetCCDDimensionsFromPositions(positions, xSize, ySize, pixelSize)
	wave positions
	variable &xSize, &ySize, &pixelSize
	
	Assert(WaveExists(positions))
	
	// given a wave containing positions, try to deduce the x and y size (in pixels) of the original data based on metadata
	// if the pixel size is known then extract that as well, otherwise return 0
	// the primary focus of this function is to isolate upstream functions from having to manipulate the wavenote directly
	
	xSize = str2num(StringByKey("X SIZE", note(positions)))
	ySize = str2num(StringByKey("Y SIZE", note(positions)))
	
	if ((NumType(xSize) == 2) || (NumType(ySize) == 2))	
		// we didn't find the xSize and/or ySize from the waveNote. 
		// Try to get the x- and y-sizes from the variables defined when loading the images
		Abort "Unable to deduce the dimensions of the original images. Try re-analyzing the data"
	endif
	
	if (numtype(NumberByKey("X PIXEL SIZE", note(positions))) != 2)
		pixelSize = NumberByKey("X PIXEL SIZE", note(positions))
	else
		pixelSize = 0
	endif
End

ThreadSafe Function GetColumnsForEmitterPositions(pos, xCol, yCol, zCol)
	wave pos
	variable &xCol, &yCol, &zCol
	
	// The different types of localization routines in the software all return their own types of parameters for every positions
	// depending on what kind of information is offered by the routine
	// This means that the column containing the x, y, and possibly z positions could be located in different columns of the positions wave
	// this function will inspect the type of positions and return the column indices containing that information by reference
	// if a particular type does not support one of these then -1 will be returned
	
	// the kind of localization used in generating the positions must be set in the wave note
	
	// special case: if LOCALIZATION METHOD is not set, then assume that the positions originate from a symmetric 2D Gaussian
	// this might be the case if the user is using positions fitted with a previous version of the package
	if (NumType(NumberByKey("LOCALIZATION METHOD", note(pos))) == 2)	// NaN
		xCol = 3
		yCol = 4
		zCol = -1
		// append the information for future use
		Note /NOCR pos, "LOCALIZATION METHOD:" + num2str(LOCALIZATION_GAUSS_FITTING) + ";"
	endif
	
	switch (NumberByKey("LOCALIZATION METHOD", note(pos)))
		case LOCALIZATION_GAUSS_FITTING:
			xCol = 3
			yCol = 4
			zCol = -1
			break
		case LOCALIZATION_GAUSS_FITTING_FIX:
			xCol = 2
			yCol = 3
			zCol = -1
			break
		case LOCALIZATION_CENTROID:
			xCol = 1
			yCol = 2
			zCol = -1
			break
		case LOCALIZATION_MULTIPLICATION:
			xCol = 2
			yCol = 3
			zCol = -1
			break
		case LOCALIZATION_ZEISSPALM:
			xCol = 2
			yCol = 3
			zCol = -1
			break
		case LOCALIZATION_ELLIPSOIDAL2DGAUSS:
			xCol = 4
			yCol = 5
			zCol = -1
			break
		case LOCALIZATION_MLEWG:
			xCol = 3
			yCol = 4
			zCol = -1
			break
		default:
			Print "Unable to deduce the kind of localized positions (check the wavenote)"
			return -1
			break
		endswitch
End

Function returnPointNearestFitPositions(positionsWave, xLoc, yLoc, nearestX, nearestY, index)
	wave positionsWave
	variable xLoc, yLoc
	variable& nearestX
	variable& nearestY
	variable& index
	
	Assert(WaveExists(positionsWave))
	Assert((xLoc >= 0) && (yLoc >= 0))
	
	// given the coordinates of a point x, y, the function returns (by reference) the coordinates of the closest matching point in positionsWave and its index in positionsWave
	variable closest_index = -1
	variable nPos = DimSize(positionsWave, 0), xCol, yCol, zCol
	
	if (nPos == 0)
		nearestX = sqrt(-1)	// return NaN if no positions are found
		nearestY = sqrt(-1)
		return 1
	endif
	
	// get the indices of the columns containing the x and y coordinates
	getColumnsForEmitterPositions(positionsWave, xCol, yCol, zCol)
	if ((xCol == -1) || (yCol == -1))
		Abort "The positions passed to returnPointNearestFitPositions() do not appear to contain any (x,y) information"
	endif
	
	MatrixOP /O/FREE W_SqDistances = magSqr(col(positionsWave, xCol) - xLoc) + magSqr(col(positionsWave, yCol) - yLoc)
	WaveStats /Q/M=1 W_SqDistances
	closest_index = V_minLoc
	
	nearestX = positionsWave[closest_index][xCol]
	nearestY = positionsWave[closest_index][yCol]
	index = closest_index
	
	return 0
End

Function /WAVE GetNPointsNearestPosition(positionsWave, xLoc, yLoc, nNearestPoints)
	wave positionsWave
	variable xLoc, yLoc, nNearestPoints
	
	Assert(WaveExists(positionsWave))
	Assert((xLoc >= 0) && (yLoc >= 0) && (nNearestPoints > 0))
	
	// given the coordinates of a point x, y, the function returns  the coordinates of the nNearestPoints closest matching points in positionsWave.
	// the coordinates are returned in an nPoints x 3 wave, with the format index / nearestX / nearestY.
	// If number of point returned will be min(nNearestPoints, nPositionsInWave).
	
	variable nPos = DimSize(positionsWave, 0)
	if (nPos == 0)
		Make /N=0 /FREE /D M_NearestPositions
		return M_NearestPositions
	endif
	
	variable nPositionsToInclude = min(nNearestPoints, nPos)
	
	// get the indices of the columns containing the x and y coordinates
	variable xCol, yCol, zCol
	getColumnsForEmitterPositions(positionsWave, xCol, yCol, zCol)
	if ((xCol == -1) || (yCol == -1))
		Abort "The positions passed to returnPointNearestFitPositions() do not appear to contain any (x,y) information"
	endif
	
	MatrixOP /O/FREE W_SqDistances = magSqr(col(positionsWave, xCol) - xLoc) + magSqr(col(positionsWave, yCol) - yLoc)
	Make /FREE/N=(nPos)/I/U W_Indices = p
	Sort W_SqDistances, W_SqDistances, W_Indices
	
	Make /FREE /N=(nPositionsToInclude, 3) /D M_NearestPositions
	M_NearestPositions[][0] = W_Indices[p]
	M_NearestPositions[][1] = positionsWave[W_Indices[p]][xCol]
	M_NearestPositions[][2] = positionsWave[W_Indices[p]][yCol]
	return M_NearestPositions
End

Function GetColumnsForLocalizationError(pos, xDeviation, yDeviation, zDeviation)
	wave pos
	variable &xDeviation, &yDeviation, &zDeviation
	
	Assert(WaveExists(pos))
	
	// The different types of localization routines in the software all return their own types of parameters for every positions
	// depending on what kind of information is offered by the routine
	// This means that the column containing the deviations on the x, y, and possibly z positions could be located in different columns of the positions wave
	// this function will inspect the type of positions and return the column indices containing that information by reference
	// if a particular type does not support one of these then -1 will be returned
	
	// the kind of localization used in generating the positions must be set in the wave note
	
	// special case: if LOCALIZATION METHOD is not set, then assume that the positions originate from a symmetric 2D Gaussian
	// this might be the case if the user is using positions fitted with a previous version of the package
	if (NumType(NumberByKey("LOCALIZATION METHOD", note(pos))) == 2)	// NaN
		xDeviation = 8
		yDeviation = 9
		zDeviation = -1
		// append the information for future use
		Note /NOCR pos, "LOCALIZATION METHOD:" + num2str(LOCALIZATION_GAUSS_FITTING) + ";"
	endif
	
	switch (NumberByKey("LOCALIZATION METHOD", note(pos)))
		case LOCALIZATION_GAUSS_FITTING:
			xDeviation = 8
			yDeviation = 9
			zDeviation = -1
			break
		case LOCALIZATION_GAUSS_FITTING_FIX:
			xDeviation = 6
			yDeviation = 7
			zDeviation = -1
			break
		case LOCALIZATION_CENTROID:
			xDeviation = -1
			yDeviation = -1
			zDeviation = -1
			break
		case LOCALIZATION_MULTIPLICATION:
			xDeviation = -1
			yDeviation = -1
			zDeviation = -1
			break
		case LOCALIZATION_ZEISSPALM:
			xDeviation = 4
			yDeviation = 4
			zDeviation = -1
			break
		case LOCALIZATION_ELLIPSOIDAL2DGAUSS:
			xDeviation = 11
			yDeviation = 12
			zDeviation = -1
			break
		case LOCALIZATION_MLEWG:
			xDeviation = 6
			yDeviation = 6
			zDeviation = -1
			break
		default:
			Abort "Unable to deduce the kind of localized positions (check the wavenote)"
			break
		endswitch
End

Function GetColumnForNFramesPresent(pos, nFramesPresentCol)
	wave pos
	variable &nFramesPresentCol
	
	Assert(WaveExists(pos))
	
	// The different types of localization routines in the software all return their own types of parameters for every positions
	// depending on what kind of information is offered by the routine
	// This means that the column containing the number of frames it is present could be located in a different column of the positions wave
	// this function will inspect the type of positions and return the column indices containing that information by reference
	// if a particular type does not support one of these then -1 will be returned
	
	// the kind of localization used in generating the positions must be set in the wave note
	
	// special case: if LOCALIZATION METHOD is not set, then assume that the positions originate from a symmetric 2D Gaussian
	// this might be the case if the user is using positions fitted with a previous version of the package
	if (NumType(NumberByKey("LOCALIZATION METHOD", note(pos))) == 2)	// NaN
		nFramesPresentCol = 11
		// append the information for future use
		Note /NOCR pos, "LOCALIZATION METHOD:" + num2str(LOCALIZATION_GAUSS_FITTING) + ";"
	endif
	
	switch (NumberByKey("LOCALIZATION METHOD", note(pos)))
		case LOCALIZATION_GAUSS_FITTING:
			nFramesPresentCol = 11
			break
		case LOCALIZATION_GAUSS_FITTING_FIX:
			nFramesPresentCol = 9
			break
		case LOCALIZATION_CENTROID:
			nFramesPresentCol = 3
			break
		case LOCALIZATION_MULTIPLICATION:
			nFramesPresentCol = 4
			break
		case LOCALIZATION_ZEISSPALM:
			nFramesPresentCol = 5
			break
		case LOCALIZATION_ELLIPSOIDAL2DGAUSS:
			nFramesPresentCol = 15
			break
		case LOCALIZATION_MLEWG:
			nFramesPresentCol = 7
			break
		default:
			Abort "Unable to deduce the kind of localized positions (check the wavenote)"
			break
		endswitch
End

Function GetColumnForIntegratedIntensity(pos, integral)
	wave pos
	variable &integral
	
	Assert(WaveExists(pos))
	
	// The different types of localization routines in the software all return their own types of parameters for every positions
	// depending on what kind of information is offered by the routine
	// This means that the column containing integrated intensity could be located in different columns of the positions wave
	// this function will inspect the type of positions and return the column indices containing that information by reference
	// if a particular type does not support one of these then -1 will be returned
	
	// the kind of localization used in generating the positions must be set in the wave note
	
	// special case: if LOCALIZATION METHOD is not set, then assume that the positions originate from a symmetric 2D Gaussian
	// this might be the case if the user is using positions fitted with a previous version of the package
	if (NumType(NumberByKey("LOCALIZATION METHOD", note(pos))) == 2)	// NaN
		integral = 1
		// append the information for future use
		Note /NOCR pos, "LOCALIZATION METHOD:" + num2str(LOCALIZATION_GAUSS_FITTING) + ";"
	endif
	
	switch (NumberByKey("LOCALIZATION METHOD", note(pos)))
		case LOCALIZATION_GAUSS_FITTING:
		case LOCALIZATION_GAUSS_FITTING_FIX:
			integral = 1
			break
		case LOCALIZATION_CENTROID:
			integral = -1
			break
		case LOCALIZATION_MULTIPLICATION:
			integral = -1
			break
		case LOCALIZATION_ZEISSPALM:
			integral = 1
			break
		case LOCALIZATION_ELLIPSOIDAL2DGAUSS:
			integral = 1
			break
		case LOCALIZATION_MLEWG:
			integral = 1
			break
		default:
			Abort "Unable to deduce the kind of localized positions (check the wavenote)"
			break
		endswitch
End

Function GetColumnForFittedWidth(pos, xWidthColumn, yWidthColumn)
	wave pos
	variable &xWidthColumn, &yWidthColumn
	
	Assert(WaveExists(pos))
	
	// The different types of localization routines in the software all return their own types of parameters for every positions
	// depending on what kind of information is offered by the routine
	// This means that the columns containing the fitted widths could be located in different columns of the positions wave
	// this function will inspect the type of positions and return the column indices containing that information by reference
	// if a particular type does not support one of these then -1 will be returned
	
	// the kind of localization used in generating the positions must be set in the wave note
	
	// special case: if LOCALIZATION METHOD is not set, then assume that the positions originate from a symmetric 2D Gaussian
	// this might be the case if the user is using positions fitted with a previous version of the package
	if (NumType(NumberByKey("LOCALIZATION METHOD", note(pos))) == 2)	// NaN
		xWidthColumn = 2
		yWidthColumn = 2
		// append the information for future use
		Note /NOCR pos, "LOCALIZATION METHOD:" + num2str(LOCALIZATION_GAUSS_FITTING) + ";"
	endif
	
	switch (NumberByKey("LOCALIZATION METHOD", note(pos)))
		case LOCALIZATION_GAUSS_FITTING:
			xWidthColumn = 2
			yWidthColumn = 2
			break
		case LOCALIZATION_GAUSS_FITTING_FIX:
			xWidthColumn = -1
			yWidthColumn = -1
			break
		case LOCALIZATION_CENTROID:
			xWidthColumn = -1
			yWidthColumn = -1
			break
		case LOCALIZATION_MULTIPLICATION:
			xWidthColumn = -1
			yWidthColumn = -1
			break
		case LOCALIZATION_ZEISSPALM:
			xWidthColumn = -1
			yWidthColumn = -1
			break
		case LOCALIZATION_ELLIPSOIDAL2DGAUSS:
			xWidthColumn = 2
			yWidthColumn = 3
			break
		case LOCALIZATION_MLEWG:
			xWidthColumn = 2
			yWidthColumn = 3
			break
		default:
			Abort "Unable to deduce the kind of localized positions (check the wavenote)"
			break
		endswitch
End

Function /S RecursiveWaveList(startDF, nameFilter, optionsFilter)
	DFREF startDF
	string nameFilter, optionsFilter
	
	// get a list of the waves in the requested data folder
	DFREF savDF = GetDataFolderDFR()
	SetDataFolder startDF
	
	string matchingWaves = WaveList(nameFilter, ";", optionsFilter)
	variable nMatches = ItemsInList(matchingWaves)
	variable i
	string matchingWavesFullPath = ""
	
	for (i = 0; i < nMatches; i+=1)
		wave matchingWaveRef = startDF:$StringFromList(i, matchingWaves)
		matchingWavesFullPath += GetWavesDataFolder(matchingWaveRef, 2 ) + ";"
	endfor
	
	// for every data folder in this one, recurse and append the result
	string DFName
	for (i = 0; ; i+=1)
		DFName = GetIndexedObjNameDFR(startDF, 4, i)
		if (strlen(DFName) == 0)
			break
		endif
		
		DFRef subFolder = startDF:$DFName
		
		matchingWavesFullPath += RecursiveWaveList(subFolder, nameFilter, optionsFilter)
		
	endfor
	
	SetDataFolder savDF
	return matchingWavesFullPath
End

Function Assert(condition, [errMessage])
	variable condition
	string errMessage
	
	string abortMessage
	
	if (!condition)
		if (ParamIsDefault(errMessage))
			abortMessage = "An assert failed"
		else
			abortMessage = "Assert failed: " + errMessage
		endif
		Debugger
		Abort abortMessage
	endif
End

Function ZapDataInFolderTree(path)
	String path

	String savDF= GetDataFolder(1)
	SetDataFolder path

	KillWaves/A/Z
	KillVariables/A/Z
	KillStrings/A/Z

	Variable i
	Variable numDataFolders = CountObjects(":",4)
	for(i=0; i<numDataFolders; i+=1)
		String nextPath = GetIndexedObjName(":",4,i)
		ZapDataInFolderTree(nextPath)
	endfor

	SetDataFolder savDF
End

Function /S PrettyPrintTimeElapsed(secondsElapsed)
	variable secondsElapsed
	
	Assert(secondsElapsed >= 0)
	
	variable days = 0, hours = 0, minutes = 0, seconds = 0
	variable remainder = secondsElapsed
	string prettyString = ""
	
	if (remainder >= 24 * 3600)
		days = floor(remainder / (24 * 3600))
		if (days > 1)
			prettyString += num2str(days) + " days, "
		else
			prettyString += num2str(days) + " day, "
		endif
		remainder -= days * 24 * 3600
	endif
	
	if (remainder >= 3600)
		hours = floor(remainder / 3600)
		if (hours > 1)
			prettyString += num2str(hours) + " hours, "
		else
			prettyString += num2str(hours) + " hour, "
		endif
		remainder -= hours * 3600
	endif
	
	if (remainder >= 60)
		minutes = floor(remainder / 60)
		if (minutes > 1)
			prettyString += num2str(minutes) + " minutes, "
		else
			prettyString += num2str(minutes) + " minute, "
		endif
		remainder -= minutes * 60
	endif
	
	// always show the seconds
	if (remainder != 1)
		prettyString += num2str(remainder) + " seconds"
	else
		prettyString += num2str(remainder) + " second"
	endif
	
	return prettyString
End

Function LBDeleteDoubleClickedRowProc(lba) : ListBoxControl
	STRUCT WMListboxAction &lba

	Variable row = lba.row
	Variable col = lba.col
	WAVE/T/Z listWave = lba.listWave
	WAVE/Z selWave = lba.selWave

	switch( lba.eventCode )
		case -1: // control being killed
			break
		case 3: // double click
			
			DeletePoints /M=0 row, 1, listWave
			if (WaveExists(selWave))
				DeletePoints /M=0 row, 1, selWave
			endif
			break
		case 4: // cell selection
		case 5: // cell selection plus shift key
			break
		case 6: // begin edit
			break
		case 7: // finish edit
			break
	endswitch

	return 0
End

// *********************************************
// ************** Progress Viewer **************
// *********************************************

Function ProgressFunctionPrototype(num, denom)
	variable num, denom
End

Function ProgressReporterFunc(num, denom)
	variable num, denom
	return ProgressWindow("Calculating", num, denom, allowAbort = 0)
End

// a hierarchic progress window based on that posted to IgorExchange by RGerkin,
// but with entirely separate code and improved functionality

constant kProgressBarWidth = 250
constant kProgressBarHeight = 20
constant kProcessLabelWidth = 100
constant kProcessTextBoxWidth = 100
constant kMargin = 10
constant kButtonHeight = 20
constant kProgressBarMaxDepth = 10
constant kAbortButtonHeight = 20
constant kAbortButtonWidth = 75

Static Function ToPixels(val)
	variable val
	// convert a dimension in points to pixels
	return val * (ScreenResolution/72)
End

Static Function ToPoints(val)
	variable val
	// convert a dimension in points to pixels
	return val * (72/ScreenResolution)
End

Function ProgressWindow(processName, num, denom, [allowAbort])
	string processName
	variable num, denom, allowAbort
	
	// if allowAbort is set then execute "Abort" when the user click the abort button
	// if not then return kUserAbort. Default is to abort.
	variable doAbort
	if (ParamIsDefault(allowAbort))
		doAbort = 1
	else
		doAbort = allowAbort
	endif
	
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:ProgressViewer
	
	String /G root:Packages:ProgressViewer:S_DisplayedProcesses
	SVAR displayedProcesses = root:Packages:ProgressViewer:S_DisplayedProcesses
	
	Variable /G root:Packages:ProgressViewer:V_abortRequested
	NVAR abortRequested = root:Packages:ProgressViewer:V_abortRequested
	
	wave /Z W_PreviousTicks = root:Packages:ProgressViewer:W_PreviousTicks
	if (WaveExists(W_PreviousTicks) == 0)
		Make /D/N=0 root:Packages:ProgressViewer:W_PreviousTicks
		wave W_PreviousTicks = root:Packages:ProgressViewer:W_PreviousTicks
	endif
	
	// the index in the ticks wave is the same as that of processName in displayedProcesses
	variable index
	index = WhichListItem(processName, displayedProcesses)
	if (index == -1)
		// this is a new process, there is no tick count at which it was last called
		// make an entry in the wave
		Redimension /N=(DimSize(W_PreviousTicks, 0) + 1) W_PreviousTicks
		// and store the current tick count before continuing on
		W_PreviousTicks[DimSize(W_PreviousTicks, 0) - 1] = ticks
	else
		if (ticks - W_PreviousTicks[index] < 10)
			return 0
		else
			W_PreviousTicks[index] = ticks
		endif
	endif
	
	string controlName
	
	// if processName contains a ':' then that will lead to errors
	if (StringMatch(processName, "*:*"))
		Abort "Error: processName cannot contain a ':')"
	endif
	if ((StringMatch(CleanupName(processName, 0), processName) != 1) || (strlen(processName) >= 30))
		Abort "Error: process names must be valid Igor names and be shorter than 30 characters)"
	endif
	
	if (abortRequested != 0)
		if (doAbort != 0)
			Abort
		else
			return kUserAbort
		endif
	endif
	
	DoWindow /F ProgressViewer
	if (V_flag != 1)
		// window does not exist
		NewPanel /N=ProgressViewer /W=(300,300, 300 + kProcessLabelWidth + kProgressBarWidth + kProcessTextBoxWidth + 4 * kMargin, 330) /K=1 as "Calculating..."
		Button BTAbort, win=ProgressViewer, pos={3 * kMargin + kProcessLabelWidth + kProgressBarWidth, kMargin}, size={kAbortButtonWidth, kAbortButtonHeight}, title="Abort", proc=BTAbortProgressViewer
		Execute /P/Q "KillDataFolder root:Packages:ProgressViewer; DoWindow /K ProgressViewer" // Automatic cleanup of data folder and progress window at the end of function execution.  
	endif
	
	// check if the current process already exists
	// otherwise set up the necessary controls
	variable controlTop, additionalWindowHeight, abortButtonTop
	if (WhichListItem(processName, displayedProcesses) == -1)
		// new process, make room
		if (ItemsInList(displayedProcesses) > kProgressBarMaxDepth)
			Abort "Error: too many processes"
		endif
		if (ItemsInList(displayedProcesses) == 0)
			// when the window is created, Igor gives it a height of at least 30 pixels
			// so that needs to be adapted for
			additionalWindowHeight = ToPoints(kProgressBarHeight + 3 * kMargin + kAbortButtonHeight - 30)
			controlTop = kMargin
			abortButtonTop = controlTop + kProgressBarHeight + kMargin
		else
			GetWindow ProgressViewer, wsize
			additionalWindowHeight = ToPoints(kProgressBarHeight + kMargin)
			controlTop = ToPixels(V_Bottom - V_top) - (kMargin + 20)
			abortButtonTop = controlTop + kMargin + kProgressBarHeight
		endif
		GetWindow ProgressViewer, wsize
		MoveWindow /W=ProgressViewer V_Left, V_Top, V_Right, V_Bottom + ToPoints(additionalWindowHeight)
		controlName = "TBT" +  processName
		TitleBox $controlName, win=ProgressViewer, pos={kMargin, controlTop}, fixedSize=1, size={kProcessLabelWidth, 20}, title=processName
		controlName = "PB" + processName
		ValDisplay $controlName, win=ProgressViewer, pos={2 * kMargin + kProcessLabelWidth, controlTop},size={kProgressBarWidth,kProgressBarHeight},limits={0,1,0},barmisc={0,0}, mode=3
		controlName = "TBV" + processName
		TitleBox $controlName, win=ProgressViewer, pos={3 * kMargin + kProcessLabelWidth + kProgressBarWidth, controlTop}, fixedSize=1, size={kProcessTextBoxWidth, 20}, title="hello"
		Button BTAbort, win=ProgressViewer, pos={3 * kMargin + kProcessLabelWidth + kProgressBarWidth, abortButtonTop}
		
		displayedProcesses += processName + ";"
	endif
	
	// update the requested process
	string progressText = num2str(num) + "/" + num2str(denom)
	controlName = "PB" + processName
	ValDisplay $controlName, win=ProgressViewer, value = _NUM:(num / denom)
	controlName = "TBV" + processName
	TitleBox $controlName, win=ProgressViewer, title=progressText
	
	DoUpdate /E=1 /W=ProgressViewer
	return 0
End

Function BTAbortProgressViewer(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	switch( ba.eventCode )
		case 2: // mouse up
			// click code here
				NVAR abortRequested = root:Packages:ProgressViewer:V_abortRequested
				abortRequested = 1
			break
	endswitch

	return 0
End

// *********************************************
// **************** Color Scale Sliders ***************
// *********************************************

Function AppendColorRangeSliders()
	
	string topGraphName = WinName(0, 1, 1)
	if (strlen(topGraphName) == 0)
		Abort "No graph window to append to"
	endif
	
	string imagesInGraph = ImageNameList(topGraphName, ";")
	if (ItemsInList(imagesInGraph) == 0)
		Abort "No image is shown in the top graph"
	endif
	if (ItemsInList(imagesInGraph) > 1)
		Abort "More than one image is present; not yet supported"
	endif
	
	// if the panel is already shown, do nothing
	DoUpdate /W=topGraphName#ColorScaleSliderPanel
	if (V_flag != 0)
		return 0
	endif
	
	string imageName = StringFromList(0, imagesInGraph)
	wave imageWave = ImageNameToWaveRef(topGraphName, imageName)
	
	// initialize the panel
	NewPanel /N=ColorScaleSliderPanel/W=(0,0,360,200) /EXT=0 /HOST=$topGraphName
	PopupMenu PMSelectColorTable,win=$topGraphName#ColorScaleSliderPanel,pos={15,8},size={205,20},bodyWidth=150,title="Color Table:"
	PopupMenu PMSelectColorTable,win=$topGraphName#ColorScaleSliderPanel,mode=1,value=CtabList(),proc=PMColorTableChanged
	Slider SLUpperLimit,win=$topGraphName#ColorScaleSliderPanel,pos={18,81},size={316,13},proc=SlidersModified
	Slider SLUpperLimit,win=$topGraphName#ColorScaleSliderPanel,limits={0,2,1},value= 0,side= 0,vert= 0
	CheckBox CBReverseColors,win=$topGraphName#ColorScaleSliderPanel,pos={243,10},size={98,14},title="Reverse colortable"
	CheckBox CBReverseColors,win=$topGraphName#ColorScaleSliderPanel,value= 0,proc=CBReverseColorTable
	GroupBox GBUpperLimit,win=$topGraphName#ColorScaleSliderPanel,pos={12,34},size={335,73},title="Upper limit"
	SetVariable SVUpperLimitSliderMin,win=$topGraphName#ColorScaleSliderPanel,pos={23,58},size={75,15},proc=SVSetVariablesModified
	SetVariable SVUpperLimitValue,win=$topGraphName#ColorScaleSliderPanel,pos={140,57},size={75,15},proc=SVSetVariablesModified
	SetVariable SVUpperLimitSliderMax,win=$topGraphName#ColorScaleSliderPanel,pos={263,57},size={75,15},proc=SVSetVariablesModified
	Slider SLLowerLimit,win=$topGraphName#ColorScaleSliderPanel,pos={19,160},size={316,13},proc=SlidersModified
	Slider SLLowerLimit,win=$topGraphName#ColorScaleSliderPanel,limits={0,2,1},value= 0,side= 0,vert= 0
	GroupBox GBLowerLimit,win=$topGraphName#ColorScaleSliderPanel,pos={13,113},size={335,73},title="Lower limit"
	SetVariable SVLowerLimitSliderMin,win=$topGraphName#ColorScaleSliderPanel,pos={24,137},size={75,15},proc=SVSetVariablesModified
	SetVariable SVLowerLimitValue,win=$topGraphName#ColorScaleSliderPanel,pos={141,136},size={75,15},proc=SVSetVariablesModified
	SetVariable SVLowerLimitSliderMax,win=$topGraphName#ColorScaleSliderPanel,pos={264,136},size={75,15},proc=SVSetVariablesModified
	
	// update the limits of the sliders to the min and max of the image
	variable minLimit = WaveMin(imageWave)
	variable maxLimit = WaveMax(imageWave)
	
	Slider SLUpperLimit,win=$topGraphName#ColorScaleSliderPanel,limits={minLimit, maxLimit, 0}
	Slider SLLowerLimit,win=$topGraphName#ColorScaleSliderPanel,limits={minLimit, maxLimit, 0}
	SetVariable SVUpperLimitSliderMax,win=$topGraphName#ColorScaleSliderPanel,value=_NUM:maxLimit
	SetVariable SVUpperLimitSliderMin,win=$topGraphName#ColorScaleSliderPanel,value=_NUM:minLimit
	SetVariable SVLowerLimitSliderMax,win=$topGraphName#ColorScaleSliderPanel,value=_NUM:maxLimit
	SetVariable SVLowerLimitSliderMin,win=$topGraphName#ColorScaleSliderPanel,value=_NUM:minLimit
		
	
	// and update the controls to the actual settings
	UpdateControlsToActualSettings(topGraphName)
End

Function UpdateControlsToActualSettings(windowName)
	string windowName
	
	string baseWindowName = StringFromList(0, windowName, "#")
	
	string imagesInGraph = ImageNameList(baseWindowName, ";")
	if (ItemsInList(imagesInGraph) == 0)
		Abort "No image is shown in the top graph"
	endif
	if (ItemsInList(imagesInGraph) > 1)
		Abort "More than one image is present; not yet supported"
	endif
	
	string imageName = StringFromList(0, imagesInGraph)
	wave imageWave = ImageNameToWaveRef(baseWindowName, imageName)
	
	// get the color scale that is currently selected
	string recreation = StringByKey("RECREATION", ImageInfo(baseWindowName, imageName, 0))
	string ctabStr = StringByKey("ctab", recreation, "=")
	
	string lowerLimitStr, upperLimitStr, ctabName, isReverseStr
	variable lowerLimit, upperLimit, isReverse
	
	SplitString  /E="\\{([^,]*),([^,]*),([^,]*),([0-9]+)\\}" ctabStr, lowerLimitStr, upperLimitStr, ctabName, isReverseStr
	
	if (GrepString(lowerLimitStr, "\\*"))
		lowerLimit = WaveMin(imageWave)
	else
		lowerLimit = str2num(lowerLimitStr)
	endif
	
	if (GrepString(upperLimitStr, "\\*"))
		upperLimit = WaveMax(imageWave)
	else
		upperLimit = str2num(upperLimitStr)
	endif
	
	isReverse = str2num(isReverseStr)
	
	// and update the controls to the actual settings
	PopupMenu PMSelectColorTable,win=$baseWindowName#ColorScaleSliderPanel,popmatch=ctabName
	CheckBox CBReverseColors,win=$baseWindowName#ColorScaleSliderPanel,value=isReverse
	Slider SLUpperLimit,win=$baseWindowName#ColorScaleSliderPanel,value=upperLimit
	Slider SLLowerLimit,win=$baseWindowName#ColorScaleSliderPanel,value=lowerLimit
	
	SetVariable SVUpperLimitValue,win=$baseWindowName#ColorScaleSliderPanel,value=_NUM:upperLimit
	SetVariable SVLowerLimitValue,win=$baseWindowName#ColorScaleSliderPanel,value=_NUM:lowerLimit

End

Function SlidersModified(sa) : SliderControl
	STRUCT WMSliderAction &sa

	switch( sa.eventCode )
		case -1: // control being killed
			break
		default:
			if( sa.eventCode & 1 ) // value set
				Variable curval = sa.curval
				string baseWindowName = StringFromList(0, sa.win, "#")
				
				// update both SetVariables
				ControlInfo /W=$sa.win SLUpperLimit
				SetVariable SVUpperLimitValue,win=$sa.win,value=_NUM:V_Value
				
				ControlInfo /W=$sa.win SLLowerLimit
				SetVariable SVLowerLimitValue,win=$sa.win,value=_NUM:V_Value
				
				UpdateColorScaling(sa.win)
			endif
			break
	endswitch

	return 0
End

Function SVSetVariablesModified(sva) : SetVariableControl
	STRUCT WMSetVariableAction &sva

	switch( sva.eventCode )
		case 1: // mouse up
		case 2: // Enter key
		case 3: // Live update
			Variable dval = sva.dval
			
			ControlInfo /W=$sva.win SVUpperLimitSliderMax
			variable upperLimitMax = V_Value
			ControlInfo /W=$sva.win SVUpperLimitSliderMin
			variable upperLimitMin = V_Value
			ControlInfo /W=$sva.win SVUpperLimitValue
			variable upperLimit = V_Value
			ControlInfo /W=$sva.win SVLowerLimitSliderMax
			variable lowerLimitMax = V_Value
			ControlInfo /W=$sva.win SVLowerLimitSliderMin
			variable lowerLimitMin = V_Value
			ControlInfo /W=$sva.win SVLowerLimitValue
			variable lowerLimit = V_Value
			
			Slider SLUpperLimit,win=$sva.win,value=upperLimit,limits={upperLimitMin, upperLimitMax, 0}
			Slider SLLowerLimit,win=$sva.win,value=lowerLimit,limits={lowerLimitMin, lowerLimitMax, 0}
			break
		case -1: // control being killed
			break
	endswitch

	return 0
End

Function PMColorTableChanged(pa) : PopupMenuControl
	STRUCT WMPopupAction &pa

	switch( pa.eventCode )
		case 2: // mouse up
			Variable popNum = pa.popNum
			String popStr = pa.popStr
			
			UpdateColorScaling(pa.win)
			
			break
		case -1: // control being killed
			break
	endswitch

	return 0
End

Function CBReverseColorTable(cba) : CheckBoxControl
	STRUCT WMCheckboxAction &cba

	switch( cba.eventCode )
		case 2: // mouse up
			Variable checked = cba.checked
			
			UpdateColorScaling(cba.win)
			
			break
		case -1: // control being killed
			break
	endswitch

	return 0
End


Function UpdateColorScaling(windowName)
	string windowName
	
	string baseWindowName = StringFromList(0, windowName, "#")
	
	string imagesInGraph = ImageNameList(baseWindowName, ";")
	if (ItemsInList(imagesInGraph) == 0)
		Abort "No image is shown in the top graph"
	endif
	if (ItemsInList(imagesInGraph) > 1)
		Abort "More than one image is present; not yet supported"
	endif
	
	string imageName = StringFromList(0, imagesInGraph)
	wave imageWave = ImageNameToWaveRef(baseWindowName, imageName)
	
	// get the color scale to use
	string ctabName
	ControlInfo /W=$baseWindowName#ColorScaleSliderPanel PMSelectColorTable
	ctabName = S_Value
	
	variable minVal, maxVal
	ControlInfo /W=$baseWindowName#ColorScaleSliderPanel SLUpperLimit
	maxVal = V_Value
	
	ControlInfo /W=$baseWindowName#ColorScaleSliderPanel SLLowerLimit
	minVal = V_Value
	
	variable isReverse
	ControlInfo /W=$baseWindowName#ColorScaleSliderPanel CBReverseColors
	isReverse = V_Value
	
	ModifyImage /W=$baseWindowName $imageName, ctab={minVal, maxVal, $ctabName, isReverse}
End
